// ====================================
//  NAME    : FWUKL1 Library
// 
//  VERSION : 2
// 
//  REVISION: 5
// 
//  DATE    : 2006/10/12
// 
//  RELEASE : 2008/04/03
// 
//  PLATFORM: PVSS II
// 
//  AUTHOR  : Gareth Rogers
// 
//  EMAIL   : rogers@hep.phy.cam.ac.uk
// ====================================

/* \mainpage UKL1 Library Documentation
 *
 * This library is used to control the UKL1 boards.
 */

// ====================
//  #USES DECLARATIONS
// ====================

/*! This UKL1 library contains all the functions to configure a UKL1 board. */
#uses "fwUkl1/fwUkl1DPInterface.ctl"

/*! Contains the call back functions that are available for the UKL1s. */
#uses "fwUkl1/fwUkl1CB.ctl"

/*! fwUkl1ExcepionHandling library is used to store and display errors that occur with fwUkl1 function calls. */
#uses "fwUkl1ExceptionHandling/fwUkl1ExceptionHandling.ctl"

/*! fwHw library is used for recipe and data point manipulation. */
#uses "fwHw.ctl"

/*! fwCcpc library is used for accessing the UKL1 registers. */
#uses "fwCcpc.ctl"

/*! fwFSMConfDB library is used to add and remove the FSM confDB object to the FSM tree. */
#uses "fwFSMConfDB/fwFsmConfDB.ctl"


// =========================
//  CONSTANTs & GLOBAL VARs
// =========================

/** @defgroup SectionConstants Constants
 *   List of constants define for various parameters used by the UKL1 controls software.
 *   All are intended for library and user use, there are no 'internal' constants.
 *  @{
 */

/*! Defines the name of the data point that holds the UKL1 exception log. */
const string FWUKL1_EXCEPTION_LOG = "Ukl1Exceptions";

/*! Defines the name of the data point type of the UKL1 hardware type. All UKL1 board data points are created of this type. */
const string FWUKL1_HW_TYPE = "UKL1";

/*! Defines the name of the recipe type that is associated with the configure action. */
const string FWUKL1_CONFIGURE_TYPE = "UKL1_Configure";
/*! Defines the name of the recipe type that is associated with the start and stop actions. */
const string FWUKL1_START_TYPE = "UKL1_Start";

/*! Broadcast FE FPGA number. Value used to write to all the FE FPGAs. */
const unsigned FWUKL1_BROADCAST_FPGA = 5;
/*! Broadcast FE FPGA channel number. Value used to write to all of a FE FPGA's channels. */
const unsigned FWUKL1_BROADCAST_CHANNEL = 10;
/*! Number of Front end (ingress) FPGAs on the RICH L1 board. */
const unsigned FWUKL1_FRONT_FPGAS = 4;
/*! Number of channels that are on each front end FPGA. */
const unsigned FWUKL1_FPGA_CHANNELS = 9;
/*! Number of channels that are on the RICH L1 board. */
const unsigned FWUKL1_BOARD_CHANNELS = 36;
/*! Number of ports on the gigabit ethernet card. */
const unsigned FWUKL1_GBE_PORTS = 4;
/*! Number of temperature sensors present on the board. */
const unsigned FWUKL1_TEMPERATURE_SENSORS = 3;

/*! This numeric constant is used to identify a production UKL1 board. */
const unsigned FWUKL1_PRODUCTION = 2;
/*! This numeric constant is used to identify a prototype UKL1 board. */
const unsigned FWUKL1_PROTOTYPE = 1;

/*! Constant that defines the size of a 32-bit data word. */
const unsigned FWUKL1_32BIT_WORD = 32;
/*! Constant that defines the size of a 16-bit data word. */
const unsigned FWUKL1_16BIT_WORD = 16;
/*! Constant that defines the size of an 8-bit data word. */
const unsigned FWUKL1_8BIT_WORD = 8;

/*! Constant to specify the type of the FSM configurator object to be added to the UKL1 tree. */
const string FWUKL1_FSMCONFDB_TYPE = "FwFSMConfDB_DAQ";

/*!
 * The constants specify the position of the register data and the extra data that is associated
 * with a register name in the map.
 */
const int FWUKL1_MAP_REG_DATA = 1;
const int FWUKL1_MAP_XTR_DATA = 2;

// @}


// ================
//  UKL1 INTERFACE
// ================

/** @defgroup SectionInterface Read/write interface.
 *   Provides access to the either the UKL1 hardware or the recipe settings for that board.
 *
 *   In both the hardware access and register access cases maps are used to associated the
 *   register name to the data read or to be written. The data returned or take by the functions
 *   is intended to be as human reable as possible and can be directly displayed on a widget
 *   or taken directly from a widget. The for of the text is described in the fwUkl1DP control
 *   library.
 *
 *   The recipe read and writes differ from hardware writes as they are grouped by recipe types.
 *   Not all hardware registers are present in all recipe types and as a result not all the
 *   hardware registers can be read from the all, if any, of the recipes.
 *   
 */
// @{

/*!
 * Writes to a given set of the UKL1 registers on a given UKL1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  inData A mapping associating the register name (string) with the data that is to be written, which is
 *           a dyn_string. The data should be in element FWUKL1_MAP_REG_DATA of the dyn_string.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful by reading back the registers
 *           and comparing it to that given.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_write(const string &ukl1Name, const mapping &inData, bool verifyWrite, dyn_string &exceptionInfo) {
  // Keep track of the exception info.
  int exInfoSize = dynlen(exceptionInfo);
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_write(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)

  // Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  // Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  // data point elements.
  mapping outData;
  _fwUkl1_parseRegistersForSet(inData, outData, exceptionInfo);
  // Check that we were actually able to parse the data. This is not necessarily fatal.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_write(): Failed to parse input data, not all the data was written.", "2", exInfoSize);

  // This is the form of the data that we require to write.
  dyn_string regNames;
  dyn_dyn_char ddcData;
  dyn_dyn_char masks;
  dyn_string writeFunctions;
  _fwUkl1_parseWriteMapping(outData, regNames, ddcData, masks, writeFunctions, exceptionInfo);
  // An error from here is not necessarily fatal.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_write(): Failed to convert all the data from that was parsed to a form that can be written. Some registers will not be written.", "2", exInfoSize);

  _fwUkl1_writeSwitch(ukl1Name, regNames, writeFunctions, ddcData, masks, verifyWrite, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_write(): Failed to perform so/all of the writes to the hardware. Not all of the settings will have been updated.", "2", exInfoSize);

  // Done.
  return;
}// fwUkl1_write()

/*!
 * Reads from a given set of the UKL1 registers on a given UKL1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  outData Returns by reference a mapping with the register name (string) as the key
 *           and a dyn_string string as the value. The dyn_string will contain the register value
 *           in the element FWUKL1_MAP_REG_DATA and a colour that identifies if the value is in
 *           a sensible range in the element FWUKL1_MAP_XTR_DATA. The map should contain as keys the
 *           registers that are to be read. The colour FwStateAttention1 is used to indicate if the
 *           read failed and FwStateAttention2 if the value is out of bounds.
 * \param  forStatus If the values read back are to be used to display the status of the register then it
 *           might have a different form to that to be used for the configuration. If TRUE then the
 *           values read back might not be able to be written to the configuration register, FALSE they
 *           must be.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_read(const string &ukl1Name, mapping &outData, bool forStatus, dyn_string &exceptionInfo) {
  // Keep track of the exception info.
  int exInfoSize = dynlen(exceptionInfo);
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_read(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)

  // This is the form of the data that we require to read.
  dyn_string regNames;
  dyn_string readFunctions;
  // Just give read mapping the list of registers that are to be read.
  // We pass FALSE as we will want any alternative hardware registers.
  // The alternative registers are for those cases where the read register
  // does not match the write register. It is none standard because the DB
  // always uses the same registers for read and write which must be the
  // same as the hardware writes.
  _fwUkl1_parseReadMapping(mappingKeys(outData), regNames, readFunctions, FALSE, exceptionInfo);
  // An error from here is not necessarily fatal.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_read(): Failed to determine the names of the some/all of the hardware registers to read from. Read operation will cotinue, but not all the expected data will be returned.", "2", exInfoSize);

  // Now we call the function that will decide how to read the data based on the information we have just received.
  // This is the map to associate the raw data with what ever extra information we are interested in passing around.
  // At the moment, none!
  mapping rawData;
  _fwUkl1_readSwitch(ukl1Name, rawData, regNames, readFunctions, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_read(): Some or all of the registers were not read back from the hardware.", "2", exInfoSize);
  
  // Now we need to parse this data. Do it regardless of whether there was a failure, it will sort out what happened with the failures.
  _fwUkl1_parseRegistersForGet(rawData, outData, forStatus, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_read(): Failed to parse all of the raw data read from the hardware, so registers probably failed to read correctly.", "2", exInfoSize);
  // Done.
  return;
}// fwUkl1_read()

/*!
 * This functions works much like read, it takes a list of abstract registers and reads their values from the hardware, but it does not
 * parse this data. Instead it just updates the HW DPs and lets a call back function handle them if necessary.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  regList List of abstract registers that are to read.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_refresh(const string &ukl1Name, const dyn_string &regList, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // First loop over the abstract registers and get the related hardware function to use to read that register.
  // Just give read mapping the list of registers that are to be read.
  dyn_string hwRegNames;
  dyn_string readFunctions;
  _fwUkl1_parseReadMapping(regList, hwRegNames, readFunctions, FALSE, exceptionInfo);
  // An error from here is not necessarily fatal.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_refresh(): Failed to determine the names of the some/all of the hardware registers to read from. Read operation will cotinue, but not all the expected data will be returned.", "2", exInfoSize);

  // Now we have to split the data to be read up such that it is read with the appropriate function.
  // This is the map to associate the raw data with what ever extra information we are interested in passing around.
  // At the moment, none!
  mapping rawData;
  _fwUkl1_readSwitch(ukl1Name, rawData, hwRegNames, readFunctions, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_refresh(): Some or all of the registers were not read back from the hardware.", "2", exInfoSize);
  
  // Done.
  return;
}// fwUkl1_refresh()

/*!
 * This saves a given set of registers to the configurationDB for a given UKL1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  inData This is the same as that described in fwUkl1_write() documention.
 * \param  recipeName The name of the recipe to save the data to. It must be of the form "RUN_TYPE|ExtraInfo/FsmAction", where:
 *   \li RUN_TYPE is a descriptiion of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION. All capitals.
 *   \li ExtraInfo is any further detail that is required to specify the name uniquely, this (and the |) are optional
 *         in the event they are present this recipe is searched for. If they are missing the recipe
 *         "RUN_TYPE/FsmAction" is searched for.
 *   \li FsmAction is the action of FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 * \param  recipeType The recipe type that it is to be saved to, this can be anything provided it already exists.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_save(const string &ukl1Name, const mapping &inData, const string &recipeName, const string &recipeType, dyn_string &exceptionInfo) {
  // Keep track of the exception info.
  int exInfoSize = dynlen(exceptionInfo);
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_save(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)

  // Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  // Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  // data point elements.
  mapping outData;
  _fwUkl1_parseRegistersForSet(ukl1Name, inData, outData, exceptionInfo);
  // Check that we were actually able to parse the data.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_save(): Failed to parse input data, not all the data was written.", "2", exInfoSize);

  // This is the form of the data that we require to write.
  dyn_string regNames;
  dyn_dyn_char ddcData;
  dyn_dyn_char masks;
  dyn_string junk; // We always use the same save function and don't care which register it uses to write.
  _fwUkl1_parseWriteMapping(outData, regNames, ddcData, masks, junk, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_save(): Failed to convert all the data from that was parsed to a form that can be written. Some registers will not be written.", "2", exInfoSize);
  
  // The data and registers have been parsed, so let us write it to the recipe.
  // Our save recipe function requires the masks to have already been appended to the data.
  const int numElements = dynlen(ddcData);
  for (int element = 1; element <= numElements; ++element) {
    dynAppend(ddcData[element], masks[element]);
  }// for element
  // Now save the recipe, domain is empty.
  _fwUkl1_save(ukl1Name, recipeName, recipeType, regNames, ddcData, "", exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_save(): Failed to save recipe for " + ukl1Name + ".", "1", exInfoSize);

  // Done.
  return;
}// fwUkl1_save()

/*!
 * Loads a recipe from a given set of the UKL1 registers on a given UKL1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  outData A mapping associating the register name with the data that is to be read and any extra
 *           information required to decode the data from the hardware register. At present the extra data
 *           is a PVSS colour name that can be displayed as well as the value. It will indicate the 'health'
 *           of the value returned.
 * \param  recipeName The name of the recipe to save the data to. It must be of the form "RUN_TYPE|ExtraInfo/FsmAction", where:
 *   \li RUN_TYPE is a descriptiion of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION. All capitals.
 *   \li ExtraInfo is any further detail that is required to specify the name uniquely, this (and the |) are optional
 *         in the event they are present this recipe is searched for. If they are missing the recipe
 *         "RUN_TYPE/FsmAction" is searched for.
 *   \li FsmAction is the action of FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 * \param  recipeType The recipe type that it is to be saved to, this can be anything provided it already exists.
 * \param  domain This is required if the recipe to be loaded has been loaded into the recipe cache by the FSMConfDB. In this case
 *           the name of the recipe will have been mangled. The recipe name given to this function should be the unmangled
 *           name and it will be converted as appropriate. This parameter can also be an empty string if no name mangling is required.
 *           Note in this case there can be no / present in the recipe name, only to delimit the run mode from the command (RUN_MODE/Command).
 * \param  forStatus If the values read back are to be used to display the status of the register then it
 *           might have a different form to that to be used for the configuration. If TRUE then the
 *           values read back might not be able to be written to the configuration register, FALSE they
 *           must be.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_load(const string &ukl1Name, mapping &outData, const string &recipeName, const string &recipeType, string domain, bool forStatus, dyn_string &exceptionInfo) {
  // Keep track of the exception info.
  int exInfoSize = dynlen(exceptionInfo);
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_load(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)

  // This is the form of the data that we require to read.
  dyn_string regNames;
  dyn_dyn_char ddcData;
  // This is the read function names that we don't care about as all are loaded with the same function.
  dyn_string junk;
  // Pass TRUE as we want to get the standard hw reg names.
  _fwUkl1_parseReadMapping(mappingKeys(outData), regNames, junk, TRUE, exceptionInfo);
  // An error from here is not necessarily fatal.
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_load(): Failed to determine the names of the some/all of the hardware registers to read from. Read operation will cotinue, but not all the expected data will be returned.", "2", exInfoSize);

  // This is the map to associate the raw data with what ever extra information we are interested in passing around.
  // At the moment, none!
  mapping rawData;
  // Now load them from the DB.
  _fwUkl1_load(ukl1Name, recipeName, recipeType, regNames, ddcData, domain, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_load(): Failed to load from some/all of the registers for " + ukl1Name + " with the _fwUkl1_load function.", "1", exInfoSize);
  // Now append the registers to the full lists for later parsing.
  // Do this even if there is an error as we don't know how much failed.
  const int numRegs = dynlen(regNames);
  const int lenData = dynlen(ddcData);
  for ( int regNum = 1; regNum <= numRegs; ++regNum ) {
    const string regName = regNames[regNum];
    if ( lenData >= regNum )
      rawData[regName] = ddcData[regNum];
    else
      rawData[regName] = makeDynChar();
  }// regNum

  // Now we need to parse this data...
  _fwUkl1_parseRegistersForGet(rawData, outData, forStatus, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_load(): Failed to parse all of the raw data read from the hardware, so registers probably failed to read correctly.", "2", exInfoSize);
  // Done.
  return;
}// fwUkl1_load()

/*!
 * This function is responsible for starting the monitoring of the given list of abstract registers.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  regList List of registers to monitor for this UKL1 board.
 * \param  exceptionInfo Error information. No change in size if no exception inforamtion is generated.
 * \return void.
 */
void fwUkl1_startMonitoring(const string &ukl1Name, const dyn_string &regList, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // First loop through the registers and get the hw regs and their monitoring functions.
  const int numRegs = dynlen(regList);
  dyn_string monFuncs;
  dyn_string hwRegList;
  for (int regNum=1; regNum<=numReg; ++regNum) {
    // Now get the hardware register list.
    const dyn_string hwRegs = fwUkl1_getMonitoringRegisterList(regList[regNum], exceptionInfo);
    if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_startMonitoring(): Could not identify the corresponding hardware registers to monitor for the register " + regList[regNum] + ". This function will not be monitored.", "2", exInfoSize)) {
      // If successful get the monitoring function.
      const string monFunc = fwUkl1_getStartMonitoringFunction(regList[regNum], exceptionInfo);
      if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_startMonitoring(): Could not identify the function to use to monitor the register " + regList[regNum] + ". This function will not be monitored.", "2", exInfoSize)) {
        for (int i=1; i<=dynlen(hwRegs); ++i) {
          if (0 < dynContains(hwRegList, hwRegs[i])) {
            dynAppend(hwRegList, hwRegs[i]);
            dynAppend(monFuncs, monFunc);
          }// if(0<dynContains(hwRegList,hwRegs[i])
        }// loop over hwRegs
      }// got mon func successfully.
    }// got hw reg successfully.
  }// for regNum
  
  // Now sort the lists for the monitoringSwitch.
  dyn_mixed lists = makeDynMixed(monFuncs, regList);
  fwUkl1_sort(lists, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_startMonitoring(): Failed to sort the hardware list, all registers should be monitored but it might take longer to setup.", "2", exInfoSize);
  
  _fwUkl1_monitoringSwitch(ukl1Name, hwRegList, monFuncs, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_startMonitoring(): Failed to start monitoring all of the requested registers.", "2", exInfoSize);
  // Done.
  return;
}// fwUkl1_startMonitoring()

/*!
 * This function is responsible for stopping the monitoring of the given list of abstract registers.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  regList List of registers to monitor for this UKL1 board.
 * \param  exceptionInfo Error information. No change in size if no exception inforamtion is generated.
 * \return void.
 */
void fwUkl1_stopMonitoring(const string &ukl1Name, const dyn_string &regList, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // First loop through the registers and get the hw regs and their monitoring functions.
  const int numRegs = dynlen(regList);
  dyn_string monFuncs;
  dyn_string hwRegList;
  for (int regNum=1; regNum<=numReg; ++regNum) {
    // Now get the hardware register list.
    // We want to monitor those registers that are being read from so take the alternative hw reg.
    // Hence we pass false.
    const string hwReg = fwUkl1_getHwRegName(regList[regNum], FALSE, exceptionInfo);
    if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_stopMonitoring(): Could not identify the corresponding hardware register to stop monitoring for the register " + regList[regNum] + ". Monitoring will not be stopped for this function.", "2", exInfoSize)) {
      // If successful get the monitoring function.
      const string monFunc = fwUkl1_getStopMonitoringFunction(regList[regNum], exceptionInfo);
      if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_stopMonitoring(): Could not identify the function to use to stop monitoring the register " + regList[regNum] + ". Monitoring will not be stopped for this function.", "2", exInfoSize)) {
        for (int i=1; i<=dynlen(hwRegs); ++i) {
          if (0 < dynContains(hwRegList, hwRegs[i])) {
            dynAppend(hwRegList, hwRegs[i]);
            dynAppend(monFuncs, monFunc);
          }// if(0<dynContains(hwRegList,hwRegs[i])
        }// loop over hwRegs
      }// got mon func successfully.
    }// got hw reg successfully.
  }// for regNum
  
  // Now sort the lists for the monitoringSwitch.
  dyn_mixed lists = makeDynMixed(monFuncs, regList);
  fwUkl1_sort(lists, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_stopMonitoring(): Failed to sort the hardware list, monitoring should be stopped for all registers but it might take longer to setup.", "2", exInfoSize);
  
  _fwUkl1_monitoringSwitch(ukl1Name, hwRegList, monFuncs, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_stopMonitoring(): Failed to stop monitoring all of the requested registers.", "2", exInfoSize);
  // Done.
  return;
}// fwUkl1_stopMonitoring()

/*!
 * This function starts the monitoring of the registers that are read directly by the ccserv.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * This function is to be called when a board is added to the system and it will monitor all but the Ingress FPGA and channel
 * register using the ccserv. The Ingress FPGA and channel registers are FIFOs and hence must be readout using a predefined
 * sequence in PVSS. They cannot be monitored by the ccserv.
 */
void fwUkl1_startMonitoringAllMonitorableRegs(const string &ukl1Name, dyn_string &exceptionInfo) {
  // Keep track of the exceptionInfo.
  int exInfoSize = dynlen(exceptionInfo);
  // Get the registers from the appropriate groups. Not worth monitoring any TTCrx stuff.
  dyn_string groups = makeDynString(  FWUKL1_GBE_CONFIG_GROUP, FWUKL1_GBE_STATUS_GROUP
                                    , FWUKL1_EGRESS_CONFIG_GROUP, FWUKL1_EGRESS_STATUS_GROUP
                                    , FWUKL1_INGRESS_CONFIG_GROUP, FWUKL1_INGRESS_STATUS_GROUP
                                    , FWUKL1_TEMPERATURE_SENSORS, FWUKL1_PROM_GROUP
                                    , FWUKL1_TFC_STATUS_GROUP
                                    , FWUKL1_E_IN_LINK_STATUS_GROUP
                                    , FWUKL1_ALARM_STATUS_GROUP
                                    );
  // This holds the names of the registers to monitor. It has to be a map even though we are only interested
  // in the names as that is what the function takes.
  mapping regMap;
  // These are given in the place of a recipe name and the registers that are reject due to not being in the recipe.
  // We are not interested in those for monitoring purposes.
  dyn_string dsJunk;
  string sJunk;
  fwUkl1_getRegistersInGroups(groups, sJunk, regMap, dsJunk, exceptionInfo);
  if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_startMonitoringAllMonitorableRegs(): Failed to get the registers for " + ukl1Name + ". No registers will be monitored on this UKL1.", "2", exInfoSize) ) {
    // Create a register and the monitoring settings for each of the found boards.
    dyn_string regList = mappingKeys(regMap);
    // Take the current monitoring settings that were configured when the registers were created.
//    if ( !fwUkl1_startMonitoring(regList) )
//      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_startMonitoringAllMonitorableRegs(): Failed to start monitoring the registers for " + ukl1Name + ". These registers will not be monitored an it must be started manually on the appropriate panel.", "2");
  }// if fwUkl1_getUkl1NamesList no exceptionInfo
  // Done.
  return;
}// fwUkl1_startMonitoringStatusRegs()

// ==============
//   INTERNAL
// ==============

/*!
 * Takes the list of hardware registers to write and their associated write functions and then calls the appropriate write.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  regNames The list of registers to write to.
 * \param  writeFunctions List of functions to use to write the data. This should correspond (by index) to the register names to write.
 * \param  ddcData List containing the dyn_char that contains the bytes that are to be written to the register. Should be associated
 *           by index to the register names to write.
 * \param  masks List containing the dyn_char that contains the bytes that are to be written to the register. Should be associated
 *           by index to the register names to write.
 * \param  verifyWrite TRUE if the data written is to be read back to confirm the write was successful, FALSE if not. This is mainly
 *           done on the ccserv side and provides little over head. Ingress FPGA and channel register will require an extra read.
 * \param  exceptionInfo Error information. No change in size if there is no exception information generated.
 * \return void.
 *
 * Where possible this function will group writes that use the same write function together.
 */
void _fwUkl1_writeSwitch(const string &ukl1Name, const dyn_string regNames, const dyn_string &writeFunctions, const dyn_dyn_char &ddcData, const dyn_dyn_char &masks, bool verifyWrite, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  int numRegs = dynlen(writeFunctions);
  // These hold the registers to be written to a specific function.
  dyn_string tmpRegNames;
  dyn_dyn_char tmpDdcData;
  dyn_dyn_char tmpMasks;
  // This is the function that the group is to be written with.
  string writefunc;
  // This is use to determine how many registers to write to a given function.
  int count;
  // This is an index that runs through all the element in the lists.
  int i = 1;
  // This is an index that is used to place the data in the appropriate place in the temp lists.
  // Run over write functions.
  while ( i <= numRegs ) {
    writefunc = writeFunctions[i];
    // Loop through the writeFunctions array and get all the matching writes.
    int j = 1;
    while ( writefunc == writeFunctions[i] ) {
      tmpRegNames[j] = ukl1Name + regNames[i];
      tmpDdcData[j]  = ddcData[i];
      tmpMasks[j]    = masks[i];
      ++i;
      ++j;
      if ( (i > numRegs) || (j > numRegs) )
        break;
    }// while(writefunc==writeFunctions[i])
    // Now write them.
    switch (writefunc) {
      case "_fwUkl1_write":
        _fwUkl1_write(tmpRegNames, tmpDdcData, masks, verifyWrite, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_writeSwitch(): Failed to write to some/all of the registers for " + ukl1Name + " with the _fwUkl1_write.", "1", exInfoSize);
        break;
      case "_fwUkl1_writeIngressFpgaChannel":
        _fwUkl1_writeIngressFpgaChannel(tmpRegNames, tmpDdcData, masks, verifyWrite, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_writeSwitch(): Failed to write to some/all of the registers for " + ukl1Name + " with the _fwUkl1_writeIngressFpgaChannel.", "1", exInfoSize);
        break;
      case FWUKL1_READ_ONLY:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_writeSwitch(): Attempted to write to read only registers for " + ukl1Name + ". Those writes will be ignored.", "2");
        break;
      default:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeSwitch(): Did not recognise the write function " + writefunc + ". Could not write the data associated with this function (first register " + tmpRegNames[1] + ").", "2");
    }// switch(writefunc)
  }// while(i<numRegs)
  // Done.
  return;
}// _fwUkl1_writeSwitch()

/*!
 * Takes the list of hardware registers and their associated read functions and then calls the appropriate read.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  rawData This is the map that will contain the dyn_char data associated with each of the hardware registers.
 *           This map should be empty and contain no keys. Any existing entries will either be overwritten or ignored
 *           and this may produce fake warnings.
 * \param  regNames List of the hardware registers to read.
 * \param  readFunction List of the functions that should be used to read the associated (by index) hardware register.
 * \param  exceptionInfo Error information. No change in size if there is no exception information is generated.
 * \return void.
 *
 * Where possible this function will group all the reads that use the same read function together.
 *
 * The read will return an entry in the map for the null register FWUKL1_NULL_HWREG. This data is empty and effectively
 * junk. It is easier if everything can rely on this register being there.
 */
void _fwUkl1_readSwitch(const string &ukl1Name, mapping &rawData, const dyn_string &regNames, const dyn_string &readFunctions, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // Now we have to split the data to be read up such that it is read with the appropriate function.
  int numRegs = dynlen(readFunctions);
  // These hold the registers to be read with a specific function.
  dyn_string tmpRegNames;
  dyn_string tmpRegNamesNoUkl1Name;
  dyn_dyn_char tmpDdcData;
  // This is the function that the group is to be read with.
  string readfunc;
  // This is use to determine how many registers to read with a given function.
  int count;
  // This is an index that runs through all the element in the lists.
  int i = 1;
  // This is an index that is used to place the data in the appropriate place in the temp lists.
  // Run over read functions.
  while (i <= numRegs) {
    readfunc = readFunctions[i];
    // Loop through the readFunctions array and get all the matching reads.
    int j = 1;
    while (readfunc == readFunctions[i]) {
      // Get the hardware register and add the UKL1 name before it.
      tmpRegNames[j] = ukl1Name + regNames[i];
      // Save the hardware register name without the UKL1 name. This makes parsing easier later.
      tmpRegNamesNoUkl1Name[j] = regNames[i];
      ++i;
      ++j;
      if (j > numRegs)
        break;
    }// while(readfunc==readFunctions[i])
    // Now read them.
    switch (readfunc) {
      case "_fwUkl1_read":
        _fwUkl1_read(tmpRegNames, tmpDdcData, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_readSwitch(): Failed to read from some/all of the registers for " + ukl1Name + " with the _fwUkl1_read function.", "1", exInfoSize);
        break;
      case "_fwUkl1_readAllUkl1Channels":
        _fwUkl1_readAllUkl1Channels(tmpRegNames, tmpDdcData, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_readSwitch(): Failed to read from some/all of the registers for " + ukl1Name + " with the _fwUkl1_readAllUkl1Channels function.", "1", exInfoSize);
        break;
      case FWUKL1_WRITE_ONLY:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_readSwitch(): Attempted to read from a write only registers for " + ukl1Name + ". Those reads will be ignored.", "2");
        break;
        
      default:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_readSwitch(): Did not recognise the read function " + readfunc + ". Could not read the data associated with this function (first register " + tmpRegNames[1] + ").", "2");
        break;
    }// switch(readfunc)
    // Now append the registers to the full lists for later parsing.
    // Do this even if there is an error as we don't know how much failed.
    const int numTmpRegs = dynlen(tmpRegNames);
    for (int regNum=1; regNum<=numTmpRegs; ++regNum) {
      const string regName = tmpRegNamesNoUkl1Name[regNum];
      rawData[regName] = tmpDdcData[regNum];
    }// regNum
  }// while(i<numRegs)
  // Add the null register.
  rawData[FWUKL1_NULL_HWREG] = makeDynChar();
  // Done.
  return;
}// _fwUkl1_readSwitch()

/*!
 * This function looks through the list of functions given to use to start or stop the monitoring and calls the
 * the relevant function for the given registers.
 *
 * \param  regNames Names of the hardware registers to monitor.
 * \param  monFunctions List of functions to use to monitor the corresponding hardware registers.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * This function will perform more efficiently if the two lists are sorted and the registers requiring
 * the same monitoring function are sequential.
 */
void _fwUkl1_monitoringSwitch(const dyn_string &hwRegNames, const dyn_string &monFunctions, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  int numRegs = dynlen(monFunctions);
  // These hold the registers to be read with a specific function.
  dyn_string tmpRegNames;
  // This is the function that the group is to be monitored with.
  string monfunc;
  // This is use to determine how many registers to monitor with a given function.
  int count;
  // This is an index that runs through all the element in the lists.
  int i = 1;
  // This is an index that is used to place the data in the appropriate place in the temp lists.
  // Run over monitor functions.
  while ( i <= numRegs ) {
    monfunc = monFunctions[i];
    // Loop through the monFunctions array and get all the matching reads.
    int j = 1;
    while ( monfunc == monFunctions[i] ) {
      // Get the hardware register.
      tmpRegNames[j] = regNames[i];
      ++i;
      ++j;
      if ( j > numRegs )
        break;
    }// while(monfunc==monFunctions[i])
    // Now read them.
    switch (monfunc) {
      case "fwCcpc_startMonitoring":
        if ( !fwCcpc_startMonitoring(tmpRegNames) )
          exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_monitoringSwitch(): Failed to start monitoring for some/all of the registers with the fwCcpc_startMonitoring function. It does not work in a distributed system.", "2", exInfoSize);
        break;

      case "fwCcpc_stopMonitoring":
        if ( !fwCcpc_stopMonitoring(tmpRegNames) )
          exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_monitoringSwitch(): Failed to stop monitoring for some/all of the registers with the fwCcpc_stopMonitoring function. It does not work in a distributed system.", "2", exInfoSize);
        break;
      
      case "fwUkl1_startFifoMonitoring":
      case "fwUkl1_stopFifoMonitoring":
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_monitoringSwitch(): Monitoring of the UKL1 FIFO registers has not yet been implemented.", "2");
        
      default:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "_fwUkl1_monitoringSwitch(): Did not recognise the monitoring function " + monfunc + ". Could not monitor the registers associated with this function (first register " + tmpRegNames[1] + ").", "2");
        break;
    }// switch(monfunc)
  }// while(i<=numRegs)
  
  // Done.
  return;
}// _fwUkl1_monitoringSwitch()

/*!
 * In order to write to specific registers on the FPGAs we must first convert the address
 * to include the address of the FPGAs from the glue card perspective.
 *
 * \param  addr Address of the register to be written to.
 * \return unsigned Full address of the FPGA register.
 */
unsigned _fwUkl1_fpgaizeAddress(unsigned addr) {
  // Based on code written by Steve Wotton, Cambridge.
  // The address sent to lb_{write|read}_hword should be above 0x20000
  // to be outside the GBE & gluecard ranges
  // The FPGA is sensitive only to bits 2-13 of the address
  // (=> the FPGA is not sensitive to the 0x2???? that indicates
  //     that an address is outside the GBE/gluecard ranges)
  // So whenever we want to address the FPGA, set the 13th bit:
  //  the FPGA will only respond to addresses for which the 13th bit has been set
  // 
  // So: << 2    translates a (32-bit) int address into (8-bit) byte address
  //    0x20000 gets us outside the address ranges of the GBE and gluecard
  //              (but will be ignored by the FPGA)
  //    0x2000  should stop the gluecard from reacting to the operation
  //              (but will be ignored by the FPGA)
  // 
  // => (int) register addresses 0x0 to 0x7ff should be usable by the FPGA
  // 
  return (0x08000000 + ( 0x800 | ( addr<<2 ) ) );
}// _fwUkl1_fpgaizeAddress()

/*!
 * This takes a hex string and converts it to a dyn_char that can be written to the UKL1 registers.
 *
 * \param  data A hex string that is to be converted into a dyn_char for writing/masking an FPGA register.
 * \return dyn_char The data that has been converted into an appropriate form for writing/masking.
 */
dyn_char _fwUkl1_fpgaizeData(const string &data) {
  // We must word swap, where a word is 16-bits, the data and add a word every other word.
  // Find the length of the string and loop over it creating a substrings for each words and swap them.
  const int numChars = strlen(data);
  const int numCharsPerWord = 4;
  // Loop over the words and create a reformed string.
  string rData = "";
  for (int index=(numChars-numCharsPerWord); index >= 0; index-=numCharsPerWord)
    rData += "0000" + substr(data, index, numCharsPerWord);// Move the end to the beginning and pad each word to 32-bits.
  // Done.
  return fwCcpc_convertHexToByte(rData);
}// _fwUkl1_fpgaizeData()

/*!
 * Writes to the PHY chip via the MDIO interface.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  phyAddr Address of the PHY chip to write to. Range 0-3.
 * \param  regAddr 5-bit address of the register on the desired PHY chip that it is to be written to.
 * \param  val Value that is to be written to this register.
 * \param  verify Read back the data written to ensure the write was successful. Note that the data written to the PHY chip
 *           is not verified, but the writes required to setup the transfer to the PHY registers are. Source documentation
 *           indicates the clear reasons for this.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_writeGbeMdioRegister(const string &ukl1Name, unsigned phyAddr, unsigned regAddr, unsigned val, bool verify, dyn_string &exceptionInfo) {
  // Keep track of errors.
  int exInfoSize = dynlen(exceptionInfo);
  bool error = FALSE;
  // This will be prepended to all registers that are read or written from.
  const string preRegStr = ukl1Name + ".GBE";
  // Holds the names of the registers to be written to and the data that is read back from them.
  dyn_string regs;
  dyn_dyn_char ddcData;
  dyn_dyn_char masks;
  dyn_int callStatusList;
  
  // First ensure that the MDIO control is enabled, should be enabled as part of initialisation.
  regs[1] = preRegStr + ".MDIOControl";
  if ( fwCcpc_read(regs, ddcData, callStatusList) && (0 != dynlen(ddcData)) ) {
    // Were able to read the MDIO control register now check if it is enabled.
    const string MDIObit = _fwUkl1_convertByteToHex(ddcData[1], 2, 1, FALSE);
    // Check for MDIO control disabled.
    if ("00" == MDIObit) {
      dynClear(regs);
      dynClear(ddcData);
      dynClear(masks);
      // MDIO control is not enabled, enable it.
      regs[1]    = preRegStr + ".MDIOControl";
      ddcData[1] = fwCcpc_convertHexToByte("00000004");
      masks[1]   = fwCcpc_convertHexToByte("00000004");
      _fwUkl1_write(regs, ddcData, masks, verify, exceptionInfo);
      if ( (dynlen(exceptionInfo) > exInfoSize) || (-1 == exInfoSize) ) {
        fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeGbeMdioRegister(): Failed to enable the MDIO control on the GBE card. Further write attempts are not possible.", "1");
        exInfoSize = dynlen(exceptionInfo);
        error = TRUE;
      }// if failed to enable MDIO control.
    }// if MDIO control not enabled or error finding out.
  }// if(fwCcpc_read())
  else {
    // Failed to read from MDIO control register, probably indicates an error.
    string errMsg = "_fwUkl1_writeGbeMdioRegister(): " + regs[1] + " encountered ";
    if ( 0 == dynlen(ddcData) )
      errMsg += "no data read back for the MDIO control register. Further encountered ";
    
    if ( 0 == dynlen(callStatusList) ) {
      errMsg += " an empty call status list, so no further information is available about the error.";
    }// if(0==dynlen(callStatusList)
    else {
      if (callStatusList[1] > 0) {
        errMsg += "a problem writing to the CCPC server, return code " + callStatusList[1] + ".";
      } else if (callStatusList[1] < 0) {
        errMsg += "a problem with PVSS.";
      } else {
        errMsg += "an unknown error.";
      }
    }// else(0==dynlen(callStatusList))
    
    // Now raise the exception.
    fwUkl1Exception_raise(exceptionInfo, "ERROR", errMsg, "1");
    exInfoSize = dynlen(exceptionInfo);
    // Mark the error.
    error = TRUE;
  }// else(fwCcpc_read())
  
  // Ensure these are all cleared for the next attempt.
  dynClear(regs);
  dynClear(ddcData);
  dynClear(masks);
  
  if (!error) {
    // The MDIO control is now enabled so we can write to the desired PHY chip.
    // Frist we write the data to be written to the first register, ensuring that the upper 16-bits are masked as requested in doc.
    masks[1]   = fwCcpc_convertHexToByte("0000ffff");
    regs[1]    = preRegStr + ".PHYData";
    string hexData = fwCcpc_convertDecToHex(val);
    // Now ensure that it is padded to the appropriate length.
    fwUkl1_padString(hexData, 4);
    ddcData[1] = fwCcpc_convertHexToByte(hexData);
    // Now write, but can't verify via this method. Would have to read from the PHY control register
    // on the relevant port and even then if performing a reset as the bit self clears it would probably fail...
    _fwUkl1_write(regs, ddcData, masks, FALSE, exceptionInfo);
    if ( (dynlen(exceptionInfo) <= exInfoSize) && (-1 != exInfoSize) ) {
      // Successfully wrote the data. Now send it.
      dynClear(regs);
      dynClear(ddcData);
      dynClear(masks);
      masks[1]   = fwCcpc_convertHexToByte("ffffffff");
      regs[1]    = preRegStr + ".PHYCtrl";
      val = 0x00110000 | ((phyAddr&0x3)<<8) | (regAddr&0x1f);
      string hexData = fwCcpc_convertDecToHex(val);
      // Now ensure that it is padded to the appropriate length.
      fwUkl1_padString(hexData, 4);
      ddcData[1] = fwCcpc_convertHexToByte(hexData);
      // Now write.
      _fwUkl1_write(regs, ddcData, masks, verify, exceptionInfo);
      if ( (dynlen(exceptionInfo) > exInfoSize) || (-1 == exInfoSize) ) {
        // Failed to trigger sending of PHY data.
        fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeGbeMdioRegister(): Failed to send data to the PHY chip. Data was placed in the buffer and MDIO control enabled, a second attempt maybe succesful.", "1");
        exInfoSize = dynlen(exceptionInfo);
      }// if(!fwCcpc_write())
    }// if(fwCcpc_write())
    else {
      // Failed to write the command for the PHY.
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeGbeMdioRegister(): Failed to write the data to be written to the PHY chip. Write aborted.", "1");
      exInfoSize = dynlen(exceptionInfo);
    }// else(fwCcpc_read())
  }// if(!error)
  
  // Done.
  return;
}// _fwUkl1_writeGbeMdioRegister()

/*!
 * Writes the dyn_char configuration data to a UKL1 Ingress FPGA channel configuration FIFO.
 *
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register.
 * \param  ddcData Each element of this array contains an array of char that make up the words that are to be written to the
 *           register in the form required by a fwCcpc_write.
 * \param  masks The masks are not used for FIFO writes and are ignored by this function.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful by reading back the registers.
 *           FALSE and no read will be done.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_writeIngressFpgaChannel(const dyn_string &registerNames, const dyn_dyn_char &ddcData, const dyn_dyn_char &masks, bool verifyWrite, dyn_string &exceptionInfo) {
  // Keep track of the amount of exception information generated.
  int exInfoSize = dynlen(exceptionInfo);
  // We can only write one channel at a time. The registerNames could
  // refer to multiple channels on multiple boards, so we must do everything
  // for each channel.
  const int numRegs = dynlen(registerNames);
  for (int i=1; i<=numRegs; ++i) {
    // Get the full register name which contains the channel and the UKL1 name.
    const string regName = registerNames[i];
    
    // Find the first full stop, which will indicate the end of the UKL1 name.
    const int stopPos = strpos(regName, ".");
    if (-1 < stopPos) {
      // Now we have found a full stop we must get the UKL1 name, which is between these two markers.
      // The last arugment is the length of the string to strip and hence is the difference
      // between the two markers.
      string ukl1Name = substr(regName, 0, stopPos);
      // Now we can get the UKL1 channel number from the register name.
      int chanPos = strpos(regName, "Channel");
      if (-1<chanPos) {
        // From the strpos we know where the string 'Channel' starts and the channel
        // number will be the next character.
        int channel = substr(regName, chanPos+strlen("Channel"));
        // Covert that channel number into some other forms we use.
        int inFpga = channel/9;
        // Prepare the FIFO for writing.
        _fwUkl1_setConfigForIngressFpgaChannel(ukl1Name, inFpga, verifyWrite, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_writeIngressFpgaChannel(): Failed to prepare the configuration FIFO for Ingress FPGA " + inFpga + ", channel " + (channel%9) + " for writting. Data not written.", "2", exInfoSize)) {
          // Write the data.
          dyn_string mailboxRegs = makeDynString(ukl1Name + ".FpgaConfig" + ".ConfigurationFifoChannel" + channel);
          dyn_dyn_char mailboxData;
          dynAppend(mailboxData, ddcData[i]);
          // We actually don't use the masks as the CCPC server just writes it into the FIFO
          // and this cuases the loss of data. It must be empty.
          dyn_dyn_char mailboxMasks;
//          dynClear(mailboxMasks);
//          dynAppend(mailboxMasks, masks[i]);
          // The given data and mask will not contain the appropriate channel to write to.
          // This must be written in the upper byte.
          // The channel number resides in bits [15..12] of the first word. Select the upper
          // byte nad shift to align with bit 12.
          // and bit shift the channel number by 4 to move it to the upper nibble of this byte.
          mailboxData[1][1] |= channel%9 << 4;
//          mailboxMasks[1][1] |= 0xf << 4;
          // Can't verify the FIFO write.
          _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, FALSE, exceptionInfo);
          if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_writeIngressFpgaChannel(): Failed to write configuration data for channel " + channel + ". Configuration not written.", "2", exInfoSize)) {
            // Now we have written the data we can trigger the sending of it.
            _fwUkl1_sendIngressChannelConfigFifo(ukl1Name, verifyWrite, exceptionInfo);
            fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_writeIngressFpgaChannel(): Triggering the transmission of the configuration data for channel " + channel + ". Data not written.", "2", exInfoSize);
          }// if _fwUkl1_write() successful.
        }// if prepared channel for writing.
      }// if(-1<chanPos)
    // If there is no channel idenfier then it can't be used to write to a channel.
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeIngressFpgaChannel: Could not determine the UKL1 channel number from the register \'" + regName + "\'. It appears to have no .ChannelX, if it does not contain a channel number it cannot be used in this function.", "2");
    }// if(-1<stopPos)
    // If there are no full stops then this can't be a valid register name!
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_writeIngressFpgaChannel: Could not determine the UKL1 name from the register \'" + regName + "\'. It has no full stops, is it correctly formed?", "2");
  }// for i
  
  // Done.
  return;
}// _fwUkl1_writeIngressFpgaChannel()

/*!
 * Prepares the configuration FIFO for writting by setting the appropriate Ingerss FPGA number and asserting and
 * deasserting the pointer resets. It ensures that the transmit trigger has been deasserted.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  inFpga Number of the Ingress FPGAs that are to be read, each element should correspond contain
 *           the number of the Ingress FPGA to read. Range:0 to 3.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_setConfigForIngressFpgaChannel(const string &ukl1Name, int inFpga, bool verifyWrite, dyn_string &exceptionInfo) {
  // Holds the various bits of data that will need to be written.
  // This will only ever write to one register.
  dyn_string mailboxRegs = makeDynString(ukl1Name + ".FpgaConfig.IngressControl");
  dyn_dyn_char mailboxData;
  dyn_dyn_char mailboxMasks;
    
  // Deassert both read and write pointer resets and the transmit trigger (zeros to bits [6..4]),
  // ensuring the broadcast bit is cleared (zeros to bit 7),
  // while writing the Ingress FPGA to be configured (bits [9..8]).
  string tmpData = fwCcpc_convertDecToHex(inFpga << 8);
  fwUkl1_padString(tmpData, 4);
  mailboxData[1] = fwCcpc_convertHexToByte(tmpData);
  // Want to clear bits [7..4] and set bit [9..8].
  mailboxMasks[1] = fwCcpc_convertHexToByte("000003F0");
  _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, verifyWrite, exceptionInfo);
    
  // Assert the read and write pointers' reset.
  mailboxData[1] = fwCcpc_convertHexToByte("00000030");
  // The mask is the same as the data to be written in this case.
  mailboxMasks[1] = mailboxData[1];
  _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, verifyWrite, exceptionInfo);
    
  // Deassert the read and write pointers' reset.
  mailboxData[1] = fwCcpc_convertHexToByte("00000000");
  // The mask is the same as before so we don't need to update it.
  _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, verifyWrite, exceptionInfo);
    
  // Done.
  return;
}// _fwUkl1_setConfigForIngressFpgaChannel()

/*!
 * Triggers the tranmission of the Ingress configuration FIFO data.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_sendIngressChannelConfigFifo(const string& ukl1Name, bool verifyWrite, dyn_string& exceptionInfo) {
  // Trigger the FIFO tranmission to Ingress FPGA from Egress FPGA.
  dyn_string mailboxRegs;
  dyn_dyn_char mailboxData;
  dyn_dyn_char mailboxMasks;
  mailboxRegs[1]  = ukl1Name + ".FpgaConfig.IngressControl";
  mailboxData[1]  = fwCcpc_convertHexToByte("00000040");
  mailboxMasks[1] = fwCcpc_convertHexToByte("00000040");
  _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, verifyWrite, exceptionInfo);

  // Deassert the FIFO tranmission trigger.
  mailboxData[1] = fwCcpc_convertHexToByte("00000000");
  // The mask is the same as before.
  _fwUkl1_write(mailboxRegs, mailboxData, mailboxMasks, verifyWrite, exceptionInfo);
  
  // Done.
  return;
}// _fwUkl1_sendIngressChannelConfigFifo()

/*!
 * Writes the given data to the given UKL1 board. The masks are applied when writing the data to ensure that only specific
 * bits of a register are written to.
 *
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register.
 * \param  ddcData Each element of this array contains an array of char that make up the words that are to be written to the
 *           register in the form required by a fwCcpc_write.
 * \param  masks Each element contains a bit mask that is applied when writing the corresponding element in the ddcData array
 *           to the corresponding register in the registerNames array. It ensure that only the desired bits are written to in
 *           the register.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful by reading back the registers.
 *           FALSE and no read will be done.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_write(const dyn_string &registerNames, const dyn_dyn_char &ddcData, const dyn_dyn_char &masks, bool verifyWrite, dyn_string &exceptionInfo) {
  // Check we are being asked to write something.
  if (0 >= dynlen(registerNames)) {
    fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_write(): List of registers to write to empty, no data written to the hardware.", "2");
    return;
  }// if(0>=dynlen(registerNames))
  // Track the exceptions.
  int exInfoSize = dynlen(exceptionInfo);
  // We will read the data back from the registers once it is written. The read data is stored here.
  dyn_dyn_char readData;
  // Holds the status of the writes.
  dyn_int callStatusList;
//  if (fwCcpc_write(registerNames, ddcData, masks, callStatusList)) {
  if (fwCcpc_writeRead(registerNames, ddcData, masks, readData, callStatusList)) {
    // See if we want to verify that the write was successfull.
    if (verifyWrite) {
      // Call the verification function.
      _fwUkl1_verifyWrite(registerNames, ddcData, masks, readData, exceptionInfo);
      fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_write(): Failed to verify data written written to some/all of the registers.", "2", exInfoSize);
    }// if(verifyWrite)
  }// if(fwCcpc_write)
  else {
    // Check where the errors occured.
    const unsigned statusLen = dynlen(callStatusList);
    if (0 != statusLen) {
      for (unsigned element=1; element<=statusLen; ++element) {
        if (0 != callStatusList[element]) {
          // Construct the error message.
          string errMsg = "_fwUkl1_write(): " + registerNames[element] + " encountered ";
          if (callStatusList[element] > 0)
            errMsg += "a problem writing to the CCPC server, return code: " + callStatusList[element] + ".";
          else if (callStatusList[element] < 0)
            errMsg += "a problem with PVSS.";
          else
            errMsg += "an unknown error.";
          exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", errMsg, "2");
        }// if(0!=callStatusList[element]
      }// for element
    }// if(0!=statusLen)
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_write(): All writes failed.", "1");
  }// else(fwCcpc_write)
  
  // Done.
  return;
}// _fwUkl1_write()

/*!
 * Reads the data from the given UKL1 board.
 *
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  ddcData Each element of this array contains an array of char that make up the words that are read from the
 *           register in the form returned by a fwCcpc_read. Returned by reference.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This function can be called to update the values in the datapoint elements and the data read back ignored. In this case the function
 * is used to convert the fwCcpc exception information into a fwUkl1 exceptionInfo and DP connects to the DPEs can parse the
 * read data if successful.
 */
void _fwUkl1_read(const dyn_string &registerNames, dyn_dyn_char &ddcData, dyn_string &exceptionInfo) {
  // Check we are being asked to read something.
  if (0 >= dynlen(registerNames)) {
    fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_read(): List of registers to read is empty, no data read from the hardware.", "2");
    return;
  }// if(0>=dynlen(registerNames))
  
  // Track the exceptions.
  int exInfoSize = dynlen(exceptionInfo);
  
  // In order to read the status information we must first assert and then deassert the read pointer
  // reset to ensure that we always start from the beginning of the FIFO. Always reset all the pointers.
  // Do it for all reads as it is only two extra writes and means we don't need a separate function
  // for the status FIFOs.
  // Need to get theUKL1 name from the register name. Can include the system name if we wish.
  // Simply remove all the characters from the beginning to the first full stop.
  // We must have at least one register name as we return if the dyn_string has zero length earlier.
  const string ukl1Name = substr(registerNames[1], 0, strpos(registerNames[1], "."));
  fwUkl1_resetIngressFpgaStatusBufPointers(ukl1Name, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_resetIngressFpgaStatusBufPointers(): Failed to reset status buffer pointers, any read status information will be invalid.", "2", exInfoSize);
  // Note what the status of the reads is.
  dyn_int callStatusList;
  // Just read don't bother checking returns the panels will indicate the problems soon enough.
  // We do this as it appears to crash if we try and use ddcData directly. If anyone ever reads
  // this comment they should try using ddcData directly and if it doesn't work figure out why!
//  dyn_dyn_char tmpDdcData;
  if (!fwCcpc_read(registerNames, ddcData, callStatusList)) {
    // There has been an error, check its seriousness
    const int sizeList = dynlen(callStatusList);
    if (0 != sizeList) {
      for (int element=1; element<=sizeList; ++element) {
        if ( 0 != callStatusList[element] ) {
          // Construct the error message.
          string errMsg = "_fwUkl1_read(): Read failed for " + registerNames[element] + ". Encountered ";
          if (callStatusList[element] > 0)
            errMsg += "a problem writing to the CCPC server, return code: " + callStatusList[element] + ".";
          else if (callStatusList[element] < 0)
            errMsg += "a problem with PVSS.";
          else
            errMsg += "an unknown error.";
          // Now raise the exception.
          exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", errMsg, "2");
        }// if(0!=callStatusList[element])
      }// for element.
    }// if(0!=sizeList)
    else {
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_read(): Failed to read, no information read back.", "1");
    }// else(0!=sizeList)
  }// if(!status)
//  ddcData = tmpDdcData;
  // Done.
  return;
}// _fwUkl1_read()

/*!
 * Read the from all of the UKL1 channels.
 *
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  ddcData Each element of this array contains an array of char that make up the words that are read from the
 *           register in the form returned by a fwCcpc_read. Returned by reference.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This will only work for the channel status registers at the moment,
 * however this is currently the only way to read any channel data so
 * it covers all use cases.
 */
void _fwUkl1_readAllUkl1Channels(dyn_string &registerNames, dyn_dyn_char &ddcData, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // In order to read each register from the all channel hardware register to the equivalent four Ingress
  // registers.
  dyn_string allRegNames;
  const int numRegs = dynlen(registerNames);
  for (int reg=1; reg<=numRegs; ++reg) {
    for (int inFpga=0; inFpga<FWUKL1_FRONT_FPGAS; ++inFpga) {
      dynAppend(allRegNames, registerNames[reg] + inFpga);
    }// for inFpga
  }// for reg
  // Over write the registerNames with the new list we have just generated.
  // We know that these can all be read with the standard read, so make that
  // assumption. If the supported register is expanded we many have to call
  // the read switch here in order to sort them out. That requires a lot more
  // function calls however.
  _fwUkl1_read(registerNames, ddcData, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_readAllUkl1Channels(): Not all of the channels' information could be read. Report value may be incomplete.", "2", exInfoSize);
  // Done.
  return;
}// _fwUkl1_readAllUkl1Channels()

/*!
 * Save a list of recipes to the given recipe, which is of the given type. The data should contain the actual value that would be
 * written to the hardware and include the mask that should be used when writing the data.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  recipeName Name of the recipe. The structure for this is described in fwUkl1_save().
 * \param  recipeType Recipe type that is to be saved. The structure for this is described in fwUkl1_save().
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register.
 * \param  ddcDataAndMask Each element of this array contains an array of char that make up the words that are to be written to the
 *           register in the form required by a fwCcpc_write and this includes the mask that is to be used for the data.
 * \param  domain This is required if the recipe to be saved has been loaded into the recipe cache by the FSMConfDB. In this case
 *           the name of the recipe will have been mangled. The recipe name given to this function should be the unmangled
 *           name and it will be converted as appropriate. This parameter can also be an empty string if no name mangling is required.
 *           Note in this case there can be no / present in the recipe name, only to delimit the run mode from the command (RUN_MODE/Command).
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void _fwUkl1_save(const string &ukl1Name, string recipeName, const string &recipeType, const dyn_string &registerNames, const dyn_dyn_char &ddcDataAndMask, string domain, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  if ( "" == ukl1Name ) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_save(): No UKL1 name give to retrieve the recipe from, no data will be retrieved.", "2");
    return;
  }// if(""==ukl1Name)
  
  // Check that we are given a recipeType and a recipeName.
  if ( ("" != recipeType) && ("" != recipeName) ) {
    // We must write to all the registers in the recipe otherwise it will populate them with the last values written to the board.
    // We must therefore get the recipe first and then update the relevant values and write it back to the recipe.
    // Holds a list of all the registers in the recipe to be written.
    dyn_string fullRegisterNames;
    // Holds a list of all the data, including masks, to be written into the recipe.
    dyn_dyn_char fullDdcData;
    // Before we try and get the recipe we should check that it exists, as if it doesn't trying to retrieve it will fail.
    dyn_string recipes = fwHw_getRecipes(ukl1Name);
    // Also note whether or not the recipe already exists.
    bool recipeAlreadyExists;
    // Check to see if the register we are going to write to exists.
    // dynContains returns either the index of the element that the string is contained within, 0 if it does not exists or -1 in the event of an error.
    if ( 0 < dynContains(recipes, recipeName) ) {
      // The recipe does exist, thus we should get it and overwrite the relevant elements.
      recipeAlreadyExists = TRUE;
      if ( -1 != fwHw_getRecipe(ukl1Name, recipeName, fullRegisterNames, fullDdcData, domain) ) {
        const int numElements = dynlen(registerNames);
        for (int element = 1; element <= numElements; ++element) {
  	  // This take the specific register name and returns the index of its occurance in the full list.
  	  // It is this index that we can use to over write the appropriate data element.
          // It assumes there is only one occurrance of a register in the recipe. This should always be the case!
  	  int index = dynContains(fullRegisterNames, registerNames[element]);
  	  // Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
          // It is possible that this data does not occur in this recipe type, hence we must protect against this.
          if ( 0 != index )
            fullDdcData[index] = ddcDataAndMask[element];
          // else
            // fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_save(): " + registerNames[element] + " was not found in recipe type. This data has not been saved.", "1");
          // Don't bother warning about this it generates an unnecessary error, just let it happen silently.
        }// for element
      }// if(-1!=fwHw_getRecipe())
      else {
       exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_save(): Failed to retrieve the recipe `" + recipeName + "' for " + ukl1Name + ". Writing the settings anyway. This could create a recipe with only partially initialised values and potentially a corrupt name.", "1");
       recipeAlreadyExists = FALSE;
     }// else(-1!=fwHw_getRecipe())      
    }// if(dynContains)
    
    if (!recipeAlreadyExists) {
      // As the recipe doesn't exist or couldn't retrieve it we are creating it. Therefore we just give the recipe all the data we have.
      // The save recipe function will place the last values written to the hardware in the other recipe registers.
      // These can be updated by other function calls.
      fullRegisterNames = registerNames;
      fullDdcData = ddcDataAndMask;
    }// if(!recipeAlreadyExists)

    // The recipe needs to know the recipe type that we are using. Has to be for the full list.
    dyn_int regTypes;
    const int fullLen = dynlen(fullRegisterNames);
    for (int i = 1; i <= fullLen; ++i) {
      // All the registers are of the same type might as well be saved as the same type so just populate array with that value. Where is memset when you need it?
      // TODO - Update this to retrieve the type from the DP that I will eventually create...
      regTypes[i] = FWHW_LBUS;
    }// for i

    // If we are given a domain then we should mangle the recipe name to write to the FSMConfigDB cached recipe.
    if ( "" != domain )
      recipeName = _fwHw_getRecipeName(fwUkl1_removeSystemName(ukl1Name), recipeName, domain, fwUkl1_removeDeviceName(ukl1Name));
    // If this function fails and the recipe already exists then there has been an error, however this function could still apparently fail if the recipe
    // doesn't already exist, but it does write the appropriate values to the recipe. Require it fails and the recipe already exists for us to indicate an error.
    if ( (-1 == fwHw_saveRecipe(ukl1Name, recipeType, recipeName, fullRegisterNames, regTypes, fullDdcData)) && recipeAlreadyExists )
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_save(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
  }// if recipe name and type not empty.
  else {
    // Create more useful error message.
    emptyName = ( "" == recipeName ? "empty":recipeName );
    emptyType = ( "" == recipeType ? "emtpy":recipeType );
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_save(): Recipe name (" + emptyName + ") and/or type (" + emptyType + ") not given for " + ukl1Name + ". Cannot save recipe.", "1");
  }// else recipe name or type empty.
  
  // Done.
  return;
}// _fwUkl1_save()

/*!
 * Loads the data from the given UKL1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  recipeName Name of the recipe. The structure for this is described in fwUkl1_save().
 * \param  recipeType Recipe type that is to be saved.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  ddcData Each element of this array contains an array of char that make up the words that are load from the
 *           recipe. Returned by reference.
 * \param  domain This is required if the recipe to be loaded has been loaded into the recipe cache by the FSMConfDB. In this case
 *           the name of the recipe will have been mangled. The recipe name given to this function should be the unmangled
 *           name and it will be converted as appropriate. This parameter can also be an empty string if no name mangling is required.
 *           Note in this case there can be no / present in the recipe name, only to delimit the run mode from the command (RUN_MODE/Command).
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_load(const string &ukl1Name, const string &recipeName, const string &recipeType, const dyn_string &registerNames, dyn_dyn_char &ddcData, string domain, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  if ( "" == ukl1Name ) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_load(): No UKL1 name give to retrieve the recipe from, no data will be retrieved.", "2");
    return;
  }// if(""==ukl1Name)
  
  // Check that we are given a recipeType and a recipeName.
  if ( ("" != recipeType) && ("" != recipeName) ) {
    // When we get the recipe data it will give us all registers in the recipe and also data.
    // This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    // Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( -1 != fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData, domain) ) {
      // Note that the length of the registerNames array is calculated for each loop.
      // This is important if we remove elements from the array, the total length does need to be updated.
      for (int element = 1; element <= dynlen(registerNames); ++element) {
        // This takes the specific register name and returns the index of its occurance in the full list.
        // Relies on the fact that the list is unique.
        // It is this index that we can use to over write the appropriate data element.
        int index = dynContains(tmpRegisterNames, registerNames[element]);
        // Now we know where the appropriate recipe data is copy it into the ddcData array.
        // It is possible that this data does not occur in this recipe type, hence we must protect against this.
        if ( 0 != index )
          ddcData[element] = tmpDdcData[index];
        // If the element doesn't exist then pad the ddcData array with an empty array.
        // This allows the parse data function to pick up on it and put a null value
        // into the into the return data.
        else
          ddcData[element] = makeDynChar();
      }// for element
    }// if(-1!=fwhw_getRecipe())
    else {
      // It is most likely that the recipe just didn't exist.
      // Populate the ddcData with nothing such that it can return nothing, but it will contain the appropriate number of elements of nothing.
      for (int element = 1; element <= dynlen(registerNames); ++element)
        ddcData[element] = makeDynChar();
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_load(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ". Recipe may not exist.", "1");
    }// else(-1!=fwHw_getRecipe())
  }// if recipe name and type not empty.
  else {
    // Create more useful error message.
    emptyName = ( "" == recipeName ? "empty":recipeName );
    emptyType = ( "" == recipeType ? "emtpy":recipeType );
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_load(): Recipe name (" + emptyName + ") and/or type (" + emptyType + ") not given for " + ukl1Name + ". Cannot load recipe.", "1");
  }// else recipe name or type empty.
    
  // Done.
  return;
}// _fwUkl1_load()
  
/*!
 * Reads from the registers on the L1 board to ensure that the data just written has been written successfully.
 *
 * \param  registerNames Names of the registers that were written to.
 * \param  writtenData The data that was written to the registers that the read back values should be compared against.
 * \param  masks Any masks that should be applied to the read data, such that only specific bits in a register are compared
 *           against the write bits.
 * \param  readData The data that was read back from the registers to compare against the written data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This function is designed to check the data written and read using the fwCcpc_writeRead function.
 */
void _fwUkl1_verifyWrite(dyn_string registerNames, dyn_dyn_char writtenData, dyn_dyn_char masks, dyn_dyn_char readData, dyn_string& exceptionInfo) {
  // Number of elements of data to loop through.
  const unsigned numRegs = dynlen(registerNames);
  // Check to see if any masks are supplied and if so flag them for use.
  bool applyMasks = ( numRegs==dynlen(masks) ? TRUE:FALSE );
  // Check the data.
  for (unsigned element = 1; element <= numRegs; ++element) {
    // Number of bytes written.
    const int dataLen  = dynlen(writtenData[element]);
    // Number of bytes read back.
    const int checkLen = dynlen(readData[element]);
    // Check they are the same size.
    if (checkLen != dataLen) {
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_verifyWrite(): " + registerNames[element] + " written and read data lengths differ when verifying write, " + dataLen + " and " + checkLen + " bytes big respectively. Read data 0x" + fwCcpc_convertByteToHex(readData[element]) + ", and written data 0x" + fwCcpc_convertByteToHex(writtenData[element]) + " (premask).", "2");
    }// if(checkLen!=data)
    else {
      // Only check the data if they are the same size.
      // Use to note if an element in the read array differ.
      bool diff = FALSE;
      // This will provide us with our loop index. The two arrays have been found to be the same size, so just set to either.
      for (unsigned index = 1; index <= dataLen; ++index) {
        // Apply the masks that are given to the read data, if masks are supplied.
        if (applyMasks)
          readData[element][index] &= masks[element][index];
        // Now check the data against the read values.
        if ( readData[element][index] != writtenData[element][index] ) {
          diff = TRUE; 
          // We can't break here as otherwise the readData will not be fully masked and that value included in the error will make no sense.
          // break;
        }
      }// for index
      // Will only raise the exception displaying the whole value and not for each differing element.
      if (diff)
        fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_verifyWrite(): " + registerNames[element] + " read data, " + fwCcpc_convertByteToHex(readData[element]) + ", and written data, " + fwCcpc_convertByteToHex(writtenData[element]) + ", differ.", "2");
    }// else(checkLen!=dataLen)
  }// for element

  // Done.
  return;
}// _fwUkl1_verifyWrite()

// @}


// ==========================
//   DATA PARSING FUNCTIONS
// ==========================

/** @defgroup SectionInterface DataParsingFunctions
 *   These functions are all provided to assist with converting the data from that read from
 *   the hardware to that display to the User or vice versa.
 */
// @{

/*!
 * Takes a dyn_char and converts it to a hex string assuming the input data is little endian.
 *
 * \param  dcData A dyn_char containing the data read back from the hardware register.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return string Returns the hex string.
 *
 * The fwCcpc_convertByteToHex function assumes big endian data.
 */
string fwUkl1_convertByteToHexLittleEndian(const dyn_char &dcData, dyn_string &exceptionInfo) {
  // We must reverse the byte ordering of the array.
  // For this we will need to know its size.
  const int numBytes = dynlen(dcData);
  // This stores the reversed bytes.
  dyn_char swappedDcData;
  if (0 < numBytes) {
    // Have a none zero number of bytes.
    // Loop through them swapping the first element for last, 2nd for 2nd to last etc..
    for (int byte=1; byte<=numBytes; ++byte)
      swappedDcData[numBytes-(byte-1)] = dcData[byte];
  }// if(0<numBytes)
  else if (0 == numBytes)
    return "";
  else {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToHexLittleEndian(): PVSS error, failed to determine size of byte array, could not convert to hex string.", "1");
    return "Fail";
  }// else failed
  
  // Done.
  return fwCcpc_convertByteToHex(swappedDcData);
}// fwUkl1_convertByteToHexLittleEndian()

/*!
 * Takes a string and ensures that it contains enough characters for the number of bytes specified
 * by numBytes. Adds `0' to the beginnig if it is too short and removes characters from the beginning
 * if it is too long. Must be in hex.
 *
 * \param  sData The string that contains the data to be checked. Returned by reference.
 * \param  numBytes The number of bytes that the returned character string should return.
 * \return void
 *
 * This is for use with strings that represent a hex number that should be a set number of bits long.
 * It assumes that the lowest order bit is the last (reading left to right) character and hence its behaviour is
 * to move the last character always towards the final character in the padded string (the 2nd, 4th, 6th etc character).
 * This is achieved by padding or removing at the beginning of the string.
 *
 * \b Example:
 * \code
 *   (...)
 *   // Declare some data e.g. User input from a panel.
 *   const string hexData1 = "1234";
 *   // Wish to use the fwCcpc library to write this to a 32-bit registers hence is needs to be zero padded
 *   // to be 4-bytes big (8-characters).
 *   fwUkl1_padString(hexData1, 4);
 *   DebugTN(hexData1); // "00001234"
 *
 *   // Declare some more data e.g. User input from a panel.
 *   const string hexData2 = "1234";
 *   // Wish to use the fwCcpc library to write this to an 8-bit registers hence it need to be 1-byte (2-characters) big.
 *   fwUkl1_padString(hexData2, 1);
 *   // Here the data is misformed and it is assumed that the least significant bit (last character) in the given number
 *   // should remain the least significant bit.
 *   DebugTN(hexData2); // "34"
 *   (...)
 * \endcode
 */
void fwUkl1_padString(string& sData, unsigned numBytes) {
  // First determine the length.
  const int datalen = strlen(sData);
  // Create a temporary string to store the original string.
  const string tmpOrig = sData;
  // Determine the number of characters that must be in the final string,
  // each character represents half a byte or nibble.
  const int numChar = numBytes*2;
  // Perform the checking and padding/trimming.
  if ( numChar > datalen ) {
    // The data string is too short so increase its length by padding at the beginning, highest order bytes.
    // A temporary string to hold the amount of padding.
    string tmpPad;
    for (int character = 1; character <= (numChar-datalen); ++character) {
      tmpPad += "0";
    }
    // Now insert the padding into the data string.
    sData = tmpPad + tmpOrig;
  }// if(numChar>datalen)
  // Remove the initial characters of the string until it reaches the appropriate length, highest order bytes.
  else if ( numChar < datalen )
    sData = substr(sData, (datalen-numChar), datalen);
  // No else do nothing, it is the right length.

  // Return the resulting string.
  return;
}// fwUkl1_padString()

/*!
 * This takes a string and will bit shift it left by the requested number of bits. It will pad the end of the string with
 * with zero bits.
 *
 * \param  theString The string to be left shifted. Returned by reference.
 * \param  shift The number of bits to shift the string by.
 * \return void.
 *
 * The primary assumption of this algorithm is that the given string is a hex representation of a number. Thus each character
 * is 4 bits when considered as an int. As in PVSS each character is actually 8 bits the shift it doesn't represent a direct
 * bit shift of the binary representation.
 *
 * Padding will be added at the end of the string, which is considered the beginning of the number, and extra characters will
 * be added at the beginning of the string, which is considered the end of the number, if the bits overflow the current length
 * of the string. No bits are lost.
 */
void fwUkl1_leftShiftString(string &theString, int shift) {
  // The alogrithm works by considering first how many whole characters to shift (1 character = 4 bits there)
  // and then how many bits to shift within each character.
  const int charToShift       = shift / 4;
  const int bitsInCharToShift = shift % 4;
  // Current number of characters in string.
  const int startLenString = strlen(theString);
  // We will break each character into a char shift each one individually.
  // Each element has enough bits to prevent overflow as we only need to shift by less bits than a character.
  // We need to keep the previous result from the loop to put the over flow in it.
  char tmpCPrevious;
  char tmpCCurrent;
  string tmpString;
  // First we will shift by the appropriate number of bits within each character.
  // We are starting at the end of the number.
  for (int character = 1; character <= startLenString; ++character) {
    // First store the character as a byte, which is has plenty of room for overflow.
    tmpCCurrent = fwCcpc_convertHexToByte(theString[character-1])[1];
    // Now shift it.
    tmpCCurrent <<= bitsInCharToShift;
    if ( 1 != character ) {
      // Now we OR the last 4 bits of the current element with the first 4 bits of the previous element.
      // This puts any overflow into the next character.
      tmpCPrevious |= tmpCCurrent >> 4;
      // Now erase the last 4 bit of the char as we are only interested in 4 bit chars.
      tmpCCurrent   &= 0x0f;
      
      // Now build the string back up from the previous value.
      // Generally we only want the last character from the returned string as this is the 4 bytes we have OR'd to
      // make the bit shift value. The first character is just padding as we give the conversion function 8 bits.
      if (2 != character)
        tmpString += fwCcpc_convertByteToHex(makeDynChar(tmpCPrevious))[1];
      // In the event we are dealing with the first created character we want all 8bits as the first character in the
      // returned string contains the overflow for the shifted string.
      else
        tmpString += fwCcpc_convertByteToHex(makeDynChar(tmpCPrevious));
      // If it is the last element then we want to save it as well.
    }// if(1!=character)
    // Save the current char for the next loop.
    tmpCPrevious = tmpCCurrent;
  }// for character
  // Need to put the last character in the temp string too. It will have not be done in the loop.
  // Always only interested in the last character.
  tmpString += fwCcpc_convertByteToHex(makeDynChar(tmpCPrevious))[1];
  
  // Overwrite theString with that we just created.
  theString = tmpString;
  // Add to the end of the string the number of whole characters we need to shift.
  int addedChar = 0;
  while (addedChar++ < charToShift)
    theString += "0";
  
  // Done.
  return;
}// fwUkl1_leftShiftString()
  
// ===============
//    INTERNAL
// ===============

/*!
 * Converts a dyn_char to a hex string. Adds the functionality to remove only a limit range from the input dyn_char.
 *
 * \param  dcData Data read from the register as a byte array.
 * \param  startBit Number of the first bit to be read from the register, where bit 0 is the least significant bit.
 * \param  length Number of bits to be read, set to 1 if only startBit is to be read. If -1 all the data is returned.
 * \param  fpga Boolean value to indicate whether we are parsing FPGA registers or not. This must be set correctly otherwise
 *           the data will not be properly interpreted.
 * \return string Hex representation of the requested bits. Will contain no leading 0s.
 *
 * The length and startBit will be cohereced to fit into the size of the dyn_char is necessary. startBit will be made less
 * than length or length greater than startBit as appropriate.
 * The reason for the FPGA flag is that the UKL1 FPGA registers are read as if they were 32-bit registers and not the 16-bit registers
 * that they are physically (this is because sequential addresses are 32-bits apart and not 16!). This function will remove the extra two
 * junk words that are read and parse only the remaining data in this case. Do not remove these extra words manually before hand.
 */
string _fwUkl1_convertByteToHex(dyn_char dcData, unsigned startBit, unsigned length, bool fpga) {
  // Check there is actually some data.
  if (0 >= dynlen(dcData))
    return "";

  // If we are dealing with an FPGA the we need to convert from 16- to 32-bit words.
  if (fpga)
    _fwUkl1_convert32To16BitWords(dcData);

  // In order to be able to span byte boundaries we need to know where we are going to start and finish in terms
  // of the appropriate byte. These constants will ease the navigation of the bit-field.
  // Determine how many bytes we have been given.
  const int numBytes = dynlen(dcData);
  // Determine how many bits we have, there are eight bits in each element as each element contains a char.
  const unsigned numBits = 8 * numBytes;
  // Perform a sanity check on the starting bit.
  if ((numBits-(unsigned)1) < startBit)
    startBit = numBits-1;
  
  // The last bit is the number of bits we wish to remove offset by the starting bit number.
  // We subtract 1 as if we just remove 1 bit then the start and end bit are the same.
  unsigned endBit = startBit + length - 1;
  // Can't end at a point passed the number of bits that we have.
  // The last bit is 1 less than the number of bits in the register as we start counting at 0.
  // numBits is starts at 1, whereas endBit starts at 0.
  if ( (numBits-(unsigned)1) < endBit )
    endBit = numBits -1;
  
  // The first byte that we need to read from. Remembering that we are using a big endian numbering scheme.
  const unsigned startByte = numBytes - (unsigned)(startBit/8);
  // This is the last byte that we need to read from.
  const unsigned endByte = numBytes - (unsigned)endBit/8;
  // This is the region to remove the data from.
  // Marks the bit in the byte to start the selection from.
  // It is initially offset into the first byte to read from.
  // After the first byte it is always the beginning of a byte.
  unsigned startBitInByte = (unsigned)startBit%8;
  // This is the last bit in the byte that we need to read from.
  // Either it is either the last bit in the byte until we are in the
  // last byte to be read from.
  unsigned endBitInByte = 7;
  
  // The selected elements from the given dyn_char are stored here before being converted to hex.
  dyn_char selectedData;
  // This the number of bytes we need to select. Make sure we round up when converting to an int.
  const int numSelectedBytes = ceil(length/8.0);
  // When we add the selected data we need to know where in the selectedData dyn_char to write it.
  // This keeps track of the byte bearing in mind we are moving from end to start of the arrays.
  unsigned currentByte = numSelectedBytes;

  // Now we need to loop from the startByte to the endByte, decrement as we are big endian.
  for (unsigned byte=startByte; byte>=endByte; --byte) {
    // This mask is used to ensure that we ignore any bits at the end of the byte we are selecting.
    // Typically we will just run to the end of the word.
    char endMask = 0xff;
    // If we are in the last byte then we won't necessarily read all the data from it.
    if (startByte == endByte) {
      endBitInByte = (unsigned)endBit%8;
      endMask = ~(0xff << (endBitInByte+1));
    }// if(startByte==endByte)
    
    // Get the byte we are intersted in and mask off the upper bits we don't want.
    char selectedByte = dcData[byte] & endMask;
    // Now align the start bit to the beginning.
    selectedByte >>= startBitInByte;
    // By this stage the bits that we are not interested must be 0 and we can have at most 1 byte.
    // Add to the selected data.
    selectedData[currentByte] = selectedByte;
    // Move onto the next byte.
    --currentByte;
    // After the first run through the loop the startBitInByte to be read from is always the beginning.
    if (startByte == byte)
      startBitInByte = 0;
  }// for byte

  // Remove any leading zeros.
  string rtrn = strltrim(fwCcpc_convertByteToHex(selectedData), "0");
  // If it is zero then it will now be empty.
  if ("" == rtrn)
    rtrn = "0";
  // Done.
  return rtrn;
}// _fwUkl1_convertByteToHex()

/*!
 * Converts the single abstract register and input data to multiple hardware channels
 * with the same data to be written.
 *
 * \param  regName Name of the abstract register to write the data to. This should be an
 *           'all' register, which represents multiple UKL1 channel abstract registers.
 * \param  sData The data to be written to all the channels.
 * \param  outData This is a mapping that links the hardware registers to the dyn_char data
 *           and masks that are to be written for the channels. Any existing data for a
 *           register bit to be written will be over written and any other registers
 *           will be untouched.
 * \param  exceptionInfo Error information. No change in size if there are no errors.
 */
void _fwUkl1_convertAbsRegChannelAllToIndividualRegs(const string &regName, const string &sData, mapping &outData, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // For the abstract register we are given we must convert it to the abstract register for each channel on the
  // UKL1 board and set the data to the same thing.
  mapping inData;
  // We will over write this string each time with the abstract register to be put in the map.
  string absRegName;
  for (int channel=0; channel<FWUKL1_BOARD_CHANNELS; ++channel) {
    // Replace the All in the abstract register with each of the channel numbers
    // we want to write to. This will give us the list of abstract registers to write to.
    absRegName = regName;
    strreplace(absRegName, "All", channel);
    inData[absRegName] = makeDynString(sData);
  }// for channel
  // Now we have constructed a new input data map we should parse it.
  _fwUkl1_parseRegistersForSet(inData, outData, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_convertAbsRegChannelAllToIndividualRegs(): Failed to parse some/all of the hardware data for the abstract register " + regName + ". Not all the data will be written to the hardware.", "2", exInfoSize);
}// _fwUkl1_convertAbsRegChannelAllToIndividualRegs()

/*!
 * Converts the data from multiple hardware registers into the output for a single abstract register.
 *
 * \param  regName The abstract register name that the data is being converted for. This should be an
 *           'all' register, which represents multiple UKL1 channel abstract registers.
 * \param  inData This is the raw data that has been read from the hardware or out of the database. The map
 *           should contain as a key the hardware register and the data is a dyn_char which is the raw
 *           data from the hardware/DB.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_string Returns a text description and colour to associate with the description. Will be Fail and FwStateAttention1
 *           in the event of an error.
 */
dyn_string _fwUkl1_convertByteToAbsRegChannelAll(const string &regName, const mapping &inData, bool forStatus, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // This will hold the output data.
  dyn_string dsData = makeDynString("Fail", "FwStateAttention1");
  // We will do this by first converting the given abstract register into a list of abstract registers that
  // we want to get the data from.
  string absRegName;
  mapping outData;
  for (int channel=0; channel<FWUKL1_BOARD_CHANNELS; ++channel) {
    // Replace the All in the abstract register with each of the channel numbers
    // we want to write to. This will give us the list of abstract registers to write to.
    absRegName = regName;
    strreplace(absRegName, "All", channel);
    outData[absRegName] = makeDynString();
  }// for channel
  // Our inData is the same as that given to the function and we have created ourselves a new output data map.
  _fwUkl1_parseRegistersForGet(inData, outData, forStatus, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_convertByteToAbsRegChannelAll(): Failed to parse some/all of the hardware data for the abstract register " + regName + ". Not all the data will be OR'd for this register.", "2", exInfoSize);
  
  // Now we have all our data in its user friendly format we should OR it together.
  // We need to loop over each entry and compare it to the first.
  dyn_string firstValue;
  dyn_string currentValue;
  // We create a list of inconsistent channels for the error message.
  string inconsistent = "The abstract register(s) -- ";
  // which we will only create if there is an error.
  bool error = FALSE;
  const dyn_string keys = mappingKeys(outData);
  const int numKeys = dynlen(keys);
  // Loop through all the entries and mark it as fail if they are not consistent.
  for (int i=1; i<=numKeys; ++i) {
    // For the first time through the loop we just update the current value
    // and move on. The previous value will be initialised from this on the
    // next loop.
    if (1 == i) {
      firstValue = outData[keys[i]];
      // Current and previous values must be the same in this case.
      continue;
    }// if(1==i)
    // Get the current entry
    currentValue = outData[keys[i]];
    // and compare it to the first.
    if (firstValue != currentValue) {
      error = TRUE;
      inconsistent += keys[i] + " (" + currentValue + "), ";
    }//if(previousValue!=currentValue)
  }// for i

  if (error) {
    // End the string.
    // Trim the excess ', ' off the end of the string
    inconsistent = strrtrim(inconsistent, ", ");
    // and now add the end of our message.
    inconsistent += " are inconsistent with " + keys[1] + "(" + firstValue + ").";
    DebugTN(inconsistent);
    // Change the return value and colour appropriately.
    dsData[FWUKL1_MAP_REG_DATA] = "Inconsistent";
    dsData[FWUKL1_MAP_XTR_DATA] = "FwStateAttention2";
  }//if(error)
  // If there was no error then just take the last output. Everything is the same as it.
  else
    dsData = currentValue;
  
  // Done.
  return dsData;
}// _fwUkl1_convertByteToAbsRegChannelAll()

/*!
 * Takes an optionally colon delimited hex MAC address, as a string, and returns a dyn_char containing
 * the MAC address in a hardware/DB compatible form.
 *
 * \param  macAddress The string that contains the MAC address to be parsed. Should be of the form XX:XX:XX:XX:XX:XX
 *           or XX:XX:XX:XX. If the final two bytes are omitted the returned dyn_char is formatted for the source MAC
 *           address, otherwise the dyn_char is formatted for the destination MAC address.
 * \param  exceptionInfo Error information. No change in size if there are no errors.
 * \return dyn_char The MAC address converted to a dyn_char suitable for writing to either the destination
 *           or source address register, which is determined by the input data format. An empty array if the
 *           data cannot be parsed.
 *
 * Expects the MAC address to be of the form 00:0e:0c:b0:2d:19 or 000e0cb02d19 i.e. 12 characters long.
 */
dyn_char _fwUkl1_convertMacToByte(string macAddress, dyn_string& exceptionInfo) {
  // Check that it contains the appropriate number of characters, either with or without colons.
  if ( (12 != strlen(macAddress)) && (17 != strlen(macAddress)) &&  // Full MAC address lengths w/out colons.
       ( 8 != strlen(macAddress)) && (11 != strlen(macAddress)) ) { // Missing remaining byte MAC address lengths w/out colons.
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): MAC address is " + strlen(macAddress) + " characters long. Should be 8, 11, 12 or 17 characters.", "1");
    return makeDynChar();
  }// MAC address suitable length.

  // Will hold the MAC address once it has been parsed.
  string parsedMacAddress;
  // Try to split the string using a colon as a delimiter.
  dyn_string tmpMac = strsplit(macAddress, ":");
  // Check it has found the appropriate number of parts.
  const int lenMac = dynlen(tmpMac);
  if ( 1 == lenMac ) {
    // String contained no colons, decide whether to add the final zero data.
    parsedMacAddress = macAddress;
    if ( 8 == strlen(macAddress) )
      parsedMacAddress = parsedMacAddress + "0000";
  }// if no colons
  else if ( 4 == lenMac ) {
    // Contains 4 parts, put them all into one string and then pad the remaining four characters with 0s.
    int status = sprintf(parsedMacAddress, "%02s%02s%02s%02s0000", tmpMac[1], tmpMac[2], tmpMac[3], tmpMac[4]);
    // Check for errors.
    if ( -1 == status ) {
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): Failed to reconstruct MAC address as hex string.", "1");
      parsedMacAddress = "Fail";
    }
  }// if source address with colons
  else if ( 6 == lenMac ) {
    // Contains 6 parts, put them all into one string.
    int status = sprintf(parsedMacAddress, "%02s%02s%02s%02s%02s%02s", tmpMac[1], tmpMac[2], tmpMac[3], tmpMac[4], tmpMac[5], tmpMac[6]);
    // Check for errors.
    if ( -1 == status ) {
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): Failed to reconstruct MAC address as hex string.", "1");
      parsedMacAddress = "Fail";
    }
  }// if destination address with colons
  else {
    // Not properly formed.
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): Found only " + lenMac + " colon delimited sections, should contain 1, 4 or 6 for a valid MAC address.", "1");
    parsedMacAddress = "Fail";
  }// else not properly formed

  // This is what we will return.
  dyn_char regData = makeDynChar();
  if ( "Fail" != parsedMacAddress ) {
    // Now we need to do two things:
    // 1) Word swap the data, this is because the fwCcpc library when converting from hex to byte will parse
    //     in increasing index through the string. This cause the last word to be written in the first elements.
    // 2) Pad the data. This is because the UKL1 registers are 16-bit registers on a 32-bit address space.
    //     The hardware write understands how to write the MAC address over consequtive registers, but
    //     the data must be in 32-bit words.
    parsedMacAddress =   "0000" + substr(parsedMacAddress, 8, 4)
                       + "0000" + substr(parsedMacAddress, 4, 4)
                       + "0000" + substr(parsedMacAddress, 0, 4);
    // Now convert it from a string to a dyn_char.
    regData = fwCcpc_convertHexToByte(parsedMacAddress);
  }// if("Fail"!=parsedMacAddress)
  // Done.
  return regData;
}// _fwUkl1_convertMacToByte()

/*!
 * Takes an array of bytes and returns them as a colon delimited MAC address.
 *
 * \param  macAddress The array containing the bytes to be parsed.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_string Contains the MAC address in element FWUKL1_MAP_REG_DATA and a colour to indicate status in FWUKL1_MAP_XTR_DATA.
 *           Returns `Fail' and FwStateAttention1 if the input data is missformed.
 *
 * This code requires that the extra bytes generated from a fwCcpc_read() are first removed, via _fwUkl1_convert32To16BitWords().
 */
dyn_string _fwUkl1_convertByteToMac(dyn_char macAddress, dyn_string& exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // First convert it from the 32bit to 16bit words. We don't care about the loss of mask
  // element from the confDB as we ignore them anyway.
  // Do this before we calculate the size of the MAC address. It also makes the later indexing more natural.
  _fwUkl1_convert32To16BitWords(macAddress);
  
  // If the masks are in the data then they will be at the front i.e. lowest element numbers.
  // Determine the number of elements in the array and count backwards from there.
  // This will allow absolute indexing whether the masks are present or not.
  const int size = dynlen(macAddress);
  // Check that it contains the appropriate number of elements.
  // Technically if it came from the configuration database it could contain the write mask tagged on the end.
  // This doesn't cause a problem as we use absolute indexes later, just require there to be 6 elements minimum (and the MAC to be in those).
  if ( 6 > size ) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToMac(): Input array contains " + size + " elements. Should be 6 or 12 elements big.", "1");
    return makeDynString("Fail", "FwStateAttention1");
  }// if(6>size)

  // Now we need to put the data into the appropriate format.
  string output;
  // This should be in a big endian order once it reaches here.
  output =  fwCcpc_convertByteToHex(macAddress[size - 5]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 4]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 3]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 2]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 1]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size]);

  // Done.
  return makeDynString(output, "FwBackgroundInputText");
}// _fwUkl1_convertByteToMac()

/*!
 * Takes a standard formed IP(4) address, as a string, and convert it to a continuous hex
 * string. Also performs syntacs checking on the given string.
 *
 * \param  ipAddress The string that contains the IP address to be parsed. Should be of the form:
 *           XXX.XXX.XXX.XXX, where X represents a decimal number, e.g. 192.168.192.192. It is possible
 *           to miss the final byte i.e. 192.168.192, this is required for the source IP address.
 *           In this case where the final byte is missing it will be replaced with zeros in the resulting hex string.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_char Contains the data in a form that can be written directly to the hardware or to the confDB.
 *           Empty if there is an error.
 */
dyn_char _fwUkl1_convertIpToByte(string ipAddress, dyn_string& exceptionInfo) {
  // Check that this does not contain more than the maximum number of bits that can be present in an IP(4) address.
  // Remember that an IP address length is very variable 1.1.1.1 compared to 255.255.255.255
  if ( 16 <= strlen(ipAddress) ) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertIpToHex(): IP address is " + strlen(ipAddress) + " characters long. Can be a maximum of 15.", "1");
    return "Fail";
  }

  // Holds the resulting IP address in hex.
  string parsedIpAddress;
  // Try to split the string using a . as a delimiter.
  dyn_string tmpIp = strsplit(ipAddress, ".");
  // Check it has found the appropriate number of parts.
  const int lenIp = dynlen(tmpIp);
  if ( 4 == lenIp ) {
    // Contains 4 parts, put them all into one string.
    for (unsigned element = 1; element <= dynlen(tmpIp); ++element) {
      // Take a section of the IP and convert it to a hex string.
      string tmpSection = fwCcpc_convertDecToHex(tmpIp[element]);
      // If it does not contain 2 characters pad the string.
      if ( 2 != strlen(tmpSection) )
	tmpSection = "0" + tmpSection;
      parsedIpAddress += tmpSection;
    }// for element
  }// if(4==lenIp)
  else if ( 3 == lenIp ) {
    // Contains 3 parts, put them all into one string and then pad the last two characters to make up the byte.
    for (unsigned element = 1; element <= dynlen(tmpIp); ++element) {
      // Take a section of the IP and convert it to a hex string.
      string tmpSection = fwCcpc_convertDecToHex(tmpIp[element]);
      // If it does not contain 2 characters pad the string.
      if ( 2 != strlen(tmpSection) )
	tmpSection = "0" + tmpSection;
      parsedIpAddress += tmpSection;
    }// for element
    // Add zeros for the remaining two characters.
    parsedIpAddress += "00";
  }// elseif(3==lenIp)
  else {
    // Not properly formed.
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertIpToHex(): Found " + lenIp + " full stop delimited sections, require 3 or 4.", "1");
    parsedIpAddress = "Fail";
  }// else

  // This is what we will return.
  dyn_char regData = makeDynChar();
  if ( "Fail" != parsedIpAddress ) {
    // Now we need to do two things:
    // 1) Word swap the data, this is because the fwCcpc library when converting from hex to byte will parse
    //     in increasing index through the string. This cause the last word to be written in the first elements.
    // 2) Pad the data. This is because the UKL1 registers are 16-bit registers on a 32-bit address space.
    //     The hardware write understands how to write the IP address over consequtive registers, but
    //     the data must be in 32-bit words.
    parsedIpAddress =   "0000" + substr(parsedIpAddress, 4, 4)
                      + "0000" + substr(parsedIpAddress, 0, 4);
    // Now convert it from a string to a dyn_char.
    regData = fwCcpc_convertHexToByte(parsedIpAddress);
  }// if("Fail"!=parsedMacAddress)
  
  // Done.
  return regData;
}// _fwUkl1_convertIpToByte()

/*!
 * Takes an array of bytes, and returns them as a full stop delimited IP address.
 *
 * \param  ipAddress The array containing the bytes to be parsed.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_string Contains the IP in element FWUKL1_MAP_REG_DATA and a colour to indicate status in FWUKL1_MAP_XTR_DATA.
 *           Returns `Fail' and FwStateAttention1 if the input data is missformed.
 *
 * This code requires that the extra bytes generated from a fwCcpc_read() are first removed, via _fwUkl1_convert32To16BitWords().
 *
 */
dyn_string _fwUkl1_convertByteToIp(dyn_char ipAddress, dyn_string& exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // First convert it from the 32bit to 16bit words. We don't care about the loss of mask
  // element from the confDB as we ignore them anyway.
  // Do this before we calculate the size of the MAC address. It also makes the later indexing more natural.
  _fwUkl1_convert32To16BitWords(ipAddress);
  
  // If the masks are in the data then they will be at the front i.e. lowest element numbers.
  // Determine the number of elements in the array and count backwards from there.
  // This will allow absolute indexing whether the masks are present or not.
  const int size = dynlen(ipAddress);
  // Check that it contains the appropriate number of elements.
  // Technically if it came from the configuration database it could contain the write mask tagged on the end.
  // This doesn't cause a problem as we use absolute indexes later, just require there to be 4 elements minimum (and the IP to be in those).
  if ( 4 > size ) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToIp(): Input array contains " + size + " elements. Should be 4 elements big.", "1");
    return makeDynString("Fail", "FwStateAttention1");
  }// if(4>size)

  // Now we need to put the data into the appropriate format.
  string output;
  output += fwCcpc_convertByteToDec(ipAddress[size - 3]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size - 2]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size - 1]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size]);

  // Done.
  return makeDynString(output, "FwBackgroundInputText");
}// _fwUkl1_convertByteToIp()

/*!
 * This takes a description of the value to be set to the hardware and 
 * This takes the data, removes the appropriate bit using _fwUkl1_convertHwRegtoAbsRegHex() and then converts that bit to the
 * appropriate text description of the register. This is then returned.
 *
 * \param  regName The abstract register name that the data is being read for.
 * \param  hwRegName Name of the hardware register that is to be written to.
 * \param  description The description of the setting that is to be added to the dcData.
 * \param  hexMask This is the hex string that is the mask for the given data. It is applied to the data before it is added to the dcData
 *           to ensure it does not go out of bounds. The mask is also added to the given mask such that it can be written to the hardware.
 * \param  dcData The dyn_char that the new value is to be written into. Returned by reference.
 * \param  dcMask The dyn_char that is the mask for the given data. Returned by reference.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The dyn_char's for the data and mask can already contain data for that hardware register and only the bit relevant to the given
 * data and mask will be updated. These dyn_char's can also be empty and the function will create them such that they are the correct size.
 *
 * No changes are made to the data or mask unless the parsing is successful.
 */
void _fwUkl1_convertAbsRegDescritionToByte(const string &regName, const string &hwRegName, const string &description, string hexMask, dyn_char &dcData, dyn_char &dcMask, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);

  // First the description needs to be converted a hex value.
  string hexData;
  // Get the arugments for the function for this register.
  dyn_dyn_string listOfArgs = fwUkl1_getExtraRegisterInformation(regName, FWUKL1_CONFIG_WRITE_XTR_INFO, exceptionInfo);
  // Must have at least three list of arugments.
  if ( 2 < dynlen(listOfArgs) ) {
    // Need two arguments.
    const int compOneEle = 2;
    const int compTwoEle = 3;
    if ( 1 < dynlen(listOfArgs[compOneEle]) ) {
      // For the moment we will only check the first case and then if necessary decode the rest of the arugments and check the second.
      string comparison  = listOfArgs[compOneEle][1];
      string value       = listOfArgs[compOneEle][2];
      if ( description == comparison )
        hexData = value;
      else {
        if ( 1 < dynlen(listOfArgs[compTwoEle]) ) {
          comparison  = listOfArgs[compTwoEle][1];
          value       = listOfArgs[compTwoEle][2];
          if ( description == comparison )
            hexData = value;
          else
            exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegDescritionToByte(): Failed to convert the setting " + description + " for the register " + regName + " as the description was not recognised.", "2");
        }// if not enough arguments.
        else
          exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegDescritionToByte(): On second comparison to convert the description (" + description + ") to a hex value for the register "+ regName + ". Require 2, given " + dynlen(listOfArgs[compTwoEle]) + ".", "2");
      }// else(description==comparison))
    }// if(1<dynlen(listOfArgs[compOneEle]))
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegDescritionToByte(): On first comparison to convert the description (" + description  + ") to a hex value for the register "+ regName + ". Require 2, given " + dynlen(listOfArgs[compOneEle]) + ".", "2");
  }// if(2<dynlen(listOfArgs))
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegDescriptionToByte(): When trying to parse the description (" + description + ") did not retrieve enough arguments from the list to parse the data. Required 3, given " + dynlen(listOfArgs) + ".", "2");
  
  // Now we should have some hex data lets convert it to a dyn_char.
  _fwUkl1_convertAbsRegHexToByte(regName, hwRegName, hexData, hexMask, dcData, dcMask, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegDescriptionToByte(): Failed to convert the hex data (" + hexData + ") for the register " + regName + " to a dyn_char to write to the hardware.", "2", exInfoSize);
  // Done.
  return;
}// _fwUkl1_convertAbsRegDescritionToByte()

/*!
 * This takes the data, removes the appropriate bit using _fwUkl1_convertHwRegtoAbsRegHex() and then converts that bit to the
 * appropriate text description of the register. This is then returned.
 *
 * \param  regName The abstract register name that the data is being converted for.
 * \param  dcData A dyn_char containing the data read back from the hardware register.
 * \param  forStatus If the description to be retrieved for the status display or configuration display.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_string Returns a text description and colour to associate with the description. Will be Fail and FwStateAttention1
 *           in the event of an error.
 */
dyn_string _fwUkl1_convertByteToAbsRegDescrition(const string &regName, const dyn_char &dcData, bool forStatus, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // This holds the output. First get the hex value, don't pad it. It confuses the comparison.
  dyn_string dsData = _fwUkl1_convertByteToAbsRegHex(regName, dcData, FALSE, forStatus, exceptionInfo);
  if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegDescrition(): Failed to retrieve the hex value from the hardware. Cannot convert it to a description.", "2", exInfoSize)) {
    // Get the arugments for the function for this reigster.
    // Need to decide whether it is for the status or not.
    string dpe = FWUKL1_CONFIG_READ_XTR_INFO;
    if (forStatus)
      dpe = FWUKL1_STATUS_READ_XTR_INFO;
    dyn_dyn_string listOfArgs = fwUkl1_getExtraRegisterInformation(regName, dpe, exceptionInfo);
    // Must have at least three list of arugments.
    if (2 < dynlen(listOfArgs)) {
      // Need two arguments.
      const int compOneEle = 2;
      const int compTwoEle = 3;
      if (1 < dynlen(listOfArgs[compOneEle])) {
        // For the moment we will only check the first case and then if necessary decode the rest of the arugments and check the second.
        string comparison  = listOfArgs[compOneEle][1];
        string description = listOfArgs[compOneEle][2];
        if ( dsData[FWUKL1_MAP_REG_DATA] == comparison )
          dsData[FWUKL1_MAP_REG_DATA] = description;
        else {
          if (1 < dynlen(listOfArgs[compTwoEle])) {
            comparison  = listOfArgs[compTwoEle][1];
            description = listOfArgs[compTwoEle][2];
            if (dsData[FWUKL1_MAP_REG_DATA] == comparison)
              dsData[FWUKL1_MAP_REG_DATA] = description;
            else
              exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegDescrition(): Failed to convert the hex value 0x" + dsData[FWUKL1_MAP_REG_DATA] + " for the register " + regName + " to a text description as the value was not recognised. The value was returned.", "2");
          }// if not enough arguments.
          else
            exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegDescrition(): Not given a large enough list of aruments. Require 2, given " + dynlen(listOfArgs[compTwoEle]) + ". Hex value of data may have been returned.", "2");
        }// else(description==comparison))
      }// if(1<dynlen(listOfArgs[1]))
      else
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegDescrition(): Not given enough arguments to parse the data. Require 2, given " + dynlen(listOfArgs[compOneEle]) + ". Hex value of data may have been returned.", "2");
    }// if(2<dynlen(listOfArgs))
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegDescrition(): Not given a large enough list of arguments. Require at least 3, given " + dynlen(listOfArgs) + ". Hex value of data may have been returned.", "2");
  }// if got hex value.
  // Done.
  return dsData;
}// _fwUkl1_convertByteToAbsRegDescrition()

/*!
 * This takes the data as a hex string and converts it to a dyn_char which can be written to the hardware register. It also converts the mask
 * from the given string to a dyn_char which can be sent to the server. The string is treated as little endian.
 *
 * \param  regName The abstract register name that the data is being read for.
 * \param  hwRegName Name of the hardware register that is to be written to.
 * \param  hexData The hex string containing the data. It does not need to be padded.
 * \param  hexMask This is the hex string that is the mask for the given data. It is applied to the data before it is added to the dcData
 *           to ensure it does not go out of bounds. The mask is also added to the given mask such that it can be written to the hardware.
 * \param  dcData The dyn_char that the new value is to be written into. Returned by reference.
 * \param  dcMask The dyn_char that is the mask for the given data. Returned by reference.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The dyn_char's for the data and mask can already contain data for that hardware register and only the bit relevant to the given
 * data and mask will be updated. These dyn_char's can also be empty and the function will create them such that they are the correct size.
 *
 * No changes are made to the data or mask unless the parsing is successful.
 */
void _fwUkl1_convertAbsRegHexToByteLittleEndian(const string &regName, const string &hwRegName, string hexData, string hexMask, dyn_char &dcData, dyn_char &dcMask, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // Get the arugments for the function for this register.
  dyn_dyn_string listOfArgs = fwUkl1_getExtraRegisterInformation(regName, FWUKL1_CONFIG_WRITE_XTR_INFO, exceptionInfo);
  // Must have at least one list of arugments.
  if ( 0 < dynlen(listOfArgs) ) {
    // Need one argument.
    if ( 0 < dynlen(listOfArgs[1]) ) {
      // The start bit is the position in the register and as we will need to place the bits within a dyn_char need to know
      // the byte (element) to put it in plus the bit within that element.
      const int startBit = listOfArgs[1][1];
      // See what register type it is.
      int regType  = fwUkl1_getHwRegType(regName, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Failed to determine the hardware register type for " + regName + ". Cannot continue with the data parsing.", "2", exInfoSize) ) {
        // Get the number of bytes that are required for the register.
        int numBytes;
        // Look in the default settings for this register, they should always be the same as the values for each board and then we don't
        // have to specify the UKL1 board specifically. 
        dyn_dyn_anytype settingsList;
        // Have to strip the leading full stop from the hardware name if it is there.
        string strippedHwRegName = hwRegName;
        if ( 0 == strpos(hwRegName, ".") )
          strippedHwRegName = substr(hwRegName, 1);
        fwHw_getDefaultSpecificSettings(FWUKL1_HW_TYPE, makeDynString(strippedHwRegName), settingsList);
        // Where we look for it depends on the register type.
        switch (regType) {
          case FWHW_LBUS:
            // The FPGA registers are given in 16-bit words we are dealing with bytes.
            // Actually as far as the hardware tool is concerned they are 32-bit words, but we will take that into account later
            // as it is more complex than just adding padding to one end. Can be considered 16-bit words.
            numBytes = (int)settingsList[1][3] * 2;// 3rd item in first register, only ever ask for 1 reg.
            break;
          case FWHW_GBE:
            // The GBE registers are given in 32-bit words but we are dealing with bytes.
            numBytes = (int)settingsList[1][2] * 4;// 2nd item in first register, only ever ask for 1 reg.
            break;
          case FWHW_I2C:
            // The I2C registers are given in bytes and we are dealing with bytes.
            numBytes = (int)settingsList[1][4];// 1st item in first register, only ever ask for 1 reg.
            break;
          case FWHW_UserSpecific:
            // The user specific registers are FIFOs in our case and we need to know how many bytes in
            // total to get, which is specified in element 4. This is the number of words and must
            // be multiplied by the size of the words.
            numBytes = (int)settingsList[1][4] * (int)settingsList[1][3];// Only ever ask for 1 reg.
            break;
          default:
            numBytes = -1;
            exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Failed to recognise the register type " + regType + ". Could not determine number of bytes required for the given register. Cannot write to register without possibly corrupting it.", "2");
            break;
        }// switch(regType)
        if ( -1 != numBytes ) {
          // Bit shift the hex data and mask.
          fwUkl1_leftShiftString(hexData, startBit);
          fwUkl1_leftShiftString(hexMask, startBit);
          // Now ensure there are enough 0's for the entire word by adding them to the beginning of the string.
          fwUkl1_padString(hexData, numBytes);
          fwUkl1_padString(hexMask, numBytes);
          // Need to swap the bytes of the data and the mask around as we are dealing with little endian data.
          string swappedHexData;
          string swappedHexMask;
          // Must be padded to a word boundary so can gauranty we have an even number of characters.
          const int numCharacters = strlen(hexData);
          // Strings are zero based!
          for (int i=numCharacters-1; i>=0; i-=2) {
            // Remember a pair of characters is a byte and we preserve the bit order in the bytes.
            swappedHexData += hexData[i-1];
            swappedHexData += hexData[i];
            swappedHexMask += hexMask[i-1];
            swappedHexMask += hexMask[i];
          }// for i
          // Now convert to a dyn_char. Exactly how depends on the register type.
          dyn_char tmpData;
          dyn_char tmpMask;
          if ( FWHW_LBUS == regType ) {
            tmpData = _fwUkl1_fpgaizeData(swappedHexData);
            tmpMask = _fwUkl1_fpgaizeData(swappedHexMask);
          }// if(FWHW_LBUS==regType)
          else {
            tmpData = fwCcpc_convertHexToByte(swappedHexData);
            tmpMask = fwCcpc_convertHexToByte(swappedHexMask);
          }// else(FWHW_LBUS==regType)

          // Now we need to correct for the FPGA data after _fwUkl1_fpgaizeData, which expects normal data!
          if ( FWHW_LBUS == regType )
            numBytes *= 2;

          // We need to ensure the data is of the correct type if empty and has at least 1 element.
          if ( 0 == dynlen(dcData) )
            dcData = makeDynChar();
          // and that the dcData is the correct size
          while (numBytes > dynlen(dcData))
            dynAppend(dcData, 0);
          
          // Same for the mask.
          if ( 0 == dynlen(dcMask) )
            dcMask = makeDynChar();
          // and that the dcData is the correct size
          while (numBytes > dynlen(dcMask))
            dynAppend(dcMask, 0);
          
          // Now loop over all the data and insert it into the existing data ensure that it doesn't exceed its bounds.
          // Same goes for the mask.
          for ( int i = 1; i <= numBytes; ++i ) {
            // The data must be prepared first by zeroing the bits we intend to over write. AND with NOT the mask to do this.
            // NOT mask gives you zeros where you want to write and ones where you don't. When you AND with the data the zeros
            // cause the bits you want to write to become zero. The ones means you leave the rest of the data untouched.
            dcData[i] &= ~tmpMask[i];
            // Now we can OR with the data (with mask taken into account) and it will set the zeros to what ever it is.
            dcData[i] |= tmpData[i] & tmpMask[i];
            // The mask is simplier as we just want to ensure the mask bits are 1.
            dcMask[i] |= tmpMask[i];
          }// for i
        }// if(-1!=numBytes)
        // No else already flagged the error.
      }// if fwUkl1_getHwRegType successful.
    }// if(1<dynlen(listOfArgs[1]))
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Not given enough arguments to determine the starting bit for the hex value. Require 1, given " + dynlen(listOfArgs[1]) + ".", "2");
  }// if(0<dynlen(listOfArgs))
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Not given a large enough list of arguments. Require 1, given " + dynlen(listOfArgs) + ".", "2");
  // Done.
  return;
}// _fwUkl1_convertAbsRegHexToByteLittleEndian()

/*!
 * This takes the data from the hardware register and converts it to a hex number based on the requirements
 * for the abstract hardware register. Ensures that the bytes are treated as little endian.
 *
 * \param  regName The abstract register name that the data is being read for.
 * \param  dcData A dyn_char containing the data read back from the hardware register.
 * \param  pad Optionally the returned hex value can be padded to the maxmimum number of bytes it can take.
 * \param  forStatus If the description to be retrieved for the status display or configuration display.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_string Returns a dyn_string containing the value read back from the hardware register for the given
 *           abstract register in FWUKL1_MAP_REG_DATA element and the colour (if any) associated with that value
 *           in FWUKL1_MAP_XTR_DATA. Will be padded to the nearest byte to the length. Will be Fail and FwStateAttention1
 *           in the event of a read failure.
 */
dyn_string _fwUkl1_convertByteToAbsRegHexLittleEndian(const string &regName, const dyn_char &dcData, bool pad, bool forStatus, dyn_string &exceptionInfo) {
  // We must reverse the byte ordering of the array.
  // For this we will need to know its size.
  const int numBytes = dynlen(dcData);
  // This stores the reversed bytes.
  dyn_char swappedDcData;
  if (0 < numBytes) {
    // Have a none zero number of bytes.
    // Loop through them swapping the first element for last, 2nd for 2nd to last etc..
    for (int byte=1; byte<=numBytes; ++byte)
      swappedDcData[numBytes-(byte-1)] = dcData[byte];
  }// if(0<numBytes)
  else if (0 == numBytes)
    return "";
  else {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegHexLittleEndian(): PVSS error, failed to determine size of byte array, could not convert to hex string.", "1");
    return "Fail";
  }// else failed
  
  // Done.
  return _fwUkl1_convertByteToAbsRegHex(regName, swappedDcData, pad, forStatus, exceptionInfo);
}// _fwUkl1_convertByteToAbsRegHexLittleEndian()

/*!
 * This takes the data as a hex string and converts it to a dyn_char which can be written to the hardware register. It also converts the mask
 * from the given string to a dyn_char which can be sent to the server.
 *
 * \param  regName The abstract register name that the data is being read for.
 * \param  hwRegName Name of the hardware register that is to be written to.
 * \param  hexData The hex string containing the data. It does not need to be padded.
 * \param  hexMask This is the hex string that is the mask for the given data. It is applied to the data before it is added to the dcData
 *           to ensure it does not go out of bounds. The mask is also added to the given mask such that it can be written to the hardware.
 * \param  dcData The dyn_char that the new value is to be written into. Returned by reference.
 * \param  dcMask The dyn_char that is the mask for the given data. Returned by reference.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The dyn_char's for the data and mask can already contain data for that hardware register and only the bit relevant to the given
 * data and mask will be updated. These dyn_char's can also be empty and the function will create them such that they are the correct size.
 *
 * No changes are made to the data or mask unless the parsing is successful.
 */
void _fwUkl1_convertAbsRegHexToByte(const string &regName, const string &hwRegName, string hexData, string hexMask, dyn_char &dcData, dyn_char &dcMask, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // Get the arugments for the function for this register.
  dyn_dyn_string listOfArgs = fwUkl1_getExtraRegisterInformation(regName, FWUKL1_CONFIG_WRITE_XTR_INFO, exceptionInfo);
  // Must have at least one list of arguments.
  if (0 < dynlen(listOfArgs)) {
    // Need one argument.
    if (0 < dynlen(listOfArgs[1])) {
      // The start bit is the position in the register and as we will need to place the bits within a dyn_char need to know
      // the byte (element) to put it in plus the bit within that element.
      const int startBit = listOfArgs[1][1];
      // See what register type it is.
      int regType  = fwUkl1_getHwRegType(regName, exceptionInfo);
      if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Failed to determine the hardware register type for " + regName + ". Cannot continue with the data parsing.", "2", exInfoSize)) {
        // Get the number of bytes that are required for the register.
        int numBytes;
        // Look in the default settings for this register, they should always be the same as the values for each board and then we don't
        // have to specify the UKL1 board specifically. 
        dyn_dyn_anytype settingsList;
        // Have to strip the leading full stop from the hardware name if it is there.
        string strippedHwRegName = hwRegName;
        if (0 == strpos(hwRegName, "."))
          strippedHwRegName = substr(hwRegName, 1);
        fwHw_getDefaultSpecificSettings(FWUKL1_HW_TYPE, makeDynString(strippedHwRegName), settingsList);
        // Where we look for it depends on the register type.
        switch (regType) {
          case FWHW_LBUS:
            // The FPGA registers are given in 16-bit words we are dealing with bytes.
            // Actually as far as the hardware tool is concerned they are 32-bit words, but we will take that into account later
            // as it is more complex than just adding padding to one end. Can be considered 16-bit words.
            numBytes = (int)settingsList[1][3] * 2;// 3rd item in first register, only ever ask for 1 reg.
            break;
          case FWHW_GBE:
            // The GBE registers are given in 32-bit words but we are dealing with bytes.
            numBytes = (int)settingsList[1][2] * 4;// 2nd item in first register, only ever ask for 1 reg.
            break;
          case FWHW_I2C:
            // The I2C registers are given in bytes and we are dealing with bytes.
            numBytes = (int)settingsList[1][4];// 1st item in first register, only ever ask for 1 reg.
            break;
          case FWHW_UserSpecific:
            // The user specific registers are FIFOs in our case and we need to know how many bytes in
            // total to get, which is specified in element 4. This is the number of words and must
            // be multiplied by the size of the words.
            numBytes = (int)settingsList[1][4] * (int)settingsList[1][3];// Only ever ask for 1 reg.
            break;
          default:
            numBytes = -1;
            exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Failed to recognise the register type " + regType + ". Could not determine number of bytes required for the given register. Cannot write to register without possibly corrupting it.", "2");
            break;
        }// switch(regType)
        if (-1 != numBytes) {
          // Bit shift the hex data and mask.
          fwUkl1_leftShiftString(hexData, startBit);
          fwUkl1_leftShiftString(hexMask, startBit);
          // Now ensure there are enough 0's for the entire word by adding them to the beginning of the string.
          fwUkl1_padString(hexData, numBytes);
          fwUkl1_padString(hexMask, numBytes);
          // Now convert to a dyn_char. Exactly how depends on the register type.
          dyn_char tmpData;
          dyn_char tmpMask;
          if (FWHW_LBUS == regType) {
            tmpData = _fwUkl1_fpgaizeData(hexData);
            tmpMask = _fwUkl1_fpgaizeData(hexMask);
          }// if(FWHW_LBUS==regType)
          else {
            tmpData = fwCcpc_convertHexToByte(hexData);
            tmpMask = fwCcpc_convertHexToByte(hexMask);
          }// else(FWHW_LBUS==regType)

          // Now we need to correct for the FPGA data after _fwUkl1_fpgaizeData, which expects normal data!
          if (FWHW_LBUS == regType)
            numBytes *= 2;

          // We need to ensure the data is of the correct type if empty and has at least 1 element.
          if (0 == dynlen(dcData))
            dcData = makeDynChar();
          // and that the dcData is the correct size
          while (numBytes > dynlen(dcData))
            dynAppend(dcData, 0);
          
          // Same for the mask.
          if (0 == dynlen(dcMask))
            dcMask = makeDynChar();
          // and that the dcData is the correct size
          while (numBytes > dynlen(dcMask))
            dynAppend(dcMask, 0);
          
          // Now loop over all the data and insert it into the existing data ensure that it doesn't exceed its bounds.
          // Same goes for the mask.
          for (int i=1; i<=numBytes; ++i) {
            // The data must be prepared first by zeroing the bits we intend to over write. AND with NOT the mask to do this.
            // NOT mask gives you zeros where you want to write and ones where you don't. When you AND with the data the zeros
            // cause the bits you want to write to become zero. The ones means you leave the rest of the data untouched.
            dcData[i] &= ~tmpMask[i];
            // Now we can OR with the data (with mask taken into account) and it will set the zeros to what ever it is.
            dcData[i] |= tmpData[i] & tmpMask[i];
            // The mask is simplier as we just want to ensure the mask bits are 1.
            dcMask[i] |= tmpMask[i];
          }// for i
        }// if(-1!=numBytes)
        // No else already flagged the error.
      }// if fwUkl1_getHwRegType successful.
    }// if(1<dynlen(listOfArgs[1]))
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Not given enough arguments to determine the starting bit for the hex value. Require 1, given " + dynlen(listOfArgs[1]) + ".", "2");
  }// if(0<dynlen(listOfArgs))
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertAbsRegHexToByte(): Not given a large enough list of arguments. Require 1, given " + dynlen(listOfArgs) + ".", "2");
  // Done.
  return;
}// _fwUkl1_convertAbsRegHexToByte()

/*!
 * This takes the data from the hardware register and converts it to a hex number based on the requirements
 * for the abstract hardware register.
 *
 * \param  regName The abstract register name that the data is being read for.
 * \param  dcData A dyn_char containing the data read back from the hardware register.
 * \param  pad Optionally the returned hex value can be padded to the maxmimum number of bytes it can take.
 * \param  forStatus If the description to be retrieved for the status display or configuration display.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_string Returns a dyn_string containing the value read back from the hardware register for the given
 *           abstract register in FWUKL1_MAP_REG_DATA element and the colour (if any) associated with that value
 *           in FWUKL1_MAP_XTR_DATA. Will be padded to the nearest byte to the length. Will be Fail and FwStateAttention1
 *           in the event of a read failure.
 */
dyn_string _fwUkl1_convertByteToAbsRegHex(const string &regName, const dyn_char &dcData, bool pad, bool forStatus, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // This will hold the output data.
  dyn_string dsData = makeDynString("Fail", "FwStateAttention1");
  // Get the arugments for the function for this register.
  string dpe = FWUKL1_CONFIG_READ_XTR_INFO;
  if (forStatus)
    dpe = FWUKL1_STATUS_READ_XTR_INFO;
  dyn_dyn_string listOfArgs = fwUkl1_getExtraRegisterInformation(regName, dpe, exceptionInfo);
  // Must have at least one list of arugments.
  if (0 < dynlen(listOfArgs)) {
    // Need two arguments.
    if (1 < dynlen(listOfArgs[1])) {
      int startBit = listOfArgs[1][1];
      int length   = listOfArgs[1][2];
      // Need to decide if it is for an LBUS register or not.
      int regType  = fwUkl1_getHwRegType(regName, exceptionInfo);
      if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegHex(): Failed to determine the hardware register type for " + regName + ". Cannot continue with the data parsing.", "2", exInfoSize)) {
        bool fpga = (regType==FWHW_LBUS ? TRUE:FALSE);
        dsData[FWUKL1_MAP_REG_DATA] = _fwUkl1_convertByteToHex(dcData, startBit, length, fpga);
        // Before we do any padding look to see if there are any limits associated with this setting.
        listOfArgs = fwUkl1_getExtraRegisterInformation(regName, FWUKL1_VALID_RANGE_INFO, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_convertByteToAbsRegHex(): Could not retrieve the limit imformation for " + regName + ". Display value with no associated colour", "", exInfoSize)) {
          if (0 == dynlen(listOfArgs))
            dsData[FWUKL1_MAP_XTR_DATA] = "FwBackgroundInputText";
          else {
            // Start at the lowest limit and break once we have found the appropriate range.
            for (int arg=1; arg<=dynlen(listOfArgs); ++arg) {
              // Need at least 2 arguments for the value and colour.
              if (1 < dynlen(listOfArgs[arg])) {
                string value = listOfArgs[arg][1];
                string colour = listOfArgs[arg][2];
                if (dsData[FWUKL1_MAP_REG_DATA] <= value) {
                  dsData[FWUKL1_MAP_XTR_DATA] = colour;
                  break;
                }// if lower than limit.
              }// if got value and colour.
              else
                exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_convertByteToAbsRegHex(): When checking the range limit for " + regName + " found only " + dynlen(listOfArgs[arg]) + " arguments, require 2. One for limit and the other for colour associated with range.", "");
            }// for arg
            // If after the loop we still haven't assigned a colour then give it one.
            if ("FwStateAttention1" == dsData[FWUKL1_MAP_XTR_DATA])
              dsData[FWUKL1_MAP_XTR_DATA] = "FwBackgroundInputText";
          }// else choose colour from range.
        }// if got some arguments.
        else
          dsData[FWUKL1_MAP_XTR_DATA] = "FwBackgroundInputText";
        
        // Now pad the data before returning it.
        if (pad) {
          // Divide by length in bits 8 and then round up. This will give the minimum number of bytes that can contain this number of bits.
          int padding = ceil(length/8.0);
          fwUkl1_padString(dsData[FWUKL1_MAP_REG_DATA], padding);
        }// if(pad)
        
        if ("Channel0.Disable" == regName) {
          DebugTN(regName);
          DebugTN("dcData -- " + fwCcpc_convertByteToHex(dcData));
          DebugTN("startBit -- " + startBit);
          DebugTN("length -- " + length);
          DebugTN("fpga -- " + fpga);
          DebugTN("result -- " + dsData);
        }

      }// if got reg type
    }// if(1<dynlen(listOfArgs[1]))
    else
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegHex(): Not given enough arguments to pass to _fwUkl1_convertByteToHex(). Require 2, given " + dynlen(listOfArgs[1]) + ".", "2");
  }// if(0<dynlen(listOfArgs))
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_convertByteToAbsRegHex(): Not given a large enough list of arguments. Require 1, given " + dynlen(listOfArgs) + ".", "2");
  // Done.
  return dsData;
}// _fwUkl1_convertByteToAbsRegHex()

/*!
 * Takes a string representing a hex number and converts it to a dyn_char assuming the input data is little endian.
 *
 * \param  dcData A dyn_char containing the data read back from the hardware register.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return dyn_char Returns the dyn_char.
 *
 * The fwCcpc_convertHexToByte function assumes big endian data.
 */
dyn_char fwUkl1_convertHexToByteLittleEndian(const string &sData, dyn_string &exceptionInfo) {
  // Convert the hex data to a dyn_char.
  dyn_char dcData = fwCcpc_convertHexToByte(sData);
  // We must reverse the byte ordering of the array.
  // For this we will need to know its size.
  const int numBytes = dynlen(dcData);
  // This stores the reversed bytes.
  dyn_char swappedDcData;
  if (0 <= numBytes) {
    // Have a none zero number of bytes.
    // Loop through them swapping the first element for last, 2nd for 2nd to last etc..
    for (int byte=1; byte<=numBytes; ++byte)
      swappedDcData[numBytes-(byte-1)] = dcData[byte];
  }// if(0<numBytes)
  else
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_convertHexToByteLittleEndian(): PVSS error, failed to determine size of byte array, could not convert to hex string.", "1");
  
  // Done.
  return swappedDcData;
}// fwUkl1_convertHexToByteLittleEndian()

/*!
 * This takes the data that is read by from the L1 board FPGAs (LBUS), which is read as a series of 32-bit words,
 * and converts it to a series of 16-bit words can then be decoded to return the register value.
 *
 * \param  dcData The dyn_char array that contains the bytes read back from the hardware, will have the upper 2-bytes
 *           removed from the array. Assume byte ordering that matches that done read by the FW CCPC library. Return
 *           by reference.
 * \return dyn_char The dyn_char that now has its bytes in an order that can be decoded by the fwCcpc library
 *           to return a string representing the hex number.
 *
 * The function performs two tasks. First it considers the byte array as a group of 32-bit numbers and removes
 * the upper two bytes to reduce them to a 16-bit number. The resulting array must then be word swapped, however
 * the words must maintain their internal byte ordering. If the register is not a mutliple of 32, which if it is a
 * valid LBUS read it must be, then it will be padded.
 */
void _fwUkl1_convert32To16BitWords(dyn_char& dcData) {
  // Convert the read data from a series of 32-bit words to a series of 16-bit words.

  // Get the number of bytes.
  int numBytes = dynlen(dcData);
  // Check that we have a multiple of 32-bits.
  while (0 != numBytes%4) {
    // The most significant word should be in the last four bytes. We need to pad this word.
    // The most significant byte of this word is the at size-4.
    int position = numBytes - numBytes%4 + 1;
    dynInsertAt(dcData, 0x00, position);
    numBytes = dynlen(dcData);
  }// while(0!=(numBytes%4))

  // The loop must run over the number of 32-bit words in the read data.
  const unsigned total32Words = dynlen(dcData)/4;
  for (unsigned word=0; word<total32Words; ++word) {
    // We need to remove the first two elements in that word.
    // This will follow the sequence of odd numbers.
    // After we delete the first element the second element becomes the first, so we always delete the same one.
    dynRemove(dcData, 2*word+1);
    dynRemove(dcData, 2*word+1);
  }// for word

  // In the event that we have read back more than one word then we will need to swap the ordering of the 16-bit
  // words contained within the array.
  const unsigned total16Words = dynlen(dcData)/2;
  if (1 < total16Words) {
    // Copy the data into a temporary array.
    const dyn_char tmpData = dcData;
    // Loop over all the 16-bit words of the array and put them back into the original in the reverse order.
    const unsigned totalElements = dynlen(dcData);
    for (unsigned word=0; word<totalElements; word+=2) {
      // Need to keep the byte ordering within the word, but word swap the array. 16-bit words.
      dcData[word+1] = tmpData[totalElements-word-1];
      dcData[word+2] = tmpData[totalElements-word];
    }// for word
  }// if(1<total16Words)
  
  // Done.
  return;
}// _fwUkl1_convert32To16BitWords()

/*!
 * Parses the given input data map and converts it to values that can be written to the hardware or saved in the ConfDB.
 *
 * \param  inData This is a map that associates a register name with the data that is to be written and any extra
 *           information required to write that data. A full description is given in fwUkl1_write().
 * \param  outData This is a map that associates a hardware register name with the dyn_char data to be written
 *           and the dyn_char mask to be applied to that data on the server end. FWUKL1_MAP_REG_DATA contains the
 *           hardware register data and FWUKL1_MAP_XTR_DATA contains the mask. These are kept separate deliberately.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void _fwUkl1_parseRegistersForSet(const mapping &inData, mapping &outData, dyn_string &exceptionInfo) {
  // Keep track of the exception information.
  int exInfoSize = dynlen(exceptionInfo);
  // We use the outData mapping to associate the register names with the data that is to be written.
  // By using a mapping we ensure that multiple entries for a given hardware register don't occur
  // and look ups should be faster than searching dyn structures and counting the index.
  // We will use the same structure as the inData with register values in the FWUKL1_MAP_REG_DATA element
  // and put the mask in the FWUKL1_MAP_XTR_DATA element. In this case everything will be dyn_char's.
  // The name of the function to use to write the data is given in FWUKL1_MAP_XTR_DATA+1.
  // Get all the register names were given and loop over them parsing the data as necessary.
  const dyn_string regNames = mappingKeys(inData);
  const int numRegs = dynlen(regNames);
  // This will come in handy in a few places.
  // They must be declared outside the loop as they need to be built up over by the appropriate register.
  dyn_char dcData, dcMask;
  // Now loop over the keys and parse all the data.
  for (int regNum=1; regNum<=numRegs; ++regNum) {
    // This is the name of the register we were given. It doesn't always correspond to the hardware
    // register name. A few correspond to the same hardware register.
    string regName = regNames[regNum];
    // This is the name of the hardware register that is being written to. Retrieve it from the DP.
    // We just want the standard register as we are writting.
    string hwRegName = fwUkl1_getHwRegName(regName, TRUE, exceptionInfo);
    if (fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to retrieve the hardware register associated with the register "  + regName + ". Will not be able to write/save this data, skipping register.", "2", exInfoSize))
      continue;
    // This is the mask as a hex string that needs to be retrieved from the DP.
    string mask = fwUkl1_getHwRegMask(regName, exceptionInfo);
    if (fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to retrieve the mask associated with the register "  + regName + ". Will not be able to write/save this data, skipping register.", "2", exInfoSize))
      continue;
    // This is the name of the function to use to write hardware register. Retrieve it from the DP.
    string writefunc = fwUkl1_getWriteFunction(regName, exceptionInfo);
    if (fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to retrieve the write function associated with the register "  + regName + ". Will not be able to write/save this data, skipping register.", "2", exInfoSize))
      continue;
    
    // Get hold of the data in the form of a string that we have been given. It will save deferencing in a few places.
    string sData = inData[regName][FWUKL1_MAP_REG_DATA];
    // If the hardware register we have been given is the null register than we should not enter it into the map.
    if (FWUKL1_NULL_HWREG != hwRegName) {
      // Check to see if the hardware register exists already and if not create it with blank defaults.
      // This is best done here and not in the case statements as it means that those that share a HW register
      // can always rely on it existing. They never have to check.
      if (!mappingHasKey(outData, hwRegName)) {
        outData[hwRegName] = makeDynMixed();
        outData[hwRegName][FWUKL1_MAP_REG_DATA]   = makeDynChar();
        outData[hwRegName][FWUKL1_MAP_XTR_DATA]   = makeDynChar();
        outData[hwRegName][FWUKL1_MAP_XTR_DATA+1] = writefunc;
      }// if(!mappingHasKey())
    
      // Must populate the values with what ever we had before such that we have add in the data correctly.
      // If the key didn't exist then we are ensured that it will be empty.
      dcData = outData[hwRegName][FWUKL1_MAP_REG_DATA];
      dcMask = outData[hwRegName][FWUKL1_MAP_XTR_DATA];
    }// if(FWUKL1_NULL_HWREG!=hwRegName)
    // Use a switch statement to decide what to do.
    switch (regName) {
      // The MAC addresses are dealt with in the same way.
      case "MacDestinationAddress":
      case "MacSourceAddress":
        // Always over write any previous information here, it just doesn't matter.
        // We also use the same for hardware and high level.
        // regName is not necessarily the name we switched on.
        outData[hwRegName][FWUKL1_MAP_REG_DATA] = _fwUkl1_convertMacToByte(sData, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_parseRegistersForSet(): Failed to parse the MAC address removed " + regName + " from the list of parsed registers.", "2", exInfoSize))
          outData[hwRegName][FWUKL1_MAP_XTR_DATA] = _fwUkl1_fpgaizeData(mask);// Don't need to bother with the mask building here only one mask ever.
        // Done.
        break;

      // The IP addresses are dealt with in the same way.
      case "IpDestinationAddress":
      case "IpSourceAddress":
        // Always over write any previous information here, it just doesn't matter.
        // We also use the same for hardware and high level.
        // regName is not necessarily the name we switched on.
        outData[hwRegName][FWUKL1_MAP_REG_DATA] = _fwUkl1_convertIpToByte(sData, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_parseRegistersForSet(): Failed to parse the IP address removed " + regName + " from the list of parsed registers.", "2", exInfoSize))
          outData[hwRegName][FWUKL1_MAP_XTR_DATA] = _fwUkl1_fpgaizeData(mask, exceptionInfo);// Don't need to bother with the mask building here only one mask ever.
        // Done.
        break;

      // These registers are all 16-bit decimal numbers. Convert them first to hex and then let them be sorted out by the hex to byte function.
      case "MepEventCount":
      case "MTU":
      case "MepTruncationHighWatermark":
      case "ManualMepGbeOutputPortNumber":
      case "GbePort0TxFifoThreshold":
      case "GbePort1TxFifoThreshold":
      case "GbePort2TxFifoThreshold":
      case "GbePort3TxFifoThreshold":
      case "GbePort0TxFifoLowWatermark":
      case "GbePort1TxFifoLowWatermark":
      case "GbePort2TxFifoLowWatermark":
      case "GbePort3TxFifoLowWatermark":
      case "GbePort0TxFifoHighWatermark":
      case "GbePort1TxFifoHighWatermark":
      case "GbePort2TxFifoHighWatermark":
      case "GbePort3TxFifoHighWatermark":
        // Convert from decimal to hex.
        sData = fwCcpc_convertDecToHex(sData);
        // No break now goes to the hex dealing with code.
      // These are all hex strings and can therefore be dealt with in the same way.
      case "PartitionId":
      case "Protocol":
      case "TypeOfService":
      case "TimeToLive":
      case "Type":
      case "Version":
      case "TfcLatencyCompensation":
      case "L1Id":
      case "TtcrxControl":
      case "GbeJtagId":
      case "GbeMACSoftReset":
      case "GbeReceivePortReset":
      case "GbeTransmitPortReset":
      case "GbeReceiveFifoEnable":
      case "GbeMode":
      case "GbeClock":
      case "GbeSPI3ConfigTrnmtGlobal":
      case "GbeSPI3ConfigRcv":
      case "GbeRxStatus":
      case "GbePHYCtrl":
      case "GbePHYWriteData":
      case "GbePHYReadData":
      case "GbeMDIOControl":
        _fwUkl1_convertAbsRegHexToByte(regName, hwRegName, sData, mask, dcData, dcMask, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to parse the given data " + sData + " (may have been converted to hex) for the register " + regName + " no data will be written for this register.", "2", exInfoSize)) {
          outData[hwRegName][FWUKL1_MAP_REG_DATA] = dcData;
          outData[hwRegName][FWUKL1_MAP_XTR_DATA] = dcMask;
        }// if converted data
        // Done.
        break;

      // These are all hex little endian hex strings and can therefore be dealt with in the same way.
      case "LogicalId":
      case "RichId":
        _fwUkl1_convertAbsRegHexToByteLittleEndian(regName, hwRegName, sData, mask, dcData, dcMask, exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to parse the given data " + sData + " (may have been converted to hex) for the register " + regName + " no data will be written for this register.", "2", exInfoSize)) {
          outData[hwRegName][FWUKL1_MAP_REG_DATA] = dcData;
          outData[hwRegName][FWUKL1_MAP_XTR_DATA] = dcMask;
        }// if converted data
        // Done.
        break;
        
      // These all are given text descriptions.
      case "ThrottleMode":
      case "ThrottlePolarity":
      case "ThrottleOutput":
      case "ForceFixedMepDestinationAddress":
      case "MepBuildingMode":
      case "TfcDecoding":
      case "GbeTpaPolling":
      case "ManualMepGbeOutputPortControl":
      case "GbePort0Enable":
      case "GbePort1Enable":
      case "GbePort2Enable":
      case "GbePort3Enable":
      case "Ingress0Enable":
      case "Ingress1Enable":
      case "Ingress2Enable":
      case "Ingress3Enable":
      case "Channel0.Disable":
      case "Channel0.HpdPixelMode":
      case "Channel0.ZeroSuppression":
      case "Channel0.EmptyHpdSuppression":
      case "Channel0.Emulator":
      case "Channel0.EmulatorEvent":
      case "Channel0.EmulatorHpdPixelMode":
      case "Channel0.PixelMasks":
      case "Channel1.Disable":
      case "Channel1.HpdPixelMode":
      case "Channel1.ZeroSuppression":
      case "Channel1.EmptyHpdSuppression":
      case "Channel1.Emulator":
      case "Channel1.EmulatorEvent":
      case "Channel1.EmulatorHpdPixelMode":
      case "Channel1.PixelMasks":
      case "Channel2.Disable":
      case "Channel2.HpdPixelMode":
      case "Channel2.ZeroSuppression":
      case "Channel2.EmptyHpdSuppression":
      case "Channel2.Emulator":
      case "Channel2.EmulatorEvent":
      case "Channel2.EmulatorHpdPixelMode":
      case "Channel2.PixelMasks":
      case "Channel3.Disable":
      case "Channel3.HpdPixelMode":
      case "Channel3.ZeroSuppression":
      case "Channel3.EmptyHpdSuppression":
      case "Channel3.Emulator":
      case "Channel3.EmulatorEvent":
      case "Channel3.EmulatorHpdPixelMode":
      case "Channel3.PixelMasks":
      case "Channel4.Disable":
      case "Channel4.HpdPixelMode":
      case "Channel4.ZeroSuppression":
      case "Channel4.EmptyHpdSuppression":
      case "Channel4.Emulator":
      case "Channel4.EmulatorEvent":
      case "Channel4.EmulatorHpdPixelMode":
      case "Channel4.PixelMasks":
      case "Channel5.Disable":
      case "Channel5.HpdPixelMode":
      case "Channel5.ZeroSuppression":
      case "Channel5.EmptyHpdSuppression":
      case "Channel5.Emulator":
      case "Channel5.EmulatorEvent":
      case "Channel5.EmulatorHpdPixelMode":
      case "Channel5.PixelMasks":
      case "Channel6.Disable":
      case "Channel6.HpdPixelMode":
      case "Channel6.ZeroSuppression":
      case "Channel6.EmptyHpdSuppression":
      case "Channel6.Emulator":
      case "Channel6.EmulatorEvent":
      case "Channel6.EmulatorHpdPixelMode":
      case "Channel6.PixelMasks":
      case "Channel7.Disable":
      case "Channel7.HpdPixelMode":
      case "Channel7.ZeroSuppression":
      case "Channel7.EmptyHpdSuppression":
      case "Channel7.Emulator":
      case "Channel7.EmulatorEvent":
      case "Channel7.EmulatorHpdPixelMode":
      case "Channel7.PixelMasks":
      case "Channel8.Disable":
      case "Channel8.HpdPixelMode":
      case "Channel8.ZeroSuppression":
      case "Channel8.EmptyHpdSuppression":
      case "Channel8.Emulator":
      case "Channel8.EmulatorEvent":
      case "Channel8.EmulatorHpdPixelMode":
      case "Channel8.PixelMasks":
      case "Channel9.Disable":
      case "Channel9.HpdPixelMode":
      case "Channel9.ZeroSuppression":
      case "Channel9.EmptyHpdSuppression":
      case "Channel9.Emulator":
      case "Channel9.EmulatorEvent":
      case "Channel9.EmulatorHpdPixelMode":
      case "Channel9.PixelMasks":
      case "Channel10.PixelMasks":
      case "Channel11.PixelMasks":
      case "Channel12.PixelMasks":
      case "Channel13.PixelMasks":
      case "Channel14.PixelMasks":
      case "Channel15.PixelMasks":
      case "Channel16.PixelMasks":
      case "Channel17.PixelMasks":
      case "Channel18.PixelMasks":
      case "Channel19.PixelMasks":
      case "Channel20.PixelMasks":
      case "Channel21.PixelMasks":
      case "Channel22.PixelMasks":
      case "Channel23.PixelMasks":
      case "Channel24.PixelMasks":
      case "Channel25.PixelMasks":
      case "Channel26.PixelMasks":
      case "Channel27.PixelMasks":
      case "Channel28.PixelMasks":
      case "Channel29.PixelMasks":
      case "Channel30.PixelMasks":
      case "Channel31.PixelMasks":
      case "Channel32.PixelMasks":
      case "Channel33.PixelMasks":
      case "Channel34.PixelMasks":
      case "Channel35.PixelMasks":
      case "Channel10.Disable":
      case "Channel11.Disable":
      case "Channel12.Disable":
      case "Channel13.Disable":
      case "Channel14.Disable":
      case "Channel15.Disable":
      case "Channel16.Disable":
      case "Channel17.Disable":
      case "Channel18.Disable":
      case "Channel19.Disable":
      case "Channel20.Disable":
      case "Channel21.Disable":
      case "Channel22.Disable":
      case "Channel23.Disable":
      case "Channel24.Disable":
      case "Channel25.Disable":
      case "Channel26.Disable":
      case "Channel27.Disable":
      case "Channel28.Disable":
      case "Channel29.Disable":
      case "Channel30.Disable":
      case "Channel31.Disable":
      case "Channel32.Disable":
      case "Channel33.Disable":
      case "Channel34.Disable":
      case "Channel35.Disable":
      case "Channel10.HpdPixelMode":
      case "Channel11.HpdPixelMode":
      case "Channel12.HpdPixelMode":
      case "Channel13.HpdPixelMode":
      case "Channel14.HpdPixelMode":
      case "Channel15.HpdPixelMode":
      case "Channel16.HpdPixelMode":
      case "Channel17.HpdPixelMode":
      case "Channel18.HpdPixelMode":
      case "Channel19.HpdPixelMode":
      case "Channel20.HpdPixelMode":
      case "Channel21.HpdPixelMode":
      case "Channel22.HpdPixelMode":
      case "Channel23.HpdPixelMode":
      case "Channel24.HpdPixelMode":
      case "Channel25.HpdPixelMode":
      case "Channel26.HpdPixelMode":
      case "Channel27.HpdPixelMode":
      case "Channel28.HpdPixelMode":
      case "Channel29.HpdPixelMode":
      case "Channel30.HpdPixelMode":
      case "Channel31.HpdPixelMode":
      case "Channel32.HpdPixelMode":
      case "Channel33.HpdPixelMode":
      case "Channel34.HpdPixelMode":
      case "Channel35.HpdPixelMode":
      case "Channel10.ZeroSuppression":
      case "Channel11.ZeroSuppression":
      case "Channel12.ZeroSuppression":
      case "Channel13.ZeroSuppression":
      case "Channel14.ZeroSuppression":
      case "Channel15.ZeroSuppression":
      case "Channel16.ZeroSuppression":
      case "Channel17.ZeroSuppression":
      case "Channel18.ZeroSuppression":
      case "Channel19.ZeroSuppression":
      case "Channel20.ZeroSuppression":
      case "Channel21.ZeroSuppression":
      case "Channel22.ZeroSuppression":
      case "Channel23.ZeroSuppression":
      case "Channel24.ZeroSuppression":
      case "Channel25.ZeroSuppression":
      case "Channel26.ZeroSuppression":
      case "Channel27.ZeroSuppression":
      case "Channel28.ZeroSuppression":
      case "Channel29.ZeroSuppression":
      case "Channel30.ZeroSuppression":
      case "Channel31.ZeroSuppression":
      case "Channel32.ZeroSuppression":
      case "Channel33.ZeroSuppression":
      case "Channel34.ZeroSuppression":
      case "Channel35.ZeroSuppression":
      case "Channel10.EmptyHpdSuppression":
      case "Channel11.EmptyHpdSuppression":
      case "Channel12.EmptyHpdSuppression":
      case "Channel13.EmptyHpdSuppression":
      case "Channel14.EmptyHpdSuppression":
      case "Channel15.EmptyHpdSuppression":
      case "Channel16.EmptyHpdSuppression":
      case "Channel17.EmptyHpdSuppression":
      case "Channel18.EmptyHpdSuppression":
      case "Channel19.EmptyHpdSuppression":
      case "Channel20.EmptyHpdSuppression":
      case "Channel21.EmptyHpdSuppression":
      case "Channel22.EmptyHpdSuppression":
      case "Channel23.EmptyHpdSuppression":
      case "Channel24.EmptyHpdSuppression":
      case "Channel25.EmptyHpdSuppression":
      case "Channel26.EmptyHpdSuppression":
      case "Channel27.EmptyHpdSuppression":
      case "Channel28.EmptyHpdSuppression":
      case "Channel29.EmptyHpdSuppression":
      case "Channel30.EmptyHpdSuppression":
      case "Channel31.EmptyHpdSuppression":
      case "Channel32.EmptyHpdSuppression":
      case "Channel33.EmptyHpdSuppression":
      case "Channel34.EmptyHpdSuppression":
      case "Channel35.EmptyHpdSuppression":
      case "Channel10.Emulator":
      case "Channel11.Emulator":
      case "Channel12.Emulator":
      case "Channel13.Emulator":
      case "Channel14.Emulator":
      case "Channel15.Emulator":
      case "Channel16.Emulator":
      case "Channel17.Emulator":
      case "Channel18.Emulator":
      case "Channel19.Emulator":
      case "Channel20.Emulator":
      case "Channel21.Emulator":
      case "Channel22.Emulator":
      case "Channel23.Emulator":
      case "Channel24.Emulator":
      case "Channel25.Emulator":
      case "Channel26.Emulator":
      case "Channel27.Emulator":
      case "Channel28.Emulator":
      case "Channel29.Emulator":
      case "Channel30.Emulator":
      case "Channel31.Emulator":
      case "Channel32.Emulator":
      case "Channel33.Emulator":
      case "Channel34.Emulator":
      case "Channel35.Emulator":
      case "Channel10.EmulatorHpdPixelMode":
      case "Channel11.EmulatorHpdPixelMode":
      case "Channel12.EmulatorHpdPixelMode":
      case "Channel13.EmulatorHpdPixelMode":
      case "Channel14.EmulatorHpdPixelMode":
      case "Channel15.EmulatorHpdPixelMode":
      case "Channel16.EmulatorHpdPixelMode":
      case "Channel17.EmulatorHpdPixelMode":
      case "Channel18.EmulatorHpdPixelMode":
      case "Channel19.EmulatorHpdPixelMode":
      case "Channel20.EmulatorHpdPixelMode":
      case "Channel21.EmulatorHpdPixelMode":
      case "Channel22.EmulatorHpdPixelMode":
      case "Channel23.EmulatorHpdPixelMode":
      case "Channel24.EmulatorHpdPixelMode":
      case "Channel25.EmulatorHpdPixelMode":
      case "Channel26.EmulatorHpdPixelMode":
      case "Channel27.EmulatorHpdPixelMode":
      case "Channel28.EmulatorHpdPixelMode":
      case "Channel29.EmulatorHpdPixelMode":
      case "Channel30.EmulatorHpdPixelMode":
      case "Channel31.EmulatorHpdPixelMode":
      case "Channel32.EmulatorHpdPixelMode":
      case "Channel33.EmulatorHpdPixelMode":
      case "Channel34.EmulatorHpdPixelMode":
      case "Channel35.EmulatorHpdPixelMode":
      case "Channel10.EmulatorEvent":
      case "Channel11.EmulatorEvent":
      case "Channel12.EmulatorEvent":
      case "Channel13.EmulatorEvent":
      case "Channel14.EmulatorEvent":
      case "Channel15.EmulatorEvent":
      case "Channel16.EmulatorEvent":
      case "Channel17.EmulatorEvent":
      case "Channel18.EmulatorEvent":
      case "Channel19.EmulatorEvent":
      case "Channel20.EmulatorEvent":
      case "Channel21.EmulatorEvent":
      case "Channel22.EmulatorEvent":
      case "Channel23.EmulatorEvent":
      case "Channel24.EmulatorEvent":
      case "Channel25.EmulatorEvent":
      case "Channel26.EmulatorEvent":
      case "Channel27.EmulatorEvent":
      case "Channel28.EmulatorEvent":
      case "Channel29.EmulatorEvent":
      case "Channel30.EmulatorEvent":
      case "Channel31.EmulatorEvent":
      case "Channel32.EmulatorEvent":
      case "Channel33.EmulatorEvent":
      case "Channel34.EmulatorEvent":
      case "Channel35.EmulatorEvent":
        _fwUkl1_convertAbsRegDescritionToByte(regName, hwRegName, sData, mask, dcData, dcMask, exceptionInfo);
        if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Failed to parse the given data " + sData + " for the register " + regName + " no data will be written for this register.", "2", exInfoSize)) {
          outData[hwRegName][FWUKL1_MAP_REG_DATA] = dcData;
          outData[hwRegName][FWUKL1_MAP_XTR_DATA] = dcMask;
        }// if converted data
        // Done.
        break;

      case "ChannelAll.Disable":
      case "ChannelAll.HpdPixelMode":
      case "ChannelAll.ZeroSuppression":
      case "ChannelAll.EmptyHpdSuppression":
      case "ChannelAll.Emulator":
      case "ChannelAll.EmulatorEvent":
      case "ChannelAll.EmulatorHpdPixelMode":
      case "ChannelAll.PixelMasks":
        // This function takes our abstract register and string data and then converts it to the appropriate
        // mask and dyn_char data for each of the hardware registers associated with the 'ChannelAll'.
        _fwUkl1_convertAbsRegChannelAllToIndividualRegs(regName, sData, outData, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegisterForSet(): Failed to parse the given data " + sData + " for the register " + regName + " no data will be written for this register.", "2", exInfoSize);
        break;
        
      default:
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForSet(): Did not recognise the register " + regName + ", this register has not been parsed and will not be written/saved.", "2");
        // Done.
        break;
    }// switch(regName)
    // If at the end of the switch we haven't added any data for the register then we should remove it.
    // This indicates that there was an error and we didn't add any data. Don't want to remove data that
    // another abstract register is writing so only remove if empty.
    if (mappingHasKey(outData, hwRegName) && (0 == dynlen(outData[hwRegName]))) {
      mappingRemove(outData, hwRegName);
    }
  }// for regNum

  // Done.
  return;
}// _fwUkl1_parseRegistersForSet()

/*!
 * Parses the given input data map of data read from the hardware/DB and converts it to values that
 * are human readable and associated with an abstract register.
 *
 * \param  inData This is the raw data that has been read from the hardware or out of the database. The map
 *           should contain as a key the hardware register and the data is a dyn_char which is the raw
 *           data from the hardware/DB.
 * \param  outData This is a map that associates a register name with the string containing the human readable
 *           value of the register and a colour representing its state. FWUKL1_MAP_REG_DATA contains the string
 *           with the register data and FWUKL1_MAP_XTR_DATA contains the colour. The outData should be populated
 *           with keys such that the function knows which registers to attempted to retrieve from the inData.
 * \param  forStatus Typically register that return text will return different values if they are to be used
 *           to display the status of the registers rather than the value for configuration. Mostly it is just
 *           a tense change i.e. Enabled (status), Enable (configuration).
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The given map should have an empty entry for the FWUKL1_NULL_HWREG constant. The data associated with
 * the entry will never be used, but the key should at least exist.
 */
void _fwUkl1_parseRegistersForGet(const mapping &inData, mapping &outData, bool forStatus, dyn_string &exceptionInfo) {
  // Keep track of the exception information.
  int exInfoSize = dynlen(exceptionInfo);
  // I don't understand why but when in the loop the for a given group of case statements the exception message is the same as the first
  // one to be written. Perhaps I am over looking something simple, perhaps it is the evils of PVSS, but what ever it is it is annoying!
  // This string holds the error messages. It ensures the proper error messages are given.
  string errMsg = "";
  // Get the register names which are the key for the outData.
  const dyn_string regNames = mappingKeys(outData);
  const int numRegs = dynlen(regNames);
  // Now loop over the keys and parse all the data.
  for (int regNum=1; regNum<=numRegs; ++regNum) {
    // This is the name of the register we were given. It doesn't always correspond to the hardware
    // register name. A few correspond to the same hardware register.
    string regName = regNames[regNum];
    // Put it in the fail state until we know better.
    outData[regName] = makeDynString("Fail", "FwStateAttention1");
    // This is the name of the hardware register that is being read from. Retrieve it from the DP.
    // Here we always take the alternative register as those are the ones used for reading from the
    // status registers.
    string hwRegName = fwUkl1_getHwRegName(regName, FALSE, exceptionInfo);
    if (fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForGet(): Failed to retrieve the hardware register associated with the register "  + regName + ". Will not be able to read this data, skipping register.", "2", exInfoSize))
      continue;
    // Now we have the hwRegName lets check to see if we have data for it.
    if (!mappingHasKey(inData, hwRegName))
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_parseRegisterForGet(): The hardware register, " + hwRegName + ", associated with the register, " + regName + ", was not found.", "2");
    else {
      if ((FWUKL1_NULL_HWREG!=hwRegName) && (1>dynlen(inData[hwRegName])))
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_parseRegisterForGet(): There is no raw data associated with the hardware register " + hwRegName + ". Cannot parse the data for the register " + regName + ", thus cannot retrieve any data for this register.", "2");
      else {
        // We should never operate on the inData directly. It will be parsed multi times, need a local copy each time.
        dyn_char dcData = inData[hwRegName];
        if ("Channel0.Disable" == regName)
          DebugTN("hwRegName -- " + hwRegName);
        // Use a switch statement to decide what to do.
        // Deal with errors at the end.
        switch (regName) {
          // MAC addresses are deal with in the same way.
          case "MacDestinationAddress":
          case "MacSourceAddress":
            outData[regName] = _fwUkl1_convertByteToMac(dcData, exceptionInfo);
            errMsg = "_fwUkl1_parseRegisterForGet(): Failed to parse the MAC address (" + regName + ").";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;
          
          // IP addresses are deal with in the same way.
          case "IpDestinationAddress":
          case "IpSourceAddress":
            outData[regName] = _fwUkl1_convertByteToIp(dcData, exceptionInfo);
            errMsg = "_fwUkl1_parseRegisterForGet(): Failed to parse the IP address (" + regName + ").";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;

          // These are all registers displayed in hex.
          case "PartitionId":
          case "Protocol":
          case "TypeOfService":
          case "TimeToLive":
          case "Type":
          case "Version":
          case "TfcLatencyCompensation":
          case "TTCId":
          case "L1Id":
          case "TTCId":
          case "EventMergerReadState":
          case "TFCShortBroadcast":
          case "EthernetFormatterState":
          case "EgressInputFIFOLowWordOccupancy0":
          case "EgressInputFIFOLowWordOccupancy1":
          case "EgressInputFIFOLowWordOccupancy2":
          case "EgressInputFIFOLowWordOccupancy3":
          case "EgressInputFIFOHighWordOccupancy0":
          case "EgressInputFIFOHighWordOccupancy1":
          case "EgressInputFIFOHighWordOccupancy2":
          case "EgressInputFIFOHighWordOccupancy3":
          case "MEPCount":
          case "TFCFIFOOccupancy":
          case "TFCLongBroadcast":
          case "L0TriggerCount":
          case "ROTBroadcastCount":
          case "TFCDestinationCount":
          case "MEPSequenceCount":
          case "EgressInputFIFOLowWordRxCount0":
          case "EgressInputFIFOLowWordRxCount1":
          case "EgressInputFIFOLowWordRxCount2":
          case "EgressInputFIFOLowWordRxCount3":
          case "EgressInputFIFOHighWordRxCount0":
          case "EgressInputFIFOHighWordRxCount1":
          case "EgressInputFIFOHighWordRxCount2":
          case "EgressInputFIFOHighWordRxCount3":
          case "EgressInputFIFOLowWordTxCount0":
          case "EgressInputFIFOLowWordTxCount1":
          case "EgressInputFIFOLowWordTxCount2":
          case "EgressInputFIFOLowWordTxCount3":
          case "EgressInputFIFOHighWordTxCount0":
          case "EgressInputFIFOHighWordTxCount1":
          case "EgressInputFIFOHighWordTxCount2":
          case "EgressInputFIFOHighWordTxCount3":
          case "TtcrxControl":
          case "GbeJtagId":
          case "GbeReceiveFifoEnable":
          case "GbeMode":
          case "GbeClock":
          case "GbeSPI3ConfigTrnmtGlobal":
          case "GbeSPI3ConfigRcv":
          case "GbeRxStatus":
          case "GbePHYCtrl":
          case "GbePHYWriteData":
          case "GbePHYReadData":
          case "GbeMDIOControl":
            // Do pad this hex data such that it is the maximum length in bytes it can be.
            outData[regName] = _fwUkl1_convertByteToAbsRegHex(regName, dcData, TRUE, forStatus, exceptionInfo);
            errMsg = "_fwUkl1_parseRegistersForGet(): Failed to parse the data from " + hwRegName + " for " + regName + ".";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;
        
          // These are all registers displayed in hex but need to ensure it is little endian.
          case "SerialNumber":
          case "LogicalId":
          case "RichId":
            // Do pad this hex data such that it is the maximum length in bytes it can be.
            outData[regName] = _fwUkl1_convertByteToAbsRegHexLittleEndian(regName, dcData, TRUE, forStatus, exceptionInfo);
            errMsg = "_fwUkl1_parseRegistersForGet(): Failed to parse the data from " + hwRegName + " for " + regName + ".";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;
            
          // These registers are displayed in dec.
          case "MepEventCount":
          case "MTU":
          case "MepTruncationHighWatermark":
          case "ManualMepGbeOutputPortNumber":
          case "TemperatureSensorTop":
          case "TemperatureSensorMiddle":
          case "TemperatureSensorBottom":
          case "GbePort0TxFifoThreshold":
          case "GbePort1TxFifoThreshold":
          case "GbePort2TxFifoThreshold":
          case "GbePort3TxFifoThreshold":
          case "GbePort0TxFifoLowWatermark":
          case "GbePort1TxFifoLowWatermark":
          case "GbePort2TxFifoLowWatermark":
          case "GbePort3TxFifoLowWatermark":
          case "GbePort0TxFifoHighWatermark":
          case "GbePort1TxFifoHighWatermark":
          case "GbePort2TxFifoHighWatermark":
          case "GbePort3TxFifoHighWatermark":
            // Don't bother to pad the data it will be converted to decimal anyway.
            outData[regName] = _fwUkl1_convertByteToAbsRegHex(regName, dcData, FALSE, forStatus, exceptionInfo);
            errMsg = "_fwUkl1_parseRegistersForGet(): Failed to parse the data from " + hwRegName + " for " + regName + ".";
            if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize) )
              outData[regName][FWUKL1_MAP_REG_DATA] = fwCcpc_convertHexToDec(outData[regName][FWUKL1_MAP_REG_DATA]);
            break;
        
          // These registers all read back bits that are converted to a description.
          case "ThrottleMode":
          case "ThrottlePolarity":
          case "ThrottleOutput":
          case "ForceFixedMepDestinationAddress":
          case "MepBuildingMode":
          case "TfcDecoding":
          case "GbeTpaPolling":
          case "ManualMepGbeOutputPortControl":
          case "TFCFIFORequest":
          case "TTCrxReady":
          case "GTDataLowWordInhibit0":
          case "GTDataLowWordInhibit1":
          case "GTDataLowWordInhibit2":
          case "GTDataLowWordInhibit3":
          case "GTDataHighWordInhibit0":
          case "GTDataHighWordInhibit1":
          case "GTDataHighWordInhibit2":
          case "GTDataHighWordInhibit3":
          case "GTCommunicationWordInhibit0":
          case "GTCommunicationWordInhibit1":
          case "GTCommunicationWordInhibit2":
          case "GTCommunicationWordInhibit3":
          case "GTDataLinkInhibit0":
          case "GTDataLinkInhibit1":
          case "GTDataLinkInhibit2":
          case "GTDataLinkInhibit3":
          case "GBETx":
          case "MEPFifo":
          case "ThrottleAsserted":
          case "GTFIFO0":
          case "GTFIFO1":
          case "GTFIFO2":
          case "GTFIFO3":
          case "GTLowFIFODataReadout0":
          case "GTLowFIFODataReadout1":
          case "GTLowFIFODataReadout2":
          case "GTLowFIFODataReadout3":
          case "GTHighFIFODataReadout0":
          case "GTHighFIFODataReadout1":
          case "GTHighFIFODataReadout2":
          case "GTHighFIFODataReadout3":
          case "TFCL0Reset":
          case "TFCL1Reset":
          case "GBETransmitPortAvailable0":
          case "GBETransmitPortAvailable1":
          case "GBETransmitPortAvailable2":
          case "GBETransmitPortAvailable3":
          case "EthernetFragmentReady":
          case "MEPReady":
          case "FragmentRequest":
          case "MEPExtraction":
          case "FragmentExtraction":
          case "SequenceError":
          case "GTFIFOTxWait0":
          case "GTFIFOTxWait1":
          case "GTFIFOTxWait2":
          case "GTFIFOTxWait3":
          case "EgressInputFIFOLowWordSequenceError0":
          case "EgressInputFIFOLowWordSequenceError1":
          case "EgressInputFIFOLowWordSequenceError2":
          case "EgressInputFIFOLowWordSequenceError3":
          case "EgressInputFIFOHighWordSequenceError0":
          case "EgressInputFIFOHighWordSequenceError1":
          case "EgressInputFIFOHighWordSequenceError2":
          case "EgressInputFIFOHighWordSequenceError3":
          case "TTCrxReadyAlarm":
          case "SequenceErrorAlarm":
          case "GbePort0Enable":
          case "GbePort1Enable":
          case "GbePort2Enable":
          case "GbePort3Enable":
          case "Ingress0Enable":
          case "Ingress1Enable":
          case "Ingress2Enable":
          case "Ingress3Enable":
          case "Channel0.Disable":
          case "Channel0.HpdPixelMode":
          case "Channel0.ZeroSuppression":
          case "Channel0.EmptyHpdSuppression":
          case "Channel0.Emulator":
          case "Channel0.EmulatorEvent":
          case "Channel0.EmulatorHpdPixelMode":
          case "Channel0.PixelMasks":
          case "Channel1.Disable":
          case "Channel1.HpdPixelMode":
          case "Channel1.ZeroSuppression":
          case "Channel1.EmptyHpdSuppression":
          case "Channel1.Emulator":
          case "Channel1.EmulatorEvent":
          case "Channel1.EmulatorHpdPixelMode":
          case "Channel1.PixelMasks":
          case "Channel2.Disable":
          case "Channel2.HpdPixelMode":
          case "Channel2.ZeroSuppression":
          case "Channel2.EmptyHpdSuppression":
          case "Channel2.Emulator":
          case "Channel2.EmulatorEvent":
          case "Channel2.EmulatorHpdPixelMode":
          case "Channel2.PixelMasks":
          case "Channel3.Disable":
          case "Channel3.HpdPixelMode":
          case "Channel3.ZeroSuppression":
          case "Channel3.EmptyHpdSuppression":
          case "Channel3.Emulator":
          case "Channel3.EmulatorEvent":
          case "Channel3.EmulatorHpdPixelMode":
          case "Channel3.PixelMasks":
          case "Channel4.Disable":
          case "Channel4.HpdPixelMode":
          case "Channel4.ZeroSuppression":
          case "Channel4.EmptyHpdSuppression":
          case "Channel4.Emulator":
          case "Channel4.EmulatorEvent":
          case "Channel4.EmulatorHpdPixelMode":
          case "Channel4.PixelMasks":
          case "Channel5.Disable":
          case "Channel5.HpdPixelMode":
          case "Channel5.ZeroSuppression":
          case "Channel5.EmptyHpdSuppression":
          case "Channel5.Emulator":
          case "Channel5.EmulatorEvent":
          case "Channel5.EmulatorHpdPixelMode":
          case "Channel5.PixelMasks":
          case "Channel6.Disable":
          case "Channel6.HpdPixelMode":
          case "Channel6.ZeroSuppression":
          case "Channel6.EmptyHpdSuppression":
          case "Channel6.Emulator":
          case "Channel6.EmulatorEvent":
          case "Channel6.EmulatorHpdPixelMode":
          case "Channel6.PixelMasks":
          case "Channel7.Disable":
          case "Channel7.HpdPixelMode":
          case "Channel7.ZeroSuppression":
          case "Channel7.EmptyHpdSuppression":
          case "Channel7.Emulator":
          case "Channel7.EmulatorEvent":
          case "Channel7.EmulatorHpdPixelMode":
          case "Channel7.PixelMasks":
          case "Channel8.Disable":
          case "Channel8.HpdPixelMode":
          case "Channel8.ZeroSuppression":
          case "Channel8.EmptyHpdSuppression":
          case "Channel8.Emulator":
          case "Channel8.EmulatorEvent":
          case "Channel8.EmulatorHpdPixelMode":
          case "Channel8.PixelMasks":
          case "Channel9.Disable":
          case "Channel9.HpdPixelMode":
          case "Channel9.ZeroSuppression":
          case "Channel9.EmptyHpdSuppression":
          case "Channel9.Emulator":
          case "Channel9.EmulatorEvent":
          case "Channel9.EmulatorHpdPixelMode":
          case "Channel9.PixelMasks":
          case "Channel10.PixelMasks":
          case "Channel11.PixelMasks":
          case "Channel12.PixelMasks":
          case "Channel13.PixelMasks":
          case "Channel14.PixelMasks":
          case "Channel15.PixelMasks":
          case "Channel16.PixelMasks":
          case "Channel17.PixelMasks":
          case "Channel18.PixelMasks":
          case "Channel19.PixelMasks":
          case "Channel20.PixelMasks":
          case "Channel21.PixelMasks":
          case "Channel22.PixelMasks":
          case "Channel23.PixelMasks":
          case "Channel24.PixelMasks":
          case "Channel25.PixelMasks":
          case "Channel26.PixelMasks":
          case "Channel27.PixelMasks":
          case "Channel28.PixelMasks":
          case "Channel29.PixelMasks":
          case "Channel30.PixelMasks":
          case "Channel31.PixelMasks":
          case "Channel32.PixelMasks":
          case "Channel33.PixelMasks":
          case "Channel34.PixelMasks":
          case "Channel35.PixelMasks":
          case "Channel10.Disable":
          case "Channel11.Disable":
          case "Channel12.Disable":
          case "Channel13.Disable":
          case "Channel14.Disable":
          case "Channel15.Disable":
          case "Channel16.Disable":
          case "Channel17.Disable":
          case "Channel18.Disable":
          case "Channel19.Disable":
          case "Channel20.Disable":
          case "Channel21.Disable":
          case "Channel22.Disable":
          case "Channel23.Disable":
          case "Channel24.Disable":
          case "Channel25.Disable":
          case "Channel26.Disable":
          case "Channel27.Disable":
          case "Channel28.Disable":
          case "Channel29.Disable":
          case "Channel30.Disable":
          case "Channel31.Disable":
          case "Channel32.Disable":
          case "Channel33.Disable":
          case "Channel34.Disable":
          case "Channel35.Disable":
          case "Channel10.HpdPixelMode":
          case "Channel11.HpdPixelMode":
          case "Channel12.HpdPixelMode":
          case "Channel13.HpdPixelMode":
          case "Channel14.HpdPixelMode":
          case "Channel15.HpdPixelMode":
          case "Channel16.HpdPixelMode":
          case "Channel17.HpdPixelMode":
          case "Channel18.HpdPixelMode":
          case "Channel19.HpdPixelMode":
          case "Channel20.HpdPixelMode":
          case "Channel21.HpdPixelMode":
          case "Channel22.HpdPixelMode":
          case "Channel23.HpdPixelMode":
          case "Channel24.HpdPixelMode":
          case "Channel25.HpdPixelMode":
          case "Channel26.HpdPixelMode":
          case "Channel27.HpdPixelMode":
          case "Channel28.HpdPixelMode":
          case "Channel29.HpdPixelMode":
          case "Channel30.HpdPixelMode":
          case "Channel31.HpdPixelMode":
          case "Channel32.HpdPixelMode":
          case "Channel33.HpdPixelMode":
          case "Channel34.HpdPixelMode":
          case "Channel35.HpdPixelMode":
          case "Channel10.ZeroSuppression":
          case "Channel11.ZeroSuppression":
          case "Channel12.ZeroSuppression":
          case "Channel13.ZeroSuppression":
          case "Channel14.ZeroSuppression":
          case "Channel15.ZeroSuppression":
          case "Channel16.ZeroSuppression":
          case "Channel17.ZeroSuppression":
          case "Channel18.ZeroSuppression":
          case "Channel19.ZeroSuppression":
          case "Channel20.ZeroSuppression":
          case "Channel21.ZeroSuppression":
          case "Channel22.ZeroSuppression":
          case "Channel23.ZeroSuppression":
          case "Channel24.ZeroSuppression":
          case "Channel25.ZeroSuppression":
          case "Channel26.ZeroSuppression":
          case "Channel27.ZeroSuppression":
          case "Channel28.ZeroSuppression":
          case "Channel29.ZeroSuppression":
          case "Channel30.ZeroSuppression":
          case "Channel31.ZeroSuppression":
          case "Channel32.ZeroSuppression":
          case "Channel33.ZeroSuppression":
          case "Channel34.ZeroSuppression":
          case "Channel35.ZeroSuppression":
          case "Channel10.EmptyHpdSuppression":
          case "Channel11.EmptyHpdSuppression":
          case "Channel12.EmptyHpdSuppression":
          case "Channel13.EmptyHpdSuppression":
          case "Channel14.EmptyHpdSuppression":
          case "Channel15.EmptyHpdSuppression":
          case "Channel16.EmptyHpdSuppression":
          case "Channel17.EmptyHpdSuppression":
          case "Channel18.EmptyHpdSuppression":
          case "Channel19.EmptyHpdSuppression":
          case "Channel20.EmptyHpdSuppression":
          case "Channel21.EmptyHpdSuppression":
          case "Channel22.EmptyHpdSuppression":
          case "Channel23.EmptyHpdSuppression":
          case "Channel24.EmptyHpdSuppression":
          case "Channel25.EmptyHpdSuppression":
          case "Channel26.EmptyHpdSuppression":
          case "Channel27.EmptyHpdSuppression":
          case "Channel28.EmptyHpdSuppression":
          case "Channel29.EmptyHpdSuppression":
          case "Channel30.EmptyHpdSuppression":
          case "Channel31.EmptyHpdSuppression":
          case "Channel32.EmptyHpdSuppression":
          case "Channel33.EmptyHpdSuppression":
          case "Channel34.EmptyHpdSuppression":
          case "Channel35.EmptyHpdSuppression":
          case "Channel10.Emulator":
          case "Channel11.Emulator":
          case "Channel12.Emulator":
          case "Channel13.Emulator":
          case "Channel14.Emulator":
          case "Channel15.Emulator":
          case "Channel16.Emulator":
          case "Channel17.Emulator":
          case "Channel18.Emulator":
          case "Channel19.Emulator":
          case "Channel20.Emulator":
          case "Channel21.Emulator":
          case "Channel22.Emulator":
          case "Channel23.Emulator":
          case "Channel24.Emulator":
          case "Channel25.Emulator":
          case "Channel26.Emulator":
          case "Channel27.Emulator":
          case "Channel28.Emulator":
          case "Channel29.Emulator":
          case "Channel30.Emulator":
          case "Channel31.Emulator":
          case "Channel32.Emulator":
          case "Channel33.Emulator":
          case "Channel34.Emulator":
          case "Channel35.Emulator":
          case "Channel10.EmulatorHpdPixelMode":
          case "Channel11.EmulatorHpdPixelMode":
          case "Channel12.EmulatorHpdPixelMode":
          case "Channel13.EmulatorHpdPixelMode":
          case "Channel14.EmulatorHpdPixelMode":
          case "Channel15.EmulatorHpdPixelMode":
          case "Channel16.EmulatorHpdPixelMode":
          case "Channel17.EmulatorHpdPixelMode":
          case "Channel18.EmulatorHpdPixelMode":
          case "Channel19.EmulatorHpdPixelMode":
          case "Channel20.EmulatorHpdPixelMode":
          case "Channel21.EmulatorHpdPixelMode":
          case "Channel22.EmulatorHpdPixelMode":
          case "Channel23.EmulatorHpdPixelMode":
          case "Channel24.EmulatorHpdPixelMode":
          case "Channel25.EmulatorHpdPixelMode":
          case "Channel26.EmulatorHpdPixelMode":
          case "Channel27.EmulatorHpdPixelMode":
          case "Channel28.EmulatorHpdPixelMode":
          case "Channel29.EmulatorHpdPixelMode":
          case "Channel30.EmulatorHpdPixelMode":
          case "Channel31.EmulatorHpdPixelMode":
          case "Channel32.EmulatorHpdPixelMode":
          case "Channel33.EmulatorHpdPixelMode":
          case "Channel34.EmulatorHpdPixelMode":
          case "Channel35.EmulatorHpdPixelMode":
          case "Channel10.EmulatorEvent":
          case "Channel11.EmulatorEvent":
          case "Channel12.EmulatorEvent":
          case "Channel13.EmulatorEvent":
          case "Channel14.EmulatorEvent":
          case "Channel15.EmulatorEvent":
          case "Channel16.EmulatorEvent":
          case "Channel17.EmulatorEvent":
          case "Channel18.EmulatorEvent":
          case "Channel19.EmulatorEvent":
          case "Channel20.EmulatorEvent":
          case "Channel21.EmulatorEvent":
          case "Channel22.EmulatorEvent":
          case "Channel23.EmulatorEvent":
          case "Channel24.EmulatorEvent":
          case "Channel25.EmulatorEvent":
          case "Channel26.EmulatorEvent":
          case "Channel27.EmulatorEvent":
          case "Channel28.EmulatorEvent":
          case "Channel29.EmulatorEvent":
          case "Channel30.EmulatorEvent":
          case "Channel31.EmulatorEvent":
          case "Channel32.EmulatorEvent":
          case "Channel33.EmulatorEvent":
          case "Channel34.EmulatorEvent":
          case "Channel35.EmulatorEvent":
            outData[regName] = _fwUkl1_convertByteToAbsRegDescrition(regName, dcData, forStatus, exceptionInfo);
            errMsg = "_fwUkl1_parseRegistersForGet(): Failed to parse the data from " + hwRegName + " for " + regName + ".";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;

          case "ChannelAll.Disable":
          case "ChannelAll.HpdPixelMode":
          case "ChannelAll.ZeroSuppression":
          case "ChannelAll.EmptyHpdSuppression":
          case "ChannelAll.Emulator":
          case "ChannelAll.EmulatorEvent":
          case "ChannelAll.EmulatorHpdPixelMode":
          case "ChannelAll.PixelMasks":
            // In this case we pass all the raw data. The all registers require information
            // from multiple hardware registers in most cases and need to look itself for them all.
            outData[regName] = _fwUkl1_convertByteToAbsRegChannelAll(regName, inData, forStatus, exceptionInfo);
            errMsg = "_fwUkl1_parseRegistersForGet(): Failed to parse the data from " + hwRegName + " for " + regName + ".";
            fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", errMsg, "2", exInfoSize);
            break;
            
  //         case "":
  //           break;
          
          // Being forgetful...
          default:
            exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_parseRegistersForGet(): Did not recognise the register " + regName + ", this register has not been parsed and will not be written/saved.", "2");
            break;
        }// switch(regName)
      }// else there is data.
    }// else there is no key!
  }// for regNum
  // Done.
  return;
}// fwUkl1_parseConfigRegistersForGet()
  
/*!
 * This takes an outData mapping from the one of the register parsing functions and converts them to two dyn_dyn_chars
 * for the register data and the mask.
 *
 * \param  outData The mapping from the parse register function.
 * \param  regNames Names of the hardware registers that are to be written to.
 * \param  regData The register data.
 * \param  masks The mask.
 * \param  writeFunctions This is a list that contains the functions that the register data and mask in the
 *           corresponding element should be written with. The writeFunctions dyn_string is sorted such
 *           that matching write functions are gaurantied to be consequtive.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The returned regData and masks are sorted with respect to the writeFunction that is to be used.
 * The writeFunctions arrays is sorted such that all matching write functions are sequential and it
 * matches the corresponding regData and mask. This is to make it easier to separate the all the data
 * and masks that are written in the same way.
 */
void _fwUkl1_parseWriteMapping(const mapping &writeData, dyn_string &regNames, dyn_dyn_char &regData, dyn_dyn_char &masks, dyn_string &writeFunctions, dyn_string &exceptionInfo) {
  // Get all the keys.
  const dyn_string keys = mappingKeys(writeData);
  const int numKeys = dynlen(keys);
  string key = "";
  for (int keyNum=1; keyNum<=numKeys; ++keyNum) {
    key = keys[keyNum];
    // Check we have the data.
    const int numDataEle = dynlen(writeData[key]);
    const int requiredEle = 3;
    if (requiredEle <= numDataEle) {
      regNames[keyNum]       = key;
      regData[keyNum]        = writeData[key][FWUKL1_MAP_REG_DATA];
      masks[keyNum]          = writeData[key][FWUKL1_MAP_XTR_DATA];
      writeFunctions[keyNum] = writeData[key][FWUKL1_MAP_XTR_DATA+1];
    }// if(1<numDataEle)
    else
      fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_parseOutMapping(): Not enough data was given for the hardware register " + key + ", this register will not be written. Require " + requiredEle + " elements given " + numDataEle + ".", "2");
  }// for keyNum
  
  // Now we need to sort the 4 arrays relative to the write function to use.
  // Hence this must be first in the array.
  dyn_mixed lists = makeDynMixed(writeFunctions, regNames, regData, masks);
  fwUkl1_sort(lists, exceptionInfo);
  // Now copy the sorted lists back again.
  writeFunctions = lists[1];
  regNames       = lists[2];
  regData        = lists[3];
  masks          = lists[4];
  
  // Done.
  return;
}// _fwUkl1_parseWriteMapping()

/*!
 * This retrieves the hardware register names that are required to read the given registers in the outData mapping.
 *
 * \param  absRegNames This is the list of registers that are to be read. It is for each of these that the associated
 *           hardware registers are retrieved.
 * \param  hwRegNames This is the returned list of hardware register names that are to be read. It will contain
 *           no duplicate entries.
 * \param  readFunctions This is the name of the function that should be used to read the hardware register.
 *           Its index in this array corresponds to the hardware register in the equivalent element in the
 *           hwRegNames array.
 * \param  standard Some of the registers require an alternative hardware register to that used by default.
 *           Passing TRUE here indicates that the standard hardware register should be used for all register,
 *           otherwise FALSE to retrieve the alternative. There is not always an alternative and in this
 *           case the standard is retrieved.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * The hardware registers are sorted with respect to the read functions this is required by the _fwUkl1_readSwitch
 * function.
 */
void _fwUkl1_parseReadMapping(const dyn_string &absRegNames, dyn_string &hwRegNames, dyn_string &readFunctions, bool standard, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // These must be clear to start otherwise it could muck up the counting.
  dynClear(hwRegNames);
  dynClear(readFunctions);
  const int numRegs = dynlen(absRegNames);
  for (int regNum=1; regNum<=numRegs; ++regNum) {
    const string hwRegName = fwUkl1_getHwRegName(absRegNames[regNum], standard, exceptionInfo);
    if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseReadMapping(): Failed to get the hardware register name for the register " + absRegNames[regNum] + ". This register will not be read.", "2", exInfoSize)) {
      // If we have the null register then we do not want to read from it.
      if (FWUKL1_NULL_HWREG == hwRegName)
        continue;
      // Now check if it is already in the array.
      if (1 > dynContains(hwRegNames, hwRegName)) {
        // Get the name of the read function.
        const string readFunc = fwUkl1_getReadFunction(absRegNames[regNum], exceptionInfo);
        if (!fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_parseReadMapping(): Failed to retrieve the function to use to read the hardware register for " + absRegNames[regNum] + ", this register cannot be read.", "2", exInfoSize)) {
          dynAppend(hwRegNames, hwRegName);
          dynAppend(readFunctions, readFunc);
        }// if got read function.
      }// if does not contain
    }// if not already add the hw reg
  }// for regNum

  // Now we need to sort the 4 arrays relative to the read function to use.
  // Hence this must be first in the array.
  dyn_mixed lists = makeDynMixed(readFunctions, hwRegNames);
  fwUkl1_sort(lists, exceptionInfo);
  // Now copy the sorted lists back again.
  readFunctions = lists[1];
  hwRegNames    = lists[2];
  
  // Done.
  return;
}// _fwUkl1_parseReadMapping()

/*!
 * Takes the given mapping element containing the data in the form for writing to the hardware register
 * and ORs the data and mask into the appropriate place in the dyn_char entries.
 *
 * \param  outDataElement The element from the mapping that corresponds to the hardware registers that the data
 *           should be added for. The data and mask elements are checked must contain at least 1 element.
 * \param  startBit First bit to write the data to in the bit field.
 * \param  sData The data that is to be written to the register.
 * \param  mask String containing the mask to be written to the register. It will appropriately aligned with the
 *           start of the data and padded to the appropriate length.
 * \param  fpga This field indicates if the data is to be written to an FPGA or not. They have a slightly different
 *           register structure that needs to be taken into account.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * Due to the way the function works we are limited to 32-bit data if the fpga is FALSE and only 16-bit if fpga is TRUE.
 */
void _fwUkl1_addWriteDataAndMask(dyn_mixed &outDataElement, int startBit, int data, int mask, bool fpga, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // Check that we have a sensible startBit. As we don't know how long the mask or data is then can't force the
  // start to be sensible.
  const int maxBit = (fpga ? 15:31);
  if (startBit >= maxBit) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_addWriteDataAndMask(): The startBit to write the data and mask to, " + startBit + ", is outside the allowed range, " + maxBit + ". Can't continue. Data and mask left unchanged.", "2");
    return;
  }// if(startBit>=maxBit);
  
  // We only want to write to the first 2 elements so we need at least that many. Not interested in the rest of the extra data.
  if (2 > dynlen(outDataElement)) {
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Not enough elements to write the data and mask to. Require at least two element given dyn_mixed contains only " + dynlen(outDataElement) + ". Inserting two elements at the beginning. Assuming 32-bit registers.", "2");
    outDataElement[FWUKL1_MAP_REG_DATA] = makeDynChar(0x00, 0x00, 0x00, 0x00);
    outDataElement[FWUKL1_MAP_XTR_DATA] = makeDynChar(0x00, 0x00, 0x00, 0x00);
  }// if(2>dynlen(outDataElement))
  // These will be used a lot.
  int datalen = dynlen(outDataElement[FWUKL1_MAP_REG_DATA]);
  int masklen = dynlen(outDataElement[FWUKL1_MAP_XTR_DATA]);
  // Need at least a byte to write to.
  if (1 > datalen) {
    // Either make it the same size as the mask or default to 32-bit.
    if (1 > masklen) {
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Write element contains no data. Assuming a 32-bit register and adding 4 bytes.", "2");
      outDataElement[FWUKL1_MAP_REG_DATA] = makeDynChar(0x00,0x00,0x00,0x00);
      datalen = dynlen(outDataElement[FWUKL1_MAP_REG_DATA]);
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Mask element contains no data. Assuming a 32-bit register and adding 4 bytes.", "2");
      outDataElement[FWUKL1_MAP_XTR_DATA] = makeDynChar(0x00,0x00,0x00,0x00);
      masklen = dynlen(outDataElement[FWUKL1_MAP_XTR_DATA]);
    }// if not mask data.
    else {
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Write element contains no data. Setting to the same size as the mask, " + masklen + ".", "2");
      int loopStart = datalen;
      int loopEnd   = masklen;
      for (int i=loopStart; i<=loopEnd; ++i)
        dynAppend(outDataElement[FWUKL1_MAP_REG_DATA], 0x00);
      datalen = dynlen(outDataElement[FWUKL1_MAP_REG_DATA]);
    }// else mask data.
  }// if no elements in write data.
  // Require to have the same number of mask elements as in the write register.
  // This will also catch the condition where there is no mask data but some write data.
  if (datalen != masklen) {
    // If the data is shorter than the mask then add elements to the data till it is the same length.
    if (datalen < masklen) {
      int loopStart = datalen;
      int loopEnd   = masklen;
      for (int i=loopStart; i<=loopEnd; ++i)
        dynAppend(outDataElement[FWUKL1_MAP_REG_DATA], 0x00);
      datalen = dynlen(outDataElement[FWUKL1_MAP_REG_DATA]);
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Appended " + (loopEnd-loopStart) + " elements to the register data such that it is the same length as the mask data (" + masklen + ").", "2");
    }// if(datalen<masklen)
    // If the mask is shorter than the data then add elements to the mask until it is the same length.
    else if (masklen < datalen) {
      int loopStart = datalen;
      int loopEnd   = masklen;
      for (int i=loopStart; i<=loopEnd; ++i)
        dynAppend(outDataElement[FWUKL1_MAP_XTR_DATA], 0x00);
      masklen = dynlen(outDataElement[FWUKL1_MAP_XTR_DATA]);
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_addWriteDataAndMask(): Appended " + (loopEnd-loopStart) + " elements to the mask data such that it is the same length as the register data (" + datalen + ").", "2");
    }// if(masklen<datalen)
  }// if(datalen!=masklen)
  
  // Now we have everything must be setup appropriately lets OR in the data and mask!
  // Arrange the data.
  data = data << startBit;
  // Arrange the mask.
  mask = mask << startBit;
  // Make sure neither over step their bounds.
  if (fpga)
    mask = mask & 0x0000ffff;
  // else is 32 the size of the reigster.
  data = data & mask;
  // Now convert it to a dyn_char, go via a hex string to ensure it is the correct size.
  string sData = fwCcpc_convertDecToHex(data);
  string sMask = fwCcpc_convertDecToHex(mask);
  // By padding the string out to a 32-bit register we will ensure that the FPGA registers are the correct lenght.
  fwUkl1_padString(sData, 4);
  fwUkl1_padString(sMask, 4);
  // Now convert to dyn_char.
  outDataElement[FWUKL1_MAP_REG_DATA] = fwCcpc_convertHexToByte(sData);
  outDataElement[FWUKL1_MAP_XTR_DATA] = fwCcpc_convertHexToByte(sMask);
  
  // Done.
  return;
}// _fwUkl1_addWriteDataAndMask()

/*!
 * Swaps two elements of a dyn structure. No operation is performed if they are the same or if one of the
 * indexes is out of bounds.
 *
 * \param  dyn The dyn structure containing the elements to be swapped.
 * \param  ele1 Index of one of the elements to swapped.
 * \param  ele2 Index of the second element to be swapped.
 * \return void.
 *
 * Not that the input elements are unsigned. If a negative number is given it will likely be
 * converted to an int be very large and in most cases out of bounds.
 */
void _fwUkl1_swapElements(dyn_anytype dyn, unsigned ele1, unsigned ele2) {
  const int len = dynlen(dyn);
  if ( (ele1 != ele2) && (ele1 <= len) && (ele2 <= len) ) {
    // Read ele1 and save it.
    anytype tmpValue  = dyn[ele1];
    // Write ele2 to ele1.
    dyn[ele1] = dyn[ele2];
    // Now write the saved ele1 to ele2.
    dyn[ele2] = tmpValue;
  }// if indexes not the same and inbounds.
  // Done.
  return;
}// _fwUkl1_swapElements()

// @}


// ==========
//   RESETS
// ==========

/** @defgroup SectionResets
 *    These are the resets that have been provided for the UKL1 board.
 *    Most of these resets are for expert use only and should be used with caution.
 *    The exception is fwUkl1_reload, which is used by the FSM for the reset
 *    command.
 */
// @{

/*!
 * Performs a reload of the FPGA firmware on the L1 board. It will reset all values back to their defaults.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_reload(const string &ukl1Name, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  _fwUkl1_toggleGpioLine(ukl1Name, FWCCPC_GPIO_LINE_6, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_reload(): Failed to reload the UKL1 firmware. There was a problem with the hardware access.", "2", exInfoSize);
  // Give the board 3 seconds to allow it to reload the firmware.
  delay(3);
  // Done.
  return;
}// fwUkl1_reload()

/*!
 * Performs a reset of the FPGAs on the L1 board. Effects still under discussion.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_reset(const string &ukl1Name, dyn_string& exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  _fwUkl1_toggleGpioLine(ukl1Name, FWCCPC_GPIO_LINE_5, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_reset(): Failed to reload the UKL1 firmware. There was a problem with the hardware access.", "2", exInfoSize);
  // Done.
  return;
}// fwUkl1_reset()

/*!
 * Performs a reset of the GBE card. If this fails it may prevent access to the GBE card.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void fwUkl1_resetGBE(const string &ukl1Name, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_resetGBE(): UKL1 name was empty, cannot proceed.", "2");
    return;
  }// if(""==ukl1Name)

  int exInfoSize = dynlen(exceptionInfo);
  int callStatus;
  // I am not certain if this GPIO line is enable for output as the others are by default, thus attempt to enable it.
  callStatus = fwCcpc_GPIOEnable(ukl1Name, FWCCPC_GPIO_LINE_4, FWCCPC_GPIO_OUTPUT);
  if (0 != callStatus) {
    // Construct the error message.
    string errMsg = "fwUkl1_resetGBE(): Failed to set the GPIO line 4 enabled for output during the reset, will continue to attempt a reset but it may fail. Encountered";
    if (callStatus > 0)
      errMsg += " a problem writing to the CCPC server, return code " + callStatus + ".";
    else if (callStatus < 0)
      errMsg += " a problem with PVSS.";
    else
      errMsg += " an unknown error.";
    // Now raise the exception.
    fwUkl1Exception_raise(exceptionInfo, "ERROR", errMsg, "1");
  }// Failed to enable the GPIO line.
  
  _fwUkl1_toggleGpioLine(ukl1Name, FWCCPC_GPIO_LINE_4, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_resetGBE(): Failed to reset the GBE card, hardware acess problems.", "2", exInfoSize);
  
  // Give the GBE card time to recover.
  delay(0,10);
  // Done.
  return;
}// fwUkl1_resetGBE()

/*!
 * Perform a partial reset of the UKL1 boards.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * Resets everything except the configuration registers e.g. logic, readout buffers, status registers etc..
 */
void fwUkl1_resetUkl1Partial(string ukl1Name, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_resetUkl1Partial(): UKL1 name was empty, cannot proceed.", "2");
    return;
  }// if(""==ukl1Name)
  
  // Used to track changes in the exception info array.
  int exInfoSize = dynlen(exceptionInfo);
  // The reset must first be asserted and then deasserted for the pointers to reset.
  // Assert the reset.
  const string fifoResetReg = ukl1Name + ".Resets.PartialL1";
  // data and mask are the same in this case.
  dyn_dyn_char dataAndMask;
  dataAndMask[1] = fwCcpc_convertHexToByte("00000020");
  _fwUkl1_write(fifoResetReg, dataAndMask, dataAndMask, TRUE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetPartialUkl1(): Failed to assert reset, attempting to deassert the reset anyway.", "2", exInfoSize);

  // Deassert the reset of the FIFO pointer.
  dyn_dyn_char clearData;
  clearData[1] = fwCcpc_convertHexToByte("00000000");
  // dataAndMask now just contains the mask as we want to write something different for the data.
  // The mask is the same as before however.
  _fwUkl1_write(fifoResetReg, clearData, dataAndMask, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetPartialUkl1(): Failed to de-assert status reset.", "2", exInfoSize);
  
  // Done.
  return;
}// fwUkl1_resetPartialUkl1()

/*!
 * Asserts and then deasserts a reset of the status buffer pointers. This must be done in order
 * to read any information from the Ingress status FIFOs.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 *
 * This will reset all the status buffer pointers simulataneously.
 */
void fwUkl1_resetIngressFpgaStatusBufPointers(string ukl1Name, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_resetIngressFpgaStatusBufPointers(): UKL1 name was empty, cannot proceed.", "2");
    return;
  }// if(""==ukl1Name)
  
  // Used to track changes in the exception info array.
  int exInfoSize = dynlen(exceptionInfo);
  // The reset must first be asserted and then deasserted for the pointers to reset.
  // Assert the reset of FIFO pointer.
  const string fifoResetReg = ukl1Name + ".Resets.IngressFpgaStatusBufferPointers";
  // data and mask are the same in this case.
  dyn_dyn_char dataAndMask;
  dataAndMask[1] = fwCcpc_convertHexToByte("0000000f");
  _fwUkl1_write(fifoResetReg, dataAndMask, dataAndMask, TRUE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetIngressFpgaStatusBufPointers(): Failed to assert status buffer pointer reset, status information will likely be invalid.", "2", exInfoSize);

  // Deassert the reset of the FIFO pointer.
  dyn_dyn_char clearData;
  clearData[1] = fwCcpc_convertHexToByte("00000000");
  // dataAndMask now just contains the mask as we want to write something different for the data.
  // The mask is the same as before however.
  _fwUkl1_write(fifoResetReg, clearData, dataAndMask, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetIngressFpgaStatusBufPointers(): Failed to de-assert status buffer pointer reset, status information will likely be invalid.", "2", exInfoSize);
  
  // Done.
  return;
}// fwUkl1_resetIngressFpgaStatusBufPointers()

/*!
 * Asserts and then deasserts a reset of the SPI3 RX read pointer.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This is for a special test feature and will almost never need to be used.
 */
void fwUkl1_resetSPI3RXReadPointer(string ukl1Name, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_resetSPI3RXReadPointer(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)
  
  // Used to track changes in the exception info array.
  int exInfoSize = dynlen(exceptionInfo);
  // The reset must first be asserted and then deasserted for the pointers to reset.
  // Assert the reset.
  const string fifoResetReg = ukl1Name + ".Resets.SPI3RxReadPointer";
  // data and mask are the same in this case.
  dyn_dyn_char dataAndMask;
  dataAndMask[1] = fwCcpc_convertHexToByte("00000010");
  _fwUkl1_write(fifoResetReg, dataAndMask, dataAndMask, TRUE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetSPI3RXReadPointer(): Failed to assert reset.", "2", exInfoSize);

  // Deassert the reset of the FIFO pointer.
  dyn_dyn_char clearData;
  clearData[1] = fwCcpc_convertHexToByte("00000000");
  // dataAndMask now just contains the mask as we want to write something different for the data.
  // The mask is the same as before however.
  _fwUkl1_write(fifoResetReg, clearData, dataAndMask, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetSPI3RXReadPointer(): Failed to de-assert status reset.", "2", exInfoSize);
  
  // Done.
  return;
}// fwUkl1_resetSPI3RXReadPointer()

/*!
 * Perform a reset of the DCMs.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Resets the Egress clock manager and generally will never need to be used except in special test cases.
 */
void fwUkl1_resetDCMs(string ukl1Name, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_resetDCMs(): UKL1 name was empty, cannot proceed.", "2");
    return;
  }// if(""==ukl1Name)
  
  // Used to track changes in the exception info array.
  int exInfoSize = dynlen(exceptionInfo);
  // The reset must first be asserted and then deasserted for the pointers to reset.
  // Assert the reset.
  const string fifoResetReg = ukl1Name + ".Resets.DCMs";
  // data and mask are the same in this case.
  dyn_dyn_char dataAndMask;
  dataAndMask[1] = fwCcpc_convertHexToByte("00000030");
  _fwUkl1_write(fifoResetReg, dataAndMask, dataAndMask, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetDCMs(): Failed to assert reset.", "2", exInfoSize);

  // Deassert the reset of the FIFO pointer.
  dyn_dyn_char clearData;
  clearData[1] = fwCcpc_convertHexToByte("00000000");
  // dataAndMask now just contains the mask as we want to write something different for the data.
  // The mask is the same as before however.
  _fwUkl1_write(fifoResetReg, clearData, dataAndMask, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetDCMs(): Failed to de-assert status reset.", "2", exInfoSize);
  
  // Done.
  return;
}// fwUkl1_resetDCMs()

/*!
 * Performs a reset of the GBE PHY chips.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Required during the GBE initialisation procedure.
 */
void fwUkl1_resetPhyChips(const string &ukl1Name, dyn_string &exceptionInfo) {
  // Keep track of errors.
  int exInfoSize = dynlen(exceptionInfo);
  // Loop over the PHY chips.
  for (int phy = 0; phy < 4; ++phy) {
    // Write the reset command to the PHY chip, don't bother to verify the writes.
    _fwUkl1_writeGbeMdioRegister(ukl1Name, phy, 0x00, 0x8140, FALSE, exceptionInfo);
    fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_resetPhyChips(): Failed to perform a reset of PHY chip number " + phy + ". GBE card may not function correctly if this was part of the initialisation.", "2", exInfoSize);
  }// for phy
}// fwUkl1_resetPhyChip()

// ==============
//    INTERNAL
// ==============

/*!
 * This function toggles the given GPIO function from low to high.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  gpioLine The line to toggle.
 * \param  exceptionInfo Error information. No change in size if no exception information is generated.
 * \return void.
 */
void _fwUkl1_toggleGpioLine(const string &ukl1Name, int gpioLine, dyn_string &exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_reload(): UKL1 name was empty, cannot proceed.", "2");
    return;
  }// if(""==ukl1Name)
  
  int exInfoSize = dynlen(exceptionInfo);
  int callStatus;
  // The reload is performed by setting the GPIO line 6 low for a few 10s of milliseconds and then setting it back to high.
  // First set it low.
  callStatus = fwCcpc_GPIOSet(ukl1Name, gpioLine, FWCCPC_GPIO_LOW);
  // Check for errors.
  if (0 == callStatus) {
    // We have driven it low, wait a few milli-seconds for the reload to take affect.
    delay(0,20);
    // Only worth trying to do this if it succeeded.
    callStatus = fwCcpc_GPIOSet(ukl1Name, gpioLine, FWCCPC_GPIO_HIGH);
    // Check for errors.
    if (0 != callStatus) {
      // Construct the error message.
      string errMsg = "_fwUkl1_toggleGpioLine(): Failed to set the GPIO line " + gpioLine + " high during a reload, unable to desassert the reload. Encountered";
      if (callStatus > 0)
        errMsg += " a problem writing to the CCPC server, return code " + callStatus + ".";
      else if (callStatus < 0)
        errMsg += " a problem with PVSS.";
      else
        errMsg += " an unknown error.";
      // Now raise the exception.
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", errMsg, "2");
    }// if set the GPIO line high.
  }// if set GPIO line low
  else {
    // Construct the error message.
    string errMsg = "_fwUkl1_toggleGpioLine(): Failed to set the GPIO line " + gpioLine + " low during a reload, unable to reload. Encountered";
    if (callStatus > 0)
      errMsg += " a problem writing to the CCPC server, return code " + callStatus + ".";
    else if (callStatus < 0)
      errMsg += " a problem with PVSS.";
    else
      errMsg += " an unknown error.";
    // Now raise the exception.
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", errMsg, "2");
  }// else could not set the GPIO line low.
  // Done.
  return;
}// _fwUkl1_toggleGpioLine()

// @}


// =========================
//  STATE MACHINE FUNCTIONS
// =========================

/** @defgroup SectionState FSM interface
 *    A group of functions that can be used to perform state machine actions
 *    for a UKL1 board. The following actions are defined:
 *    \li Configure - Loads of the configuration registers on the UKL1 board with the settings defined in the database.
 *          It will leave the board in the READY state. It is slow transition.
 *    \li Start - Moves the board into the RUNNING state. Enables data to flow through the UKL1 board. Fast transition.
 *    \li Stop  - Moves the board into the READY state. Prevents data flow through the UKL1 board. It will ignore TFC
 *          commands and cannot output over the GBE card.
 *    \li Reset  - Performs a reset of the UKL1. Currently the actions of this are ill defined.
 *    \li Reload - Reloads the FPGA (Ingress/FE and Egress/BE) firmware from the onboard memory.
 *    \li Recover - Currently does nothing.
 *
 *    None of the defined functions change the FSM state, this is left to the User to ensure the FSM is in the appropriate
 *    state after the function has been called.
 *  @{
 */

/*!
 * Gets the state of the current UKL1.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \return string State that the UKL1 board is currently in, empty string if the board is not present
 *           and hence there is no associated state.
 */
string fwUkl1_getUkl1State(const string& ukl1Name) {
  // Holds the state to be returned.
  string sState = "";
  // Check the DPE exists.
  const string dpeName = ukl1Name + ".status";
  if ( dpExists(dpeName) )
    dpGet(dpeName, sState);
  // No else as we have initialised ourselves to the appropriate case for DPE not existing.
  return sState;
}// fwUkl1_getUkl1State()

/*!
 * Performs a transition from one state to another. Takes the parameters that are passed into the equivalent function
 * in the FSM, as well as an exceptionInfo parameter.
 *
 * \param  domain FSM domain that the device is in e.g. RICH2_L1, the RICH2 control unit.
 * \param  device Name of the device that the command is to be performed for.
 * \param  command Name of the action that is to be performed.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_doCommand(const string &domain, const string &device, const string &command, dyn_string &exceptionInfo) { 
  int exInfoSize = dynlen(exceptionInfo);
  
  // Note that we rely on Stop coming after configure in the switch statement list.
  // Otherwise any order can be used.
  switch(command) {
    case "Configure": {
      // Inform the world that we are configuring.
      dpSet(device+".status","CONFIGURING");
    
      // Holds the name of the recipe that is to be loaded.
      string recipeName = "";
      // Determine what the parameter was given and hence the recipe name.
      string runType;
      fwDU_getCommandParameter(domain, device, "RUN_TYPE", runType);
      if ( "" == runType )
        fwCU_getParameter(domain, "RUN_TYPE", runType);
      recipeName += runType;
      recipeName += "/";
      recipeName += command;
    
      // Determine the partition ID. Will retrieve a decimal number stored in a string.
      string sPartId;
      fwDU_getCommandParameter(domain, device, "PART_ID", sPartId);
      if ( "" == sPartId )
        fwCU_getParameter(domain, "PART_ID", sPartId);
      // Convert the string to an int so it can be made into a hex number.
      sPartId = fwCcpc_convertDecToHex( (int)sPartId );
      // Determine the MTU from the run info.
      string runInfo;
      fwDU_getCommandParameter(domain, device, "RUN_INFO_DP", runInfo);
      int mtu = fwUkl1_getMtuFromRunInfo(runInfo, exceptionInfo);
      fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_doCommand(): Failed to retrieve the ethernet MTU size from the given run info data point. Taking value from the recipe.", "2", exInfoSize);
      
      // Configure the device.
      fwUkl1_configure(device, recipeName, domain, sPartId, mtu, exceptionInfo);
      // Once configured we want to stop. Stop must alway be below configure in if list! Don't use `if else'!
      if ( fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the configure action for " + device + ", transition to READY state failed.", "1", exInfoSize) ) {
        dpSet(device+".status","ERROR");
        break;
      }// if an error occured don't transition to stop!
      // Note that we only break if there is an error, if all is well continue and perform a stop.
    }// case "Configure"
	   
    case "Stop": {
      // Start and stop are configured from hardcoded recipe names. The FSM provides no choice in the matter.
      fwUkl1_stop(device, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the stop action for " + device + ", transition to READY state failed.", "1", exInfoSize) )
        dpSet(device+".status","READY");
      else
        dpSet(device+".status","ERROR");
      // Done.
      break;
    }// case "Stop"
  
    case "Start": {
      // Start and stop are configured from hardcoded recipe names. The FSM provides no choice in the matter.
      fwUkl1_start(device, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the start action for " + device + ", transition to RUNNING state failed.", "1", exInfoSize) )
        dpSet(device+".status","RUNNING");
      else
        dpSet(device+".status","ERROR");
      // Done.
      break;
    }// case "Start"
  
    case "Reload": {
      // Reloads the firmware for the FPGAs.
      fwUkl1_reload(device, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the reload action for " + device + ", transition to NOT_READY state failed.", "1", exInfoSize) )
        dpSet(device+".status","NOT_READY");
      else
        dpSet(device+".status","ERROR");
      // Done.
      break;
    }// case "Reload"

    case "Reset": {
      // As part of the LHCb FSM reset policy it is more sensible to do a reload here rather than a reset.
      fwUkl1_reload(device, exceptionInfo);
      // fwUkl1_reset(device, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the reset action for " + device + ", transition to NOT_READY state failed.", "1", exInfoSize) )
        dpSet(device+".status","NOT_READY");
      else
        dpSet(device+".status","ERROR");
      // Done.
      break;
    }// case "Reset"

    case "Recover": {
      // A recover should do the same as a reset, but only operate on those UKL1s that are in the error state.
      fwUkl1_reload(device, exceptionInfo);
      if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "fwUkl1_doCommand(): Failed to perform the recover action for " + device + ", transition to NOT_READY state failed.", "1", exInfoSize) )
        dpSet(device+".status","NOT_READY");
      else
        dpSet(device+".status","ERROR");
      // Done.
      break;
    }// case "Reconver"
  
    case "FAKE_READY": {
      // Just set to ready.
      dpSet(device+".status", "READY");
      // Done.
      break;
    }// case "FAKE_READY"
      
    default:{
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "fwUkl1_doCommand(): Unrecognised command received - " + command + " for " + device + ", no action taken.", "2");
      break;
    }// default
  }// switch (command)
  // Done.
  return;
}// fwUkl1_doCommand()

/*!
 * Loads a specific recipe to a UKL1 board from the configuration database.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  recipe Name of the recipe that is to be applied to the UKL1 board.
 * \param  domain Name of the FSM domain that the action was triggered from.
 *           If specified the recipe loaded by the "Configurator", if it exists, will be used.
 * \param  partitionId ID of the partition that this UKL1 board is within. Should be a hex string.
 *           If an empty string is passed the value stored in the recipe will be used.
 * \param  ethMtu Size of the ethernet maximum transmission unit that is to be used. If set to -1 then the value in the recipe
 *           will be used, otherwise it is overwritten.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 board to a configured state, it is typically a long operation and
 * it is recommended that a fwUkl1_stop() is performed once completed. It will not update the
 * state of the UKL1, this must be done by the User.
 */
void fwUkl1_configure(const string &ukl1Name, string recipe, const string &domain, string partitionId, int ethMtu, dyn_string& exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)
  // Holds the registers that can be set with this recipe.
  dyn_string regs;
  // Holds the recipe data.
  dyn_dyn_char data;
  // Keeps track of the number of errors that occured.
  int exInfoSize = dynlen(exceptionInfo);

  // When performing the configure we will always configure as much as possible and indicate any errors via
  // exception info. Boards may still be mostly usable.
  
  // See if the UKL1 name contains the system name. If it does, get it and then remove the system name.
  // We must pass the UKL1 name with the system name present otherwise it will appear not to be in the list
  // of hardware in the recipes. This causes lots and lots of difficulties!!!
  string sysName     = fwUkl1_removeDeviceName(ukl1Name);
  string sysUkl1Name = ukl1Name;
  // If it doesn't just guess.
  if ( "" == sysName ) {
    sysName     = fwUkl1_guessSystemName( fwUkl1_getFSMCUName() );
    sysUkl1Name = sysName + ukl1Name;
  }// if(""==sysName)
  
  // Convert the recipe name that we were given into a FSMConfDB compatible name.
  // don't! getRecipe does that all for you!!!
  // recipe = _fwHw_getRecipeName(ukl1Name, recipe, domain);
  
  // For the moment we will still configure the hardcoded settings of the GBE from value defined settings and not DB.
  // Probably will never use the DB.
  // This performs some configuration routines for the GBE, setting the GBE into its running mode.
  _fwUkl1_configureHardcodedSettings(sysUkl1Name, exceptionInfo);
  // Check for errors.
  if ( (dynlen(exceptionInfo) > exInfoSize) || (-1 == exInfoSize) ) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to initialise GBE or TTCrx.", "1");
    exInfoSize = dynlen(exceptionInfo);
  }// if _fwUkl1_configureHardcodedSettings() failed
  
  // The source IP and MAC addresses must be configured dynamically from the serial number.
  _fwUkl1_configureSourceAddresses(sysUkl1Name, exceptionInfo);
  if ( (dynlen(exceptionInfo) > exInfoSize) && (-1 != exInfoSize) ) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to initialise some/all of the source IP and MAC addresses, and L1 ID.", "1");
    exInfoSize = dynlen(exceptionInfo);
  }// if _fwUkl1_configureSourceAddresses()
  
  // Overwrite the partitionID setting by here if we are given one.
  // Must be done before we get the recipe! Do need to switch the recipe name for the cache one here as setEgressFpgaConfigRegisters()
  // won't do it for us. No domains are passed...
  dyn_string recipeNameAndType = makeDynString(recipe, FWUKL1_CONFIGURE_TYPE);
  dyn_string regs;
  dyn_string data;
  dyn_dyn_char ddcData;
  dyn_dyn_char masks;
  if ( "" != partitionId ) {
    dynAppend(regs, "PartitionId");
    dynAppend(data, partitionId);
  }// if(""!=partitionId)
  if ( -1 != ethMtu ) {
    dynAppend(regs, "DataFragmentSize");
    dynAppend(data, ethMtu);
  }// if(-1!=ethMtu)
  // Save in the recipe but don't verify the write. Want to be speedy :)
  fwUkl1_parseEgressFpgaConfigRegistersForSet(sysUkl1Name, regs, data, ddcData, masks, exceptionInfo);
  if ( !fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_configure(): To parse the values to be set for the ethernet MTU and partition ID. These could not be set from the FSM and will be taken from the recipe instead.", "2", exInfoSize) ) {
    // Our save recipe function requires the masks to have already been appended to the data.
    const int numElements = dynlen(ddcData);
    for (int element = 1; element <= numElements; ++element) {
      dynAppend(ddcData[element], masks[element]);
    }// for element
    // Now save the recipe
    _fwUkl1_save(sysUkl1Name, recipe, FWUKL1_CONFIGURE_TYPE, regs, ddcData, domain, exceptionInfo);
    fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "fwUkl1_configure(): Failed to update the recipe for " + ukl1Name + " with the ethernet MTU and partition ID taken from the FSM. Using the values found in the recipe.", "1", exInfoSize);
  }// if parsed MTU and partitionID.
  
  // Now retrieve the recipe so we can write it to the hardware.
  dynClear(regs);
  dynClear(ddcData);
  // Get the recipe that we wish to apply, from the cache hence domain!
  if ( -1 != fwHw_getRecipe(sysUkl1Name, recipe, regs, ddcData, domain) ) {
    // Now we have the recipe we must remove the channel configuration from it for later application.
    // If successful this will modify the regs and data such that fwHw_applyRecipe() will only apply the Egress
    // and GBE settings. If it is not successful everything will be applied, but the channel configuration data
    // will not be set.
    _fwUkl1_configureIngressChannelSettings(sysUkl1Name, regs, ddcData, exceptionInfo);
    if ( (dynlen(exceptionInfo) > exInfoSize) && (-1 != exInfoSize) ) {
      fwUkl1Exception_raise(exceptionInfo, "WARNING", "fwUkl1_configure(): Failed to configure the UKL1 input channel specific settings from the recipe `" + recipe + "'.", "2");
      exInfoSize = dynlen(exceptionInfo);
    }// if _fwUkl1_configureSourceAddresses()
    
    // Apply the recipe with the Egress and GBE data.
    if ( -1 == fwHw_applyRecipe(sysUkl1Name, regs, ddcData) ) {
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to apply the recipe `" + recipe + "' for the Egress and GBE settings.", "1");
      exInfoSize = dynlen(exceptionInfo);
    }// if(-1==fwHw_applyRecipe())
  }// if(-1!=fwHw_getRecipe())
  else {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to get the recipe `" + recipe + "'. Cannot configure the recipe specific settings.", "1");
    exInfoSize = dynlen(exceptionInfo);
  }// else(-1!=fwHw_getRecipe())

  // Apply the pixel masks recipe by here when we have some.
  // Don't apply them in the channel configuration as we have to get them from a recipe.

  // Done.
  return;
}// fwUkl1_configure()

/*!
 * Start does only two things and that is to enable the TFC decoding on the UKL1 and enable the link from
 * the BE FPGA to the GBE card for data output.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 to the start state where it is ready to take data. It will not update the state
 * once complete this is upto the User.
 */
void fwUkl1_start(string ukl1Name, dyn_string &exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_start(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)
  // Holds the registers that can be set with this recipe.
  dyn_string regs;
  // Holds the recipe data.
  dyn_dyn_anytype data;
  
  // Get and apply the recipe.
  // We just hardcode the recipe, as we always use the same one. There is no domain argument given as it is
  // a user recipe and hence the FSMConfDB object will not mangle its name.
  const string recipe = "Start";
  if ( -1 == fwHw_getApplyRecipe(ukl1Name, recipe) ) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_start(): Failed to get and apply the recipe `" + recipe + "'.", "1");
  }// if(-1==fw_getApplyRecipe())

  // Done.
  return;
}// fwUkl1_start()

/*!
 * Stop does only two things and that is to disable the TFC decoding on the UKL1 and disable the link from
 * the BE FPGA to the GBE card for data output.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 to the start state where it will not be able to see triggers sent from the TFC and also
 * is not capable of outputting data. It will not update the state once complete this is upto the User.
 */
void fwUkl1_stop(string ukl1Name, dyn_string &exceptionInfo) {
  // Check we have a UKL1 name, this really upsets things if we don't.
  if ("" == ukl1Name) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_stop(): UKL1 name was empty, cannot proceed.", "1");
    return;
  }// if(""==ukl1Name)
  // Holds the registers that can be set with this recipe.
  dyn_string regs;
  // Holds the recipe data.
  dyn_dyn_anytype data;
  
  // Get and apply the recipe.
  // We just hardcode the recipe, as we always use the same one. There is no domain argument given as it is
  // a user recipe and hence the FSMConfDB object will not mangle its name.
  const string recipe = "Stop";
  if ( -1 == fwHw_getApplyRecipe(ukl1Name, recipe) ) {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_stop(): Failed to get and apply the recipe `" + recipe + "'.", "1");
  }// if(-1==fw_getApplyRecipe())

  // Done.
  return;
}// fwUkl1_stop()

/*!
 * Recovers the UKL1 board from an error state. This will return the UKL1 to an unconfigured state and the values in
 * the registers cannot be guarantied to be the same as before recovering. The User should update the FSM state appropriately.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * The interface to this function is unlikely to change, but it might.
 */
void fwUkl1_recover(string ukl1Name, dyn_string& exceptionInfo) {
  // Do something to recover from an error...
  // this is actually quite useful as it is, it allows us to return to NOT_READY without doing anything.
  return;
}// fwUkl1_recover()

/*!
 * Configures the last byte of the source IP address and the fifth byte of the source MAC address from the ID number in the
 * EPROM. It also sets the L1 ID to the least significant byte of the serial number.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_configureSourceAddresses(const string& ukl1Name, dyn_string& exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // Get the logical ID from the board.
  int logicalId = fwUkl1_getLogicalId(ukl1Name, exceptionInfo);
  if ( -1 == logicalId ) {
    // This is going likely to be the prototype board, if the board genuinely doesn't exist then we will encounter some errors.
    // The prototype board can have a logical ID of 40, it is high to indicate what happened.
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_configureSourceAddresses(): Configuring board without a logical ID, using 40.", "1");
  }// if(-1==logicalId)

  // This prefixes the register names.
  const string prefix = ukl1Name + ".EgressFpga.";
  // These are required when writing to the board registers.
  dyn_string regNames;
  dyn_dyn_char data;
  dyn_dyn_char masks;
  
  // Convert the logical number to a string. For the IP this is fine, as it is a decimal number.
  // For the MAC address this is also fine as we wish the MAC address hex numbers to appear visually as the same
  // as the IP address i.e. if the logical number is 10, both the IP and MAC address will appear with 10
  // (192.168.14.10 and 00:BB:CC:DD:10:00) rather than a normal display (192.168.14.10 and 00:BB:CC:DD:0A:00).
  // This was requested by Niko and the Online group.
  const string boardNum = logicalId;
  // Construct the MAC address. Pad the first 4 bytes with 0s and the last byte also.
  // It must have extra 16-bit words added in order to convert the data from 16- to 32-bit word for writing to the registers.
  regNames[1] = prefix + "MacSourceAddress";
  data[1] = fwCcpc_convertHexToByte("0000" + boardNum + "000000000000000000");
  // Need a mask for a 96-bit word write, 2 32-bit words.
  // The last two 32-bit words (actually last two bytes but 16- to 32-bit funniness comes into play)
  // are reserved. The first of these bytes we must configure, the other must be set to zero and the UKL1 firmware will deal with it.
  masks[1] = makeDynChar(0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00);
  // Now construct the IP address. Need to pad the first 3 bytes with 0s, plus the extra padding to convert from 16- to 32-bit
  // words for the register writes.
  regNames[2] = prefix + "IpSourceAddress";
  data[2] = fwCcpc_convertHexToByte("000000" + boardNum + "00000000");
  // Need a 48-bit mask for this.
  // The last 8-bits are reserved and not User configurable. These are written what we now write.
  masks[2] = makeDynChar(0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00);
  // Now set the serial number to the L1 ID.
  string serialNumber = fwUkl1_getUkl1SerialNumber(ukl1Name, exceptionInfo);
  if ( "Fail" == serialNumber ) {
    // This is going likely to be the prototype board, if the board genuinely doesn't exist then we will encounter some errors.
    // The prototype board can have a serial number of 40, it is high to indicate what happened.
    serialNumber = "60110040";
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_configureSourceAddresses(): Configuring board without a serial number, using 0x60110040.", "1");
  }// if("Fail"==serialNumber)
  
  // Now get the last two characters (byte) from this number for the L1 ID.
  regNames[3] = prefix + "L1Id";
  string l1Id = substr(serialNumber, strlen(serialNumber)-2, 2);
  fwUkl1_padString(l1Id, 4);
  data[3] = fwCcpc_convertHexToByte(l1Id);
  // Need an 8-bit mask for this.
  masks[3] = makeDynChar(0x00, 0x00, 0x00, 0xff);

  // Now write this data. Have to bypass the BE FPGA write as this can't set the bytes we want to.
  // Use the _fwUkl1_write instead as this takes the raw byte data and can optionally verify the writes.
  // Don't verify writes until the update to the CCPC library fixes the problem where writeRead returns the previous written value.
  _fwUkl1_write(regNames, data, masks, FALSE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_configureSourceAddresses(): Failed to configure the source IP and MAC addresses, and the L1 ID. Events may not be routed correctly.", "1", exInfoSize);
  
  // Done.
  return;
}// _fwUkl1_configureSourceAddresses()

/*!
 * Configures all the settings on the GBE that are never changed by the User and it calls a function directly on the ccserv to
 * perform this initialisation. It also setups the single TTCrx register that is hardcoded to a UKL1 specific setting.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_configureHardcodedSettings(const string& ukl1Name, dyn_string& exceptionInfo) {
  // Used to track the size of the exception info array to see if new exception have been added.
  int exInfoSize = dynlen(exceptionInfo);

  // These are used just to make it a bit clearer in the function calls what we are doing.
  const bool writeRecipe = FALSE;
  dyn_string recipeNameAndType = makeDynString();
  const bool verifyWrite = FALSE;
  
  // Contains the registers and data that are to be written.
  mapping data;

  // Set up the TTCrx control register as desired.
  data["TtcrxControl"] = "a1";
  // Add the MDIO enable here also.
  fwUkl1_write(ukl1Name, data, TRUE, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_configureHardcodedSettings(): Some of the hardcoded settings have not been set correctly.", "2", exInfoSize);

  // Reset the GBE card via the GPIO line.
  fwUkl1_resetGBE(ukl1Name, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_configureHardcodedSettings(): Failed to reset the GBE card.", "2", exInfoSize);
    
  // Reset the PHY chips first.
  fwUkl1_resetPhyChips(ukl1Name, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_configureHardcodedSettings(): Failed to reset the GBE PHY chips, they may not function correct.", "2", exInfoSize);
  
  // Now ask the ccserv to perform the GBE setup routine. Does lots of initialisation that has been hardcoded into the ccserv.
  if (_fwCcpc_commonOperation(ukl1Name, "231", true) != "0")
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHardcodedSettings(): Failed to initialise the GBE it is very unlikely that it will function correctly. There was a problem sending the command to the ccserv.", "2");
  
  // Done.
  return;
}// _fwUkl1_configureHardcodedSettings()

/*!
 * This writes the Ingress input channel configuration settings to the UKL1s.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  regs This should contain all the registers that are found in a UKL1_Configure type recipe.
 *           The function will separate out the Channel settings and the other settings returning, by reference,
 *           the other settings that this function does not write. These can then be applied by fwHw_applyRecipe().
 * \param  data This should contain all the data that is retrieved from a recipe and the function will, as for
 *         regs, separate out the data it wants, apply it, and return the remaining data by reference for use with
 *         fwHw_applyRecipe().
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_configureIngressChannelSettings(const string& ukl1Name, dyn_string& regs, dyn_dyn_char& data, dyn_string& exceptionInfo) {
  // Keep track of the number of exceptions that have been generated.
  int exInfoSize = dynlen(exceptionInfo);
  // Determine where in the full register list are the registers that are set for channels.
  dyn_bool chanRegPos = patternMatch("*.Channel?.*", regs);
  if ( 0 < dynlen(chanRegPos) ) {
    // Holds the names of the registers that we wish to apply and the corresponding data.
    dyn_string channelRegs;
    dyn_dyn_char channelData;
    // Note where we put the register and data into the arrays.
    int index = 0;
    // Now loop through the returned bool array and determine where the channel data is.
    for (int reg = 1; reg <= dynlen(chanRegPos); ++reg) {
      if (chanRegPos[reg]) {
        ++index;
        // Copy the desired register to the channel register names array and then remove the register.
        channelRegs[index] = regs[reg];
        dynRemove(regs, reg);
        dynRemove(chanRegPos, reg);
        // also the data
        channelData[index]  = data[reg];
        dynRemove(data, reg);
        // Now decrement the reg number.
        --reg;
      }// if(chanRegPos[reg])
      // No need for an else as we want to leave the other register untouched.
    }// for reg
    // We can just apply the settings to the appropriate channel.
    const int numChanRegs = dynlen(channelRegs);
    // Check to see if we matched against anything.
    if ( 0 == numChanRegs )
      exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_configureIngressChannelSettings(): Failed to find any configuration settings for the Ingress FPGA channels. These have not been configured.", "2");
    
    // so we know who to configure.
    unsigned channel = -1;
    unsigned inFpga  = -1;
    for (int chanReg = 1; chanReg <= numChanRegs; ++chanReg) {
      // Get the channel and Ingress number.
      // This is the pattern that sscanf must look for. The system name can cause a problem as generally
      // it is not in the ukl1Name variable but it is in the recipe. The format thus is different and everything
      // becomes bad...
      const string format = fwUkl1_removeSystemName(ukl1Name) + ".IngressFpga%u.Channel%u.ConfigurationFifo";
      if ( 2 == sscanf(fwUkl1_removeSystemName(channelRegs[chanReg]), format, inFpga, channel) ) {
        // The FALSE is because we are not going to verify the writes. Want recipe application to be as fast as possible.
        _fwUkl1_setIngressFpgaChannelConfigRegisters(ukl1Name, channel, inFpga, channelData[chanReg], false, exceptionInfo);
        fwUkl1Exception_checkAndRaise(exceptionInfo, "WARNING", "_fwUkl1_configureIngressChannelSettings(): Failed to configure Ingress FPGA " + inFpga + " channel " + channel + ". This channel may not work correctly.", "2", exInfoSize);
      }// if(2!=sscanf())
      else
        exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_configureIngressChannelSettings(): Failed to determine the channel and Ingress FPGA to write the configuration data to for register `" + channelRegs[chanReg] + ". sscanf() failed to identify the channel (" + channel + ")and Ingress FPGA (" + inFpga + ") number in that string.", "2");
    }// for chanReg
  }// if(0<dynlen(chanRegPos)
  // In this case we couldn't find any channel configuration registers!
  // Warn people and then apply the full recipe anyway.
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "_fwUkl1_configureIngressChannelSettings(): Could not find any channel configuration settings in this recipe. No input channel configuration was done.", "2");
  // Done.
  return;
}// _fwUkl1_configureIngressChannelSettings()

/*!
 * This updates the partition ID in the recipe from the value given. Typically this value will have been retrieved
 * from the FSM and will have been passed down from the top level.
 *
 * \param  partitionId The partition ID to be set.
 * \param  recipe Name of the recipe to update the partition ID in. Generally this will be the recipe cached by the FSMConfDB.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void _fwUkl1_updatePartitionIdInRecipe(const string &ukl1Name, string partitionId, const string &recipe, const string &domain, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  const string sysName = "";
  
  // Create an array to hold the name of the recipe and the type of that recipe that we are writing to.
  // Here we should operate directly on the recipe loaded into the cache by the FSMConfDB object.
  dyn_string recipeNameAndType = makeDynString(recipe, "UKL1_Configure");
  // Name of the partition ID register that will be written to.
  dyn_string regNames;
  regNames[1] = ukl1Name + ".EgressFpga.PartitionId";
  // Now convert the partitionId to a byte array with mask.
  // Must be 8 characters big to be a valid ID, which is 4 bytes.
  fwUkl1_padString(partitionId, 4);
  // This needs to be written to two 32-bit registers with the two 16-bits words of the given partition ID
  // in each. Need to pad the beginning with four 0 characters and between the 1st and 2nd words.
  updatedPartitionId = "0000" + substr(partitionId, 4, 4) + "0000" + substr(partitionId, 0, 4);
  // Now convert it to a byte array for writing.
  dyn_dyn_char ddcDataAndMask;
  ddcDataAndMask[1] = fwCcpc_convertHexToByte(updatedPartitionId);
  // and append a 64-bit mask.
  dynAppend(ddcDataAndMask[1], makeDynChar(0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff));
  _fwUkl1_saveRecipe(sysName + ukl1Name, recipeNameAndType, regNames, ddcDataAndMask, FWHW_LBUS, domain, exceptionInfo);
  fwUkl1Exception_checkAndRaise(exceptionInfo, "ERROR", "_fwUkl1_updatePartitionIdInRecipe(): Failed to update the partition ID stored in the recipe with that given as a run time parameter (0x" + partitionId + ")in the FSM. Partition ID stored in the recipe will be written.", "1", exInfoSize);

  // Done.
  return;
}// _fwUkl1_updatePartitionIdInRecipe()

// @}


// ===================
//  SUPPORT FUNCTIONS
// ===================

/** @defgroup SectionSupport Support funtions
 *     These are miscellaneous functions that provide functionality that does not fit into any of the other categories.
 *  @{
 */

/*!
 * This takes a dyn_mixed where each element should be a list. The list in the first element is sorted
 * with the other lists also being sorted also being sorted in the same order.
 *
 * \param  lists This is the list of dyn types that are to be sorted. They are sorted relative to the
 *           first element of this dyn_mied type.
 * \param  exceptionInfo Error information. No change in size if there is no exception information generated.
 * \return void.
 *
 * The reference list must contain elements that can be compared and the other lists must
 * have at least as many elements. In the event there are too few elements in one list compared to the
 * reference then the function will fail with no sorting having taken place. In there event are too many
 * elements then those extra elements will not be sorted but the function will not fail nor warn of this.
 */
void fwUkl1_sort(dyn_mixed &lists, dyn_string &exceptionInfo) {
  // Just see who is first in the list then place ever function with the same next to it until there
  // are no more, by swapping with the appropriate element. When there are no more move onto the next type.
  // This is the list we are using as the reference for the sorting.
  const int ref = 1;
  // Need to now the number of lists to sort.
  const int numLists = dynlen(lists);
  // Need to know the size of the list we are sorting.
  const int numItems = dynlen(lists[ref]);
  // Start with the first element.
  anytype sortVal;
  // This is the element that we need to put the function in.
  int writeIndex = 1;
  while ( writeIndex < numItems ) {
    // Now we have a function to look for.
    sortVal = lists[ref][writeIndex];
    // Shuffle on one as we want to write matching functions in the next element.
    ++writeIndex;
    // Read all the remaining elements and swap in the occurances of this function.
    // Move on the write index and keep reading the rest of the elements looking for something to swap.
    for (int readIndex = writeIndex; readIndex <= numItems; ++readIndex) {
      if ( sortVal == lists[ref][readIndex] ) {
        // Swap the elements in each of the lists.
        for (int listNum = 1; listNum <= numLists; ++listNum)
          _fwUkl1_swapElements(lists[listNum], writeIndex, readIndex);
        // Now move on our write pointer.
        ++writeIndex;
      }// if (func===writeFunctions[readIndex])
    }// for readIndex
    // Once we have read through once we should move on one to where we will look for our next comparison function.
    ++writeIndex;
  }// while(writeIndex<numItems)
  // Done.
  return;
}// fwUkl1_sort()
  
/*!
 * Retrieves the ethernet MTU setting from the given run info data point. In the event of an error returns -1.
 *
 * \param  runInfo The run info data point to retrieve the value from. The DPE and necessary full stops will be added
 *           by the function.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return int The ethernet maximum transmission unit size as given by the FSM.
 */
int fwUkl1_getMtuFromRunInfo(string runInfo, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  // We will return -1 if everything is not successful.
  int mtu = -1;
  // Save the full path to the DPE.
  const string runInfoMTU = runInfo + ".SubDetectors.ethMTU";
  // Check it exists.
  if ( dpExists(runInfoMTU) ) {
    // Retrieve the value from the DP.
    dpGet(runInfoMTU, mtu);
    // Couldn't get the MTU.
    dyn_errClass error = getLastError();
    if ( 0 != dynlen(error) )
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getMtuFromRunInfo(): Could not retrieve the ethernet MTU size from the run info data point (" + runInfoMTU + "), it must exist at least. More details will be in the PVSS log.", "1");
  }// if(dpExists())
  else
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getMtuFromRunInfo(): Failed to retrieve the ethernet MTU from the run info datapoint as the DPE '" + runInfoMTU + "' did not exist.", "1");
  
  // Done.
  return mtu;
}// fwUkl1_getMtuFromRunInfo()

/*!
 * Based on the name of the CU it guess what the system number should be and then asks for the
 * name of that system.
 *
 * \param  cu Name of the control unit for this group of UKL1s.
 * \return string System name with an ending colon or empty string.
 *
 * Basically this function uses the following assumptions
 *   \li RICH1_L1 - This is RICH1 and has a system number of 136.
 *   \li RICH2_L1 - This is RICH2 and has a system number of 176.
 *   \li RICH3_L1 - This is any test system and has a system number of 176.
 *   \li Anything else will be assigned a system number of 0.
 * As the PVSS function getSystemName is used it will return the name with a ending colon.
 * If there is no connection to the guessed system number an empty string will be return, which
 * can be safely used for the system name and it will work provided everything is running on the
 * same local system.
 */
string fwUkl1_guessSystemName(string cu) {
  // Holds the guessed system ID.
  int sysId = 0;
  switch (cu) {
    case FWUKL1_RICH1_CU:
      sysId = FWUKL1_RICH1_SYS_NUM;
      break;
    case FWUKL1_RICH2_CU:
      sysId = FWUKL1_RICH2_SYS_NUM;
      break;
    case FWUKL1_RICH3_CU:
      sysId = FWUKL1_RICH3_SYS_NUM;
      break;
    default:
      // Do nothing it was initialised to 0.
      break;
  }// switch(cu)
  // Done.
  return getSystemName(sysId);
}// fwUkl1_guessSystemName()

/*!
 * This takes a recipe name and tries to determine if there is a domain name present in the recipe. If it is
 * found then it is removed and the recipe name is returned in a demanagled state.
 *
 * \param  recipeName Name of the recipe to retrieve the domain from and to demanagle.
 * \return string A string containing the domain name. It is an empty string if no domain name is found.
 *
 * The aim of this function is to take a recipe name that has been managled by the FSMConfDB object and demangle it.
 * The demangling is reversible as the domain name is returned by the function and this is with the demanagled
 * recipe name is all that is required to managle the name such that it is suitable to be used with the FSMConfDB
 * recipes.
 *
 * If the domain cannot be found then an empty string is returned and the recipeName left unchanged. If the recipe
 * name is believed to be valid then no error is raised. If the recipe name is believed to be invalid then
 * the exceptionInfo dyn_string will have information exception information added to it.
 */
string fwUkl1_removeDomainName(string &recipeName, dyn_string &exceptionInfo) {
  int exInfoSize = dynlen(exceptionInfo);
  string domain = "";
  // Split the recipe name using / as a delimiter.
  const dyn_string splitName = strsplit(recipeName, "/");
  const int splitLen = dynlen(splitName);
  if ( 4 == splitLen ) {
    // In this case return the domain and put the recipe name back in its original form.
    domain = splitName[2];
    for ( int i = 3; i <= splitLen; ++i ) {
      recipeName += splitName[i];
      if ( i != splitLen )
        recipeName += "/";
    }// for i
  }// if domain present.
  // No domain present and the recipe name appears to be invalid.
  else if ( 2 != dynlen(splitName) )
    exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getDomainNameFromRecipeName(): Failed to find the required number of / delimited sections, 4. Found only " + splitLen + " sections. Cannot retrieve the domain name and the recipe name appears invalid. Given recipe name was " + recipeName + ".", "2");
  // Ignore the case where we have two parts as this is a correctly formed none managled recipe name.

  // Done.
  return domain;
}// fwUkl1_getDomainNameFromRecipeName()

/*!
 * Strips the system name from the given string. Will remove only the system name. If no system name is present
 * then the effects are undefined, unless there are no semi-colons present in the string.
 *
 * \param  dpName String containing the DP name that the system name is to be removed from.
 * \return string dpName minus the beginning system name.
 */
string fwUkl1_removeSystemName(string dpName) {
  // Will hold the stripped DP name, will rely on it being initialised to NULL.
  string strippedDpName = "";

  // First break up the string into its semi-colon parts.
  const dyn_string semiColonDelimited = strsplit(dpName, ':');
  // Determine how many parts we have.
  const int numSec = dynlen(semiColonDelimited);
  if ( 1 == numSec ) {
    // No semi-colons found just return the original string.
    strippedDpName = dpName;
  }// if(1==numSec)
  else if ( 2 <= numSec ) {
    // Start at the second element, bypassing the system name and then append the other elements.
    for (int sec = 2; sec <= numSec; ++sec) {
      strippedDpName += semiColonDelimited[sec];
    }// for sec
  }// elseif(2<=numSec)
  else {
    // Must be no elements, just return the original string which is probably NULL :S
    strippedDpName = dpName;
  }// else

  // Done.
  return strippedDpName; 
}// fwUkl1_removeSystemName()

/*!
 * Strips the rest of the details from a DP name leaving only the system name and a colon.
 * If there is no system name present an empty string is returned.
 *
 * \param  dpName String containing the DP name that the system name is to selected from.
 * \return string The system name with a colon e.g. "R2DAQL1:", empty string is not present.
 */
string fwUkl1_removeDeviceName(string dpName) {
  // Will hold the stripped DP name, will rely on it being initialised to NULL.
  string strippedDpName = "";

  // First break up the string into its semi-colon parts.
  const dyn_string semiColonDelimited = strsplit(dpName, ':');
  // Determine how many parts we have.
  const int numSec = dynlen(semiColonDelimited);
  // Found at least two sections so the first should be the system name and then add the colon. Lost during the splitting.
  if ( 2 <= numSec )
    strippedDpName = semiColonDelimited[1] + ":";

  // Done.
  return strippedDpName; 
}// fwUkl1_removeDeviceName()

/*!
 * Takes the recipe name in the form RUN_MODE/Action and returns the recipe type the recipe should be save as.
 *
 * \param  recipeName The name of the recipe as it is save in the configuration data base.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string The recipe type that the recipe should be saved under. An empty string is returned
 *           if the recipe type cannot be determined, along with exception information detailing the problem.
 *
 * The recipe type required is determine from the Action part of the recipe name, if this is not present it
 * cannot be determined. The return value can be used directly in any recipe saving function. The only exception
 * to this are the start and stop recipes, which take the name "Start" and "Stop" respectively. These are user
 * recipes and their type is already known.
 * \n
 * This function works entirely from the recipe name and no gauranties are given that it is the actual type.
 * If the recipe is not named correctly or is currently saved under the wrong type this function cannot
 * know this.
 */
string fwUkl1_getRecipeType(const string& recipeName, dyn_string& exceptionInfo) {
  // Holds the recipe type.
  string recipeType = "";
  // Check to see if we have a name.
  if ( "" != recipeName ) {
    // Try and split the recipe name using a `/' as a delimiter.
    dyn_string tmpRunModeAndType = strsplit(recipeName, "/");
    // Check to see we have a name and type.
    const int numEle = dynlen(tmpRunModeAndType);
    if ( 1 < numEle ) {
      // It must have at least two elements for it to be of the appropriate form
      // (taking into account the fwFSMConfDB form with many /).
      // The type is always in the last element regardless of form.
      if ( "Configure" == tmpRunModeAndType[numEle] )
        recipeType = FWUKL1_CONFIGURE_TYPE;
      else if ( ("Start" == tmpRunModeAndType[numEle]) || ("Stop" == tmpRunModeAndType[numEle]) )
        recipeType = FWUKL1_START_TYPE;
      // Only need to mark the error with else as the element is already an empty string as promised.
      else
        fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getRecipeType(): Could not determine the recipe type from the recipe name `" + recipeName + "'. The action is believed to be `" + tmpRunModeAndType[2] + "'. No recognisable action present in the name (must be of the form RUN_MODE/Action).", "1");
    }// if(1<numEle)
    else if ( 1 == numEle ) {
      // This could be the user recipes Start and Stop, we know these types. Otherwise it is an error.
      if ( ("Start" == recipeName) || ("Stop" == recipeName) )
        recipeType = FWUKL1_START_TYPE;
      else
        fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getRecipeType(): Could not determine the recipe type from the recipe name `" + recipeName + "'. The recipe name does not appear to cotain an action and it is not recognised `" + recipeName+ "'. No recognisable action present in the name (must be of the form RUN_MODE/Action or called Start or Stop).", "1");
    }// elseif(1==numEle)
    else {
      // Can't do anything with either, mark the error and move on. The return values are already initialised to their error states.
      fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getRecipeType(): Could not determine the recipe type from the recipe name `" + recipeName + "'. No forward slash, /, present in the name to delimit the run mode and action (must be of the form RUN_MODE/Action).", "1");
    }// else
  }// if(""!=recipeName)
  else {
    fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_getRecipeType(): Could not identify recipe type as the recipe name string was empty.", "1");
  }// else(""!=recipeName)
  
  // Done.
  return recipeType;
}// fwUkl1_parseRecipeNameAndIdentifyType()

/*! --- !!!DEPRECIATED!!! ---
 * Updates the library constant FWUKL1_SYS_NUMBER with the value given to this function. If the value given
 * is equal to the current system number no changes are made.
 *
 * \param  sysNum The new number of the PVSS system.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This function updates the text in this library (const int FWUKL1_SYS_NUMBER = ?;), so it should not be called frequently
 * as it has the potential to corrupt the PVSS code in the file. If the system number set is not the current system number
 * of the PVSS project then all functions that require DP access will likely fail as they can no longer retrieve the system name.
 * The library will have to be reloaded after this function has been called in order for any changes to propagate.
 *
 * This function has been depreciated, it is just no longer needed. The system number can never be set anywhere
 * as the RICH1 and RICH2 libraries have the same name, thus the top level FSM will only ever load them once,
 * thus RICH1 and 2 will always use the same copy and see the same system number. Same goes for globals!
 * It is kept simply because a lot of effort went into writing it!!!!
 */
// void fwUkl1_setSystemNumber(int sysNum, dyn_string &exceptionInfo) {
  // Tracks the exception information.
//   int exInfoSize = dynlen(exceptionInfo);
  // This returns the full path and name of the library we are looking for.
//   const string libPath = getPath(LIBS_REL_PATH, "fwUkl1/fwUkl1.ctl");
  // Attempt to load the file into a string.
//   string sLib;
//   if ( fileToString(libPath, sLib) ) {
    // So we have the file as a rather large string.
    // We don't know the current value of the so ask sscans for it.
//     const string baseStr = "const int FWUKL1_SYS_NUMBER = ";
//     int curSysNum;
    // sscanf doesn't seem to like wildcards at the beginning of the string and hence instead we have to create
    // a substring from the point where the constant definition start onwards.
    // Don't worry about substr and strpos failing as sscanf will indicate an error if they fail.
//     if ( -1 != sscanf( substr(sLib, strpos(sLib, baseStr)), baseStr + "%d;*", curSysNum) ) {
      // There is no need to continue as the system number requested is the same as that to be set.
      // Note there are no file handles to clean up at this point so we can just return.
//       if (sysNum == curSysNum)
//         return;
      // Need to find the occurance of the constant
//       const string searchStr    = "const int FWUKL1_SYS_NUMBER = " + curSysNum + ";";
      // and replace it with this.
//       const string replacement  = "const int FWUKL1_SYS_NUMBER = " + sysNum + ";";
      // Perform the search and check the result.
//       int numReplacements = strreplace(sLib, searchStr, replacement);
//       if ( 1 == numReplacements ) {
        // Now try and write the new version of the library to file.
//         const string tmpPath = libPath + ".tmp";
//         file tmpFile = fopen(tmpPath, "w");
//         if ( 0 == ferror(tmpFile) ) {
          // Now we have opened a temporary file we should write the updated library to it.
//           if ( 0 == fputs(sLib, tmpFile) ) {
            // Flush the file and close it.
//             fflush(tmpFile);
//             fclose(tmpFile);
            // Now copy the temporary file across.
//             if ( !copyFile(tmpPath, libPath) )
//               exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Failed to update the library file fwUkl1.ctl with the new system number. Over write \'" + libPath + "\' with manually \'" + tmpPath + "\'.", "1");
//           }// if(0==fputs())
//           else
//             exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Could not write to the temporary file, aborting update of the system number.", "1");
          // Try and delete the temporary file.
//           if ( -1 == remove(tmpPath) )
//             exInfoSize = fwUkl1Exception_raise(exceptionInfo, "WARNING", "fwUkl1_setSystemNumber(): Could not delete the temporary file \'" + tmpPath + "\'. Please delete it manually.","2");
//         }// if(0==ferror())
//         else
//           exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Could not open a temporary file to save the updated library to. Failed to update the system number.", "1");
//       }// if 1 replacement
//       else if ( 1 < numReplacements )
//         exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Replace more than one entry of the system number in the library. It is likely that the library has been corrupted hence the update has not been saved. Pleasew updated the system number manually.", "1");
//       else
//         exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Could not find current entry of the system number in the file and hence could not update it.", "1");
//     }// if(-1!=sscanf())
//     else
//       exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Failed to identify the current system number from the library. Cannot update the number as we could not find it.", "1");
//   }// if(fileToString())
//   else
//     exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_setSystemNumber(): Failed to open the library \'" + libPath + "\' into a string to update the system number. This must be done manually.", "1");
//   
  // Done.
//   return;
// }// fwUkl1_setSystemNumber()

/*! --- !!!DEPRECIATED!!! ---
 * Starts a PVSS manager that is defined by the inputs manager and options. It is based on pmonStartManager(), which is provided
 * by PVSS but does not appear to work from within a PVSS00ctrl manager. The function has been modified to allow this.
 *
 * \param  manager The name of the PVSS manager to be started e.g. PVSS00ui.
 * \param  options The options that are passed to that manager in the PVSS console, sometime refered to as the command.
 * \param  showErrDialog If true PVSS will produce an error dialog in the event that the attempt to start the manager fails (due to an
 *            error connecting to pmon). This cannot work from within a PVSS00ctrl script, only a PVSS00ui script.
 * \param  openTcp Opens a TCP port in order to communicate with pmon if TRUE. This defaults to FALSE in pmonStartManager(),
 *            however it appears to be necessary for this to be TRUE when running inside a PVSS00ctrl manager. Perhaps it works
 *            within a PVSS00ui manager when FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * This function appears to work, but often still causes errors to be produced in the PVSS log. A fix is being worked on.
 * This function is probably no longer going to be maintained. It is left here simply as a starting point if this is ever
 * needed again... It also now won't work until the system name is properly determined...
 */
// void fwUkl1_startManager(const string &manager, const string &options, bool showErrDialog, bool openTcp, dyn_string &exceptionInfo) {
  // Keeps track of the errors.
//   int exInfoSize = dynlen(exceptionInfo);
//   
  // First we must identify the manager to start, ask the fwInstallation library to figure it out.
  // The postion of the manager that it found, this also doubles as an error code.
  // The function does have a return code but it is couple to an error in the position.
//   int manPos = -1;
//   fwInstallation_getManagerIndex(manager, options, manPos);
//   if ( -1 != manPos ) {
    // It appears that the PVSS console is counting from 0 and getMangerIndex returns the position not in the console
    // but the index in an array of all the entries. Hence it is out by 1!
//     --manPos;
    // Name of the host that is running the PVSS project.
//     string host;
    // Port that pmon is running on the host.
//     int port;
    // Get the host and port details for this project.
//     string sysName = "";// THIS NEEDS TO BE FOUND!!!
//     int iErr = paGetProjHostPort(sysName, host, port);
//     if (!iErr) {
      // Get the username and password for the project.
//       string sUser = "";
//       string sPassword = "";
//       fwInstallation_getPmonInfo(sUser, sPassword);
      // The command to be sent to pmon.
//       string command = "";
//       sprintf(command, "SINGLE_MGR:START %d", manPos);
//       command = sUser+"#"+sPassword+"#"+command;
//   
      // Don't show the error dialog, it can't be used in 
//       if ( pmon_command(command, host, port, showErrDialog, true) )
//         exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_startManager(): Failed to start the manager at position " + manPos + ". Tried to connect to pmon on host \'" + host + "\' on port " + port + ". The manager will have to be started manually.", "1");
        // Start the manager.
//         bool err = FALSE;
//         pmonStartManager(err, fwUkl1_getSystemName(), manPos, sUser, sPassword);
//         if (err) {
//           exInfoSize = fwUkl1Exception_raise(exceptionInfo, "FATAL", "fwUkl1_startManager(): Failed to start the post installation manager. Could not start the post install PVSS panel, this must now be done manually. The FSM will also need to be configured manually as it cannot be setup correctly until the constants have been set.", "1");
//         }// if(err)
//     }// if(!iErr)
//     else {
//       exInfoSize = fwUkl1Exception_raise(exceptionInfo, "ERROR", "fwUkl1_startManager(): Failed to retrieve host and port for the project " + fwUkl1_getSystemName() + ". The manager could not be started and will need to be started manually.", "1");
//     }// else(!iErr)
//   }// if(-1!=pos)
//   else {
//     exInfoSize = fwUkl1Exception_raise(exceptionInfo, "FATAL", "fwUkl1.postInstall(): Failed to identify the post installation manager to start. Could not start the post install PVSS panel, this must now be done manually. The FSM will also need to be configured manually as it cannot be setup correctly until the constants have been set.", "1");
//   }// else(-1!=pos)
// 
  // Done.
//   return;
// }// fwUkl1_startManager()

// @}
